#------------------------------------------------------------------------------
# Support functions for windows based icons
# Mostly copied from winico.tcl part of winico package
#------------------------------------------------------------------------------
set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/IconSupport.tcl,v 1.13 2004/02/05 05:02:58 wilkason Exp $
}
if {[catch {package require Winico} result]} {
   puts "No built-in or available Winco, using local DLL"
   load [file join [file dirname [info script]] myWinico winico03.dll]
}
#------------------------------------------------------------------------------
# Routines below are in the winico distribution
#------------------------------------------------------------------------------
proc winico_seticon { w icofile } {
   set icofile [winico_file $icofile]   ;# calls either wrapped or stub file
   set ico [winico create $icofile]
   set screendepth [winfo screendepth .]
   set bigsize "32x32"
   set bigpos -1
   set bigdepth 0
   set smallsize "16x16"
   set smallpos -1
   set smalldepth 0
   foreach i [winico info $ico] {
      array set opts $i
      set depth    $opts(-bpp)
      set pos      $opts(-pos)
      set geometry $opts(-geometry)
      if { $geometry=="$bigsize" && $depth<=$screendepth } {
         if { $depth>$bigdepth } {
            set bigpos $pos
            set bigdepth $depth
         }
      } elseif { $geometry=="$smallsize" && $depth<=$screendepth } {
         if { $depth>$smalldepth } {
            set smallpos $pos
            set smalldepth $depth
         }
      }
   }
   if { $bigpos==-1 && $smallpos==-1 } {
      puts stderr "couldn't find $bigsize and $smallsize icons in $icofile"
      return $ico
   } elseif { $bigpos==-1 } {
      set bigpos $smallpos
      puts stderr "couldn't find $bigsize icons in $icofile"
   } elseif { $smallpos==-1 } {
      set smallpos $bigpos
      puts stderr "couldn't find $smallsize icons in $icofile"
   }
   winico setwindow $w $ico big   $bigpos
   winico setwindow $w $ico small $smallpos
   return $ico
}
proc winico_delall {} {
   foreach i [winico info] { winico delete $i }
}
proc winico_loadicon { w symbol } {
   set ico [winico load $symbol]
   winico setwindow $w $ico big
   winico setwindow $w $ico small
}
#------------------------------------------------------------------------------
# Function   : winico_file
# Description: Dummy proc to just return the file passed in
#              replaced with wrap code if wrapped
#              wrapped code will write then read icon file to disk
# Author     : Tom Wilkason
# Date       : 1/27/2001
#------------------------------------------------------------------------------
if {[hasNamespace ::starkit]} {
   proc winico_file {icofile} {
      # Icon file must be located in substDir (i.e. same as main script)
      # Just read it from the wrapped location and write it out to
      # the TEMP folder for use by the icon utilities.
      if {[catch {
         # In
         set fin [open $icofile r]
         fconfigure $fin -translation binary
         # Out
         set tmpfile [file join $::env(TEMP) [file tail $icofile]]
         # puts "icon: $icofile with temp file as $tmpfile"
         if {! [file exists $tmpfile] } {
            set fout [open $tmpfile w]
            fconfigure $fout -translation binary
            # Move file to temp folder.
            puts -nonewline $fout [read $fin]
            close $fout
            set libMtime [file mtime $tmpfile]
            puts "Created $tmpfile with time [clock format $libMtime]"
         }
         close $fin
      } result]} {
         puts stderr "getIconScript error: $result\n$::errorInfo"
         return {}
      }
      return "$tmpfile"
   }
} else {
   proc winico_file {name} {
      # The snackAmp.ico HAS to be in the same folder
      # as the snackAmp.tcl file.
      return [file join $::TA(scriptHome) $name]
   }
}
#------------------------------------------------------------------------------
# Function   : iconSupport
# Description: Add icons in taskbar for Windows
# Author     : Tom Wilkason
# Date       : 1/9/2001
#------------------------------------------------------------------------------
proc iconSupport {args} {
   global TA
   if {[catch {
      set TA(IconFile) [file join $TA(snackRoot) icons snackAmp.ico]
      update
      #saLog "Installing icons '$TA(IconFile)'"
      set Icon [winico_seticon $::SA $TA(IconFile)]
      set TA(Icon) $Icon
      winico text $Icon {Snack Music Player, LMB to pause, RMB for miniPlayer}
      winico taskbar add $Icon -callback "package require miniPlayer;miniPlayer::iconCallback %m $Icon %x %y"
   } result]} {
      debug $result
   }
}
# proc iconText {text} {
#    global TA
#    if {[info exists TA(Icon)]} {
#       winico text $TA(Icon) $text
#    }
# }
package provide IconSupport 1.0


