#------------------------------------------------------------------------------
# autoPlayList Package 1.0
#
# Auto Playlist functions for snackAmp Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#
#------------------------------------------------------------------------------
set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/autoPlayList.tcl,v 1.78 2004/07/20 23:32:59 wilkason Exp $
}
package require snID3

namespace eval APL {
   variable folderList
   upvar #0 snackAmpSettings(folderList) folderList
   #upvar #0 snackAmpSettings(folderPos) folderPos
   variable folderPos -1
   variable folderNext
   variable folderPrev
   variable icons
   variable Table
   variable TableCols
   variable paneOrient
   variable lastAfter 0
   set icons [list "saImages::Tree.Folder" "saImages::Tree.ROM" "saImages::mp3_player"]
   set ::TA(Title,$::RMT) "/Removable Media"
   variable rmName $::TA(Title,$::RMT)
   option add *FolderTree*Font  treeFont startup
   option add *FolderTree*Menu.font  menu startup
   # Trace variables used during settings when adding/removing
   # a column from the table.
   foreach {key} [array names ::snackAmpSettings col,*,show] {
      trace variable ::snackAmpSettings($key) w APL::rebuildTableAfterIdle
   }
}
#------------------------------------------------------------------------------
# Function   : APL::rebuildTableAfterIdle
# Description: When setting the snackAmpSettings array, this callback will
#              be called for each column. We only want to rebuild once, so
#              we keep pushing it off until the last call.
# Author     : Tom Wilkason
# Date       : 8/7/2003
#------------------------------------------------------------------------------
proc APL::rebuildTableAfterIdle {variable index mode} {
   variable lastAfter
   global $variable
   after cancel $lastAfter
   set lastAfter [after 200 [list APL::rebuildTable .wFC $variable $index $mode]]
}
#------------------------------------------------------------------------------
# Function   : APL::Title
# Description: Change the title to the last file
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc APL::Title {args} {
   catch {
      wm title .wFC "SnackAmp Auto-Playlist Manager - [file rootname [file tail $::snackAmpSettings(lastFolderFile)]]"
   }
}
#------------------------------------------------------------------------------
# Function   : saveSize
# Description: Save the APL window sizes
# Author     : Tom Wilkason
# Date       : 1/9/2001
#------------------------------------------------------------------------------
proc APL::saveSize {args} {
   if {![catch {winfo toplevel $::TREE} tl]} then {
      set ::snackAmpSettings(treeSize) [wm geometry $tl]
      catch {set ::snackAmpSettings(FileListSize) [lindex [pane $tl.sv.lb -size] 0]}
   }
}
##
# Handlers for entry widgets
#
proc APL::invokeFindNextFile {w} {
   APL::findNextSong $::TA(SAFind)
   entryHistory::push $w
}
proc APL::invokeFindNextFolder {w} {
   APL::findNextFolder $::TA(SAFind)
   entryHistory::push $w
}
proc APL::invokeFilterTracks {w} {
   APL::setSearchFilter
   entryHistory::push $w
}
#------------------------------------------------------------------------------
# Function   : folderTree
# Description: Toss up a folder tree linked to a variable
# Author     : Tom Wilkason
# Date       : 1/9/2001
#------------------------------------------------------------------------------
proc APL::FolderTree {args} {
   package require saTree
   global TA snackAmpSettings
   variable folderNext
   variable folderPrev
   variable paneOrient
   ##
   # Create the folder tree entries if needed
   #
   set W .wFC
   if {[winfo exists $W]} {
      wm deiconify $W
      #raise $W      ;# This causes a noticible delay on KDE
      focus $W
      return
   }
   Wait 1
   if {[aplFolder::entries]==0} {
      readAPLFile $snackAmpSettings(lastFolderFile)
   }

   toplevel $W -cursor $::TA(arrow) -class FolderTree
   wm withdraw $W
   trace variable ::snackAmpSettings(lastFolderFile) w APL::Title
   trace variable ::TA(highlightPath) w {after idle [list APLcover::packCoverArt $::TA(highlightPath)]}
   APL::Title
   wm protocol $W WM_DELETE_WINDOW {
      APL::close
   }
   ##
   # Button bar
   #
   if {$::tcl_platform(platform)=="windows"} {
      frame $W.mb -relief groove -bd 2
   } else {
      frame $W.mb
   }
   pack $W.mb -side top -fill x -pady 3

   # Save Button
   pack [button $W.mb.save -image saImages::Save -command {macros::saveCurrentAPL 0}]\
      -padx 3 -side left -fill y
   balloon::button $W.mb.save "Save the current settings"

   # Save As button
   pack [button $W.mb.saveas -image saImages::SaveAs -command {macros::SaveFolderFile}]\
    -padx 3 -side left -fill y
   balloon::button $W.mb.saveas "Save Auto-Playlist under a new file name"

   # Apply button
   pack [button $W.mb.apply -image saImages::Apply -command {macros::applyCurrentAPL 0}]\
     -padx 3 -side left -fill y
   balloon::button $W.mb.apply "Update the playlist with the current settings"

   menubutton $W.mb.files -relief flat -image saImages::apllist -menu $W.mb.files.list

   # Refresh button
   pack [button $W.mb.refresh -image saImages::refresh -command {APL::indexFolderChildren $TA(highlightPath)}]\
     -padx 3 -side left -fill y
   balloon::button $W.mb.refresh "Re-Index selected folder and sub-folders"
   if {$gui::hasImg} {
      # Coverart Toggle
      pack [button $W.mb.covart -image saImages::coverart \
         -command APLcover::ToggleArt] \
         -padx 3 -side left -fill y
      balloon::button $W.mb.covart "Show or hide cover art for selected folder"
      set TA(aplArtButton) $W.mb.covart
   }

   # Filelist menu
   pack [button $W.mb.togpl -image saImages::FileList \
      -command "APL::setFileListToShow $W.sv.lb $W.mb.togpl"] \
      -padx 3 -side left -fill y
   balloon::button $W.mb.togpl "Depress to show files under the selected folder"

   # Initially, we toggle the setting to start the statemachine
   set ::snackAmpSettings(hideFileList) [expr {! ($::snackAmpSettings(hideFileList)&0x1)}]

   # Play button
   pack [button $W.mb.play -image saImages::noatun2-16 -command {aplLB::playSelectedFiles}] \
      -padx 3 -side left -fill y
   balloon::button $W.mb.play "Play Selection(s)"

   # Play button
   pack [button $W.mb.queue -image saImages::queue -command {aplLB::insertSelectedFiles}] \
      -padx 3 -side left -fill y
   balloon::button $W.mb.queue "Queue Selection(s)"

   # Level Buttons
   foreach {level} {0 1 2 3 4 5} {
      pack [radiobutton $W.mb.level$level -width 2 -text [expr {$level+1}] -value $level \
         -padx 0 -pady 0 -bd 1 -variable ::snackAmpSettings(treeDepth) -indicatoron 0 -relief flat \
          -relief flat -command {trackList::populateHighighted 1 1}] \
         -padx 0 -fill both -side left
      balloon::button $W.mb.level$level "Show an additional $level levels below the selected folder"
   }

   # Edit cell button
   pack [button $W.mb.edit -image saImages::edit-16 \
      -command trackList::changeEdit] \
      -padx 3 -side left
   balloon::button $W.mb.edit "Toggle Cell Editing"
   # Rating +/-
   pack [button $W.mb.plus -image saImages::iconthumbsup \
      -command {aplLB::bumpRating $::TA(highlightPath) 100 1}] \
      -padx 0 -side left -fill y
   balloon::button $W.mb.plus "Increase Rating of selected tracks or folders"
   pack [button $W.mb.minus -image saImages::iconthumbsdown \
      -command {aplLB::bumpRating $::TA(highlightPath) 100 -1}] \
      -padx 0 -side left -fill y
   balloon::button $W.mb.minus "Decrease Rating of selected tracks or folders"


   set trackList::editCellButton $W.mb.edit

   # Back button
   pack [button $W.mb.prev -image saImages::NavBack -command APL::prevSelection] \
      -padx 3 -side left -fill y
   balloon::button $W.mb.prev "Change to prior selected folder"


   # Dropdown history button (redundant with menu)
   pack [menubutton $W.mb.prevHistory -relief flat -image saImages::dropdownImage -menu $W.mb.prevHistory.list] -side left -fill y
   balloon::button $W.mb.prevHistory  "List of previous locations"
   # History menu, one for button and one for RMB
   menu $W.mb.prevHistory.list -postcommand [list APL::cbHistoryMenu $W.mb.prevHistory.list]

   # Forward Button
   pack [button $W.mb.next -image saImages::NavForward -command APL::nextSelection] \
      -padx 3 -side left -fill y
   balloon::button $W.mb.next "Change to next selected folder"

   ##
   # Find next entry
   #
   pack [button $W.mb.findFile -image saImages::findfile \
      -command [list APL::invokeFindNextFile $W.mb.find]] -side left -fill y
   balloon::button $W.mb.findFile "Press to search for next file containing text to the right."

   pack [button $W.mb.findLab -image saImages::findfolder \
      -command [list APL::invokeFindNextFolder $W.mb.find]] -side left -fill y
   balloon::button $W.mb.findLab "Press to search for next folder containing text to the right."

   pack [entryHistory::create $W.mb.find ::TA(SAFind) ::snackAmpSettings(findHistory)] \
      -side left -expand yes -fill x
   set ::SAFind [entryHistory::hull $W.mb.find]

   bind $::SAFind <Control-Return> [list APL::invokeFindNextFile $W.mb.find]
   bind $::SAFind <Return> [list APL::invokeFindNextFolder $W.mb.find]

   ##
   # Filter tracks
   #
   pack [button $W.mb.ckfilter -image saImages::filter \
      -command [list APL::invokeFilterTracks $W.mb.filter]] \
      -padx 3 -side left -fill y
   balloon::button $W.mb.ckfilter "Press search all media for tracks matching pattern to the right"

   pack [listOptionMenu $W.mb.searchWhat ::snackAmpSettings(FilterWhat) {Any File Title Artist Album Note Genre Folder Track Rate Tag Rating} 6] \
      -side left -expand no -fill x
   balloon::define $W.mb.searchWhat.e "Select what attribute to filter on"

   pack [entryHistory::create $W.mb.filter ::TA(SAFilter) ::snackAmpSettings(filterHistory)] \
      -side left -expand yes -fill x
   set ::SAFilter [entryHistory::hull $W.mb.filter]

   pack $::SAFilter -side left -expand yes -fill x
   bind $::SAFilter <Return> {APL::setSearchFilter}
   bind $::SAFilter <Key> {APL::incrSearchFilter %K}
   pack [button $W.mb.clfilter -text Reset \
      -command {APL::SelectFile $::TA(highlightPath)}] \
      -padx 3 -side left -fill y
   balloon::button $W.mb.clfilter "Clear search entry and return to tree view of tracks"

   balloon::define $::SAFind "Enter text then hit return to search for next folder with matching text. Enter Ctrl-Return to search for next song with matching text. You may use wildcards for searching."
   balloon::define $::SAFilter "Enter text then depress the filter button to the left to restrict track names that contain this text."
   # Save the button names
   set folderNext $W.mb.next
   set folderPrev $W.mb.prev

   set Folders [menu $W.mb.files.list -tearoff 1 -postcommand "[list gui::APLFiles $W.mb.files.list]" ]
   set Removal [menu $W.mb.files.list.remove -postcommand "[list gui::APLFilesRemove $W.mb.files.list.remove]"]

   $Folders add command -label "Save Auto-Playlist As..." -command SaveFolderFile
#   menu $Folders.subsort
#   gui::aplSubsortMenu $Folders
   $Folders add cascade -label "Remove..." -menu $Removal
   $Folders add separator

#   pack [button $W.mb.close -image saImages::Close -command  {APL::close} \
#      -borderwidth 1] -side left
#   balloon::button $W.mb.close "Close this window"

   ##
   # Label and File List check button


   ##
   # Frame to hold the tree/filelist
   #
   pack [frame $W.sv -bd 2] -side top -fill both -expand true
   #
   set TA(ck,selected)     "Checked: 0"
   set TA(ck,unselected)   "Unchecked: 0"
   set TA(ck,total)        "Total: 0"
   set TA(ck,totalFolders) "Files: 0"
   set TA(ck,totalTime)    "Time: 00:00:00"
   set TA(ck,totalSize)    "Size: 0"

   pack [frame $W.file -relief flat] -fill x
   set i 0
   # TFW: Can't get grid to force them to the right like I would think -sticky e would
   set statusWidgets {
      selected   12 left
      unselected 14 left
      total      10 left
   }
   # Fast free space
   if {[string length [info command ntFreeSpace]] > 0} {
      set TA(ck,free)         "Free 0"
      lappend statusWidgets free 13 left
   }
   lappend statusWidgets totalFolders 10 right totalSize 16 right totalTime 14 right
   foreach {item wid where} $statusWidgets {
      pack [label $W.file.$item \
         -textvariable TA(ck,$item) -width $wid  -bd 1 -relief sunken -padx 2 -pady 1] \
         -side $where -padx 1
   }
   # todo: Bindsource on the checked label to drag checked selected folder tree files
   # todo: Bindsource on the total label to drag entire selected folder tree files
   if {![catch {package require tkdnd}] } then {

      dnd bindsource $W.file.selected text/uri-list {APL::dndSrcCheckedTreeDrag}
      bind $W.file.selected <Control-1> {dnd drag %W -actions copy;break}
      balloon::define $W.file.selected "Control-Drag tracks from highlighted folder(s)"

      dnd bindsource $W.file.total text/uri-list {APL::dndSrcTotalTreeDrag}
      bind $W.file.total <Control-1> {dnd drag %W -actions copy;break}
      balloon::define $W.file.total "Control-Drag selected tracks from highlighted folder(s)"
      bindtags $W [concat dndTag [bindtags $W]]
   }
   ##
   # Frame to hold the tree and scrollbars, pane packs it
   #
   frame $W.sv.tree -bd 1 -relief sunken ;# "placed" by pane'r

   set ::TREE $W.sv.tree.w  ;# Canvas
   scrollbar $W.sv.tree.sb -orient vertical -command [list $::TREE yview]

   pack $W.sv.tree.sb -side right -fill y
   ##
   # Build the guts of the tree (menus are created here)
   #
   APL::buildAPL $::TREE
   mscroll $::TREE
   ##
   # File listbox/Table
   #
   set ::PLB [trackList::buildTable $W.sv.lb ::snackAmpSettings]
   ##
   # Create the appropriate bindings
   #
   aplLB::bindings
   bindings::assignHotKeys
   APL::treeBind
   # Keys for the APL search box
   APLFilterAssign <Return> {
      APL::findNextFolder $::TA(SAFind)
   }
   APLFilterAssign <Control-Return> {
      APL::findNextSong $::TA(SAFind)
   }
   ##
   # Top Menus
   #
   APL::aplMenu $W
   ##
   # Pane them
   #
   package require saPane
   pane $W.sv.tree -orient $paneOrient -dynamic 0
   APL::setFileListToShow $W.sv.lb $W.mb.togpl
   ##
   # Table/Listbox Menus
   #
   aplLB::menus
   # Expand initially to folder level on menu
   if {[info exists ::snackAmpSettings(treeSize)]} {
      wm geometry $W $::snackAmpSettings(treeSize)
   }
   # Open from the top 1 layer down
##########   update idletasks
   Tree::open $::TREE / 1
   wm deiconify $W
   focus $W
   after 100 [list myIcon $W]
   Wait 0
   after 100 APL::bindTkDND
}

#------------------------------------------------------------------------------
# Bindings for TKDND
#------------------------------------------------------------------------------
proc APL::bindTkDND {args} {
   ##
   # DND
   #
   if {![catch {package require tkdnd}] } then {

      # When active over the tree, the following <Drag> event will be called
      # This kills the drop operation however
      #dnd bindtarget $::TREE text/uri-list <Drag> {
      #   Tree::setselection %W [Tree::labelat %W %x %y]
      #}
      # When dropping files in tree...
      dnd bindtarget $::TREE text/uri-list <Drop> {APL::dndDropOnTree %x %y %D %A %T %d %m}
      #%m is modifier key pressed
      # Define the APL Listbox as a source, to drag to one of
      #- LB -> Insert files
      #- TREE -> Move files
      #- Explorer -> copy files
      dnd bindsource [$::PLB bodypath] text/uri-list {APL::dndSrcLBDrag %W}
      bind [$::PLB bodypath] <Control-1> {dnd drag %W -actions copy;break}

      # Bind the drag to the mouse, for shift must release shift prior to the
      # drop (for windows anyway)
      # Bind the tree as a valid source to drag data from
      # relying on mouse binding to put the folder/tracks in the array
      dnd bindsource $::TREE text/uri-list {APL::dndSrcTreeDrag}
      bind $::TREE <Control-1> {dnd drag %W -actions copy;break}
   }
}
#------------------------------------------------------------------------------
# Function   : APL::rebuildTable
# Description: Rebuild the table when snackAmpSettings columns change, called
#              from a trace writing to one of the column settings.
#              Rebuild if the col changed from the current value.
# Author     : Tom Wilkason
# Date       : 5/27/2002
#------------------------------------------------------------------------------
proc APL::rebuildTable {W _variable index mode} {
   Trace
   global TA
   if {[catch {
      variable paneOrient
      upvar #0 $_variable settings
      set what $W.sv.lb
      # Add check to see if cols really need to be reconfigured
      if {[winfo exists $what]} {
         ##
         # Make sure we have the columns in the correct order
         #
         set col2Use [list]
         foreach {xx col xx} [$::PLB cget -columns] {
            lappend col2Use $col
         }
         set ::TA(cols) [orderList $::TA(cols) $col2Use]
         set wasOpen [winfo ismapped $what]
         foreach {xx col mode} [split $index ,] {break} ;# mode s/b "show"
         set geom [wm geometry $W]
         set ::snackAmpSettings(FileListSize) [lindex [pane $what -size] 0]
         destroy $what
         set trackList::TABLEDATA [list]

         set ::PLB [trackList::buildTable $what ::snackAmpSettings]
         if {$trackList::editCellState} {
            trackList::changeEdit
         }
         ##
         # Table/Listbox Menus
         #
         aplLB::menus
         ##
         # Create the appropriate bindings
         #
         aplLB::bindings
         if {$wasOpen} {
            pane $what -orient $paneOrient -dynamic 0 \
               -size $::snackAmpSettings(FileListSize)
            aplLB::UpdateList $::TA(highlightPath) 1
         }
         wm geometry $W $geom
         APL::bindTkDND
      }
   } result] } then {
      saLog "$::errorInfo"
   }
}
#------------------------------------------------------------------------------
# Function   : APL::close
# Description: Close the window asking to save changes if needed
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc APL::close {args} {
   $::TA(treeButton) configure -relief flat
   if {[Tree::isDirty $::TREE]} {
      if {[tk_messageBox -type yesno -icon question -message "Do you want to save your current Auto-playlist?"] == "yes"} {
         saveCurrentAPL 0
      }
   }
   wm withdraw .wFC
   APL::saveSize
   db::Commit
   trackList::saveColumnWidths
}
#----------------------------------------------------------------------------
# Grab some cover art and dump it into the selected folder
# Also handles native files dropped (called once per file) -> not implemented yet
#----------------------------------------------------------------------------
proc APL::dndDropOnTree {x y urls args} {
   Trace
   foreach {url} $urls {
      set lbl [Tree::labelat $::TREE $x $y]
      if {[string length $lbl]} {
         if {[catch {APL::GetTreeFolder $::TREE $lbl} folderName]} {
            saLog $folderName
            return
         }
         switch -- [file extension $url] {
            .jpg -
            .png {
               macros::copyImageToFolder $url $folderName
            }
            default {saLog "$url is not supported here"}
         }
      }
   }
}

#------------------------------------------------------------------------------
# Function   : APL::cbDndPLBToTree
# Description: Callback on drag-n-drop target from PLB to Tree
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc APL::cbDndPLBToTree {X Y} {
   Trace
   set x [expr {$X - [winfo rootx $::TREE]}]
   set y [expr {$Y - [winfo rooty $::TREE]}]
   set lbl [Tree::labelat $::TREE $x $y]
   if {$lbl != {}} {
      if {![catch {APL::GetTreeFolder $::TREE $lbl} folderName]} {
         # We let go over a node, ask to copy files
         if {[catalog::isMediaPlayer $folderName $::TA(highlightPath)]} {
            if {[catalog::copyClipboardFiles $folderName] > 0} {
               catch {
                  aplLB::UpdateList $::TA(highlightPath) [winfo ismapped $::PLB]
                  APL::treeSelect $::TA(highlightPath)
               }
            }
         } else {
            if {[catalog::moveClipboardFiles $folderName] > 0} {
               catch {
                  aplLB::UpdateList $::TA(highlightPath) [winfo ismapped $::PLB]
                  APL::treeSelect $::TA(highlightPath)
               }
            }
         }
      }
   }
}
#------------------------------------------------------------------------------
# Function   : APL::cbDndPLBToLB
# Description: Callback on drag-n-drop target from PLB to LB
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc APL::cbDndPLBToLB {x y} {
   global LB
   Trace
   if {$x>0} {
      set new [$LB nearest [expr {$y - [winfo rooty $LB]}]]
   } else {
      set new end
   }
   if {[catch {selection get -selection CLIPBOARD} tryFiles]} {
      return
   }
   clipboard clear
   set files [list]
   # Ignore errors due to funky clipboard data
   foreach {file} $tryFiles {
      if {[lsearch -exact $files $file] < 0} {
         lappend files $file
      }
   }
   LB::insertList $new $files
   return 1
}
#------------------------------------------------------------------------------
# Function   : APL::cbDndTreeToClipboard
# Description: Callback on drag-n-drop target from Tree to PlayList
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc APL::cbDndTreeToClipboard {X Y} {
   Trace
   set src $::TA(aplDNDSource)
   clipboard clear
   set data [list]
   ##
   # Would be nice to grab the data in the sort order of the tablelist
   #
   foreach {entry} [db::filesMatchingPattern "$src/*"] {
      lappend data $entry
   }
   clipboard append [lsort -dictionary $data]
}
#------------------------------------------------------------------------------
# Function   : APL::cbDndTreeToTree
# Description: Callback when moving folders in the APL tree
# Author     : Tom Wilkason
# Date       : 6/3/2002
#------------------------------------------------------------------------------
proc APL::cbDndTreeToTree {W lbl} {
   if {![catch {APL::GetTreeFolder $W $lbl} folderName]} {
      set newName [file join $folderName [file tail $::TA(aplDNDSource)]]
      if {![string equal $::TA(aplDNDSource) $folderName] && ![string equal $::TA(aplDNDSource) $newName]} {
         set ans [tk_messageBox -type yesno -icon question \
               -message "Are you sure you want to move/copy \n'$::TA(aplDNDSource)'\n  to\n'$newName'"]
         if {$ans=="yes"} {
            if {[catch {
                  if {[catalog::isMediaPlayer $newName $::TA(aplDNDSource)]} {
                     catalog::copyFolder $::TA(aplDNDSource) $newName
                  } else {
                     catalog::moveFolder $::TA(aplDNDSource) $newName
                  }
                  if {$::snackAmpSettings(dndExpand)} {
                     # Highlight and expand the new target
                     APL::treeOpen [file dirname $newName]
                     APL::treeSelect $newName
                     APL::viewHighlighted $::TREE
                  }
               } result] \
                  } then {
               tk_messageBox -type ok -icon warning \
                     -message "Failed to move/copy $::TA(aplDNDSource) under $folderName\n$result\n$::errorInfo"
            }
         } else {
            APL::treeSelect $::TA(aplDNDSource)
         }
         set ::TA(aplDND) 0
      }
   }
}

#------------------------------------------------------------------------------
# Function   : APL::findNextSong
# Description: Highlight the next song in tree with matching text
# Author     : Tom Wilkason
# Date       : 3/28/2002
#------------------------------------------------------------------------------
proc APL::findNextSong {text} {
   if {[catch {aplLB::getActiveFullName} fullName]} {
      set pos 0
   } else {
      # Locate where song is now, and search from there
      set pos [db::whereInDb $fullName]
      if {$pos < 0} {
         set pos 0
      }
   }
   # Search from current place down
   set npos [db::matchInDb "*$text*" [incr pos]]
   if {$npos >= 0} {
      SelectFile [db::fileByPosition $npos] 1
      return 1
   } else {
      # search from top
      set npos [db::matchInDb "*$text*" 0]
      if {$npos >= 0} {
         SelectFile [db::fileByPosition $npos] 1
         return 1
      }
   }
   return 0
}
#------------------------------------------------------------------------------
# Function   : APL::findAllSongs
# Description: Locate all tracks containing the text
# Author     : Tom Wilkason
# Date       : 3/28/2002
#------------------------------------------------------------------------------
proc APL::findAllSongs {text} {
   return 0
}
#------------------------------------------------------------------------------
# Function   : APL::positionFolder
# Description: Bound to cursor keys, this moves the selection within the tree
# Author     : Tom Wilkason
# Date       : 3/28/2002
#------------------------------------------------------------------------------
proc APL::positionFolder {how} {
   set current $::TA(highlightPath)
   set new ""
   # Move to the appropriate folder
   switch -- $how {
      prev {
         set folders [aplFolder::getFolders [file dirname $current] 1]
         set pos [sortedSearch $folders $current]
         if {$pos > -1} {
            set new [lindex $folders [incr pos -1]]
         }
      }
      next {
         set folders [aplFolder::getFolders [file dirname $current] 1]
         set pos [sortedSearch $folders $current]
         if {$pos > -1} {
            set new [lindex $folders [incr pos 1]]
         }
      }
      left {
         # Go up one
         set new [file dirname $current]
      }
      right {
         set folders [aplFolder::getFolders $current 1]
         set pos [sortedSearch $folders $current]
         if {$pos > -1} {
            set new [lindex $folders [incr pos 1]]
         }
      }
      default {
      }
   }
   if {[file exists $new]} {
      SelectFile $new 1
   }
}
#------------------------------------------------------------------------------
# Function   : APL::findNextFolder
# Description: Highlight the next folder in tree with matching text
# Author     : Tom Wilkason
# Date       : 3/28/2002
#------------------------------------------------------------------------------
proc APL::findNextFolder {text} {
   set folders [aplFolder::foldersMatchingPattern * 1]
   if {[catch {sortedSearch $folders $::TA(highlightPath)} pos]} {
      set pos -1
   }
   # Now search from current place
   foreach {folder} [lrange $folders [incr pos] end] {
      if {[string match -nocase "*$text*" [file tail $folder]]} {
         SelectFile $folder 1
         return 1
      }
   }
   ##
   # Didn't find from current pos, start from top (wrap)
   #
   foreach {folder} [lrange $folders 0 $pos] {
      if {[string match -nocase "*$text*" [file tail $folder]]} {
         SelectFile $folder 1
         return 1
      }
   }
}
#------------------------------------------------------------------------------
# Function   : APL::prevSelection
# Description: Move to the last know selected folder
# Author     : Tom Wilkason
# Date       : 3/28/2002
#------------------------------------------------------------------------------
proc APL::prevSelection {args} {
   variable folderList
   variable folderPos
   variable folderNext
   variable folderPrev
   if {$folderPos > 0} {
      incr folderPos -1
      set folder [lindex $folderList $folderPos]
      SelectFile $folder 0
   }
   APL::setNavButtons $folderNext $folderPrev
}
#------------------------------------------------------------------------------
# Function   : APL::prevSelection
# Description: Move to the next know selected folder
# Author     : Tom Wilkason
# Date       : 3/28/2002
#------------------------------------------------------------------------------
proc APL::nextSelection {args} {
   variable folderList
   variable folderPos
   variable folderNext
   variable folderPrev
   if {($folderPos+1)  < [llength $folderList] && $folderPos >= 0} {
      incr folderPos 1
      set folder [lindex $folderList $folderPos]
      SelectFile $folder 0
   }
   APL::setNavButtons $folderNext $folderPrev
}
#------------------------------------------------------------------------------
# Function   : insertUnique
# Description: Insert an item in the of a list if it is unique
#              folderPos points to location in list of current selection
#              if list is empty, it will be -1 (initial condition)
#              When called, the next entry will be inserted into the list
#              after the current item and the pointer incremented.
#              This only will occur if the item at the current position
#              does not equal the item to be inserted
# Author     : Tom Wilkason
# Date       : 3/28/2002
#------------------------------------------------------------------------------
proc APL::insertUnique {item} {
   variable folderList
   variable folderPos
   variable folderNext
   variable folderPrev
   set curItem [lindex $folderList $folderPos]
   if {![string equal $curItem $item]} {
      incr folderPos  ;# Point to next item
      set folderList [linsert $folderList $folderPos $item]
   }
   APL::setNavButtons $folderNext $folderPrev
}
#------------------------------------------------------------------------------
# Function   : APL::historyMenu
# Description: Display a history menu
# Author     : Tom Wilkason
# Date       : 3/5/2002
#------------------------------------------------------------------------------
proc APL::cbHistoryMenu {menu} {
   variable folderList
   variable folderPos
   variable folderNext
   variable folderPrev

   $menu delete 0 end
   # Clamp menu size to 20 high
   set mh 20
   set pos $folderPos
   set end [expr {[llength $folderList]-1}]
   set min [expr {$pos - $mh/2}]
   # If closer than mh/2 from end, offset min back by delta
   if {($end-$pos) < $mh/2} {
      incr min [expr {($end-$pos) - $mh/2}]
   }
   # Clamp min
   if {$end < $mh || $min < 0} {
      set min 0
   }
   # Display up to $mh menus
   set i 0
   set cnt 0
   foreach item $folderList {
      if {$i >= $min && $cnt <= $mh} {
         $menu add radiobutton -variable APL::folderPos -value $i \
            -label "$i  $item" -command [list APL::positionHistory $i]
         incr cnt
      }
      incr i
   }
}
#------------------------------------------------------------------------------
# Function   : APL::positionHistory
# Description: Position the folder at some location and update the buttons
# Author     : Tom Wilkason
# Date       : 3/5/2002
#------------------------------------------------------------------------------
proc APL::positionHistory {location} {
   variable folderList
   variable folderPos
   variable folderNext
   variable folderPrev
   set folder [lindex $folderList $folderPos]
   SelectFile $folder 0
   APL::setNavButtons $folderNext $folderPrev

}
#------------------------------------------------------------------------------
# Function   : APL::setNavButtons
# Description: Set the state of the navigation buttons
# Author     : Tom Wilkason
# Date       : 3/5/2002
#------------------------------------------------------------------------------
proc APL::setNavButtons {next prev} {
   variable folderList
   variable folderPos
   # Configure previous button
   if {$folderPos > 0} {
      $prev config -state normal
   } else {
      $prev config -state disabled
   }
   # configuration next button
   if {$folderPos < ([llength $folderList]-1)} {
      $next config -state normal
   } else {
      $next config -state disabled
   }
}
#------------------------------------------------------------------------------
# Function   : APL::popWatch
# Description: Tempoaraly put a busy watch up
# Author     : Tom Wilkason
# Date       : 3/5/2002
#------------------------------------------------------------------------------
proc APL::popWatch {{time 300}} {
   APL::busyCursor 1
   update idletasks
   after 250 APL::busyCursor 0
}

#------------------------------------------------------------------------------
# Fire a filter action on incoming key strokes, only entered characters
# will fire it.
#------------------------------------------------------------------------------
proc APL::incrSearchFilter {Key} {
   if {$::snackAmpSettings(incrKeyFilter)} {
      after cancel APL::setSearchFilter
      if {[string length $Key]==1 && [string length $::TA(SAFilter)]>1} {
         after 400 APL::setSearchFilter
      }
   }
}
#------------------------------------------------------------------------------
# Toggle the filter button and update the track listing
#------------------------------------------------------------------------------
proc APL::setSearchFilter {args} {
   APL::setFilter 1
   # aplLB::UpdateList "" 1
   trackList::populate "" $::PLB 1 1

}
#----------------------------------------------------------------------------
# Set the filter to some mode
#----------------------------------------------------------------------------
proc APL::setFilter {how} {
   set ::TA(useSAFilter) $how
}

#------------------------------------------------------------------------------
# Function   : APL::aplMenu
# Description: Create the main menu for the APL Window, reuse
#              pop-up menus for the tree and table.
# Author     : Tom Wilkason
# Date       : 5/27/2002
#------------------------------------------------------------------------------
proc APL::aplMenu {W} {
   global SA
   set M $W.aplMenu
   bindings::init $W menuBindings
   # This forces a moden L&F, top menus are show
   # as raised, sub-menus are hightlight selected
   menu $M \
      -activebackground [$SA cget -background] \
      -activeforeground [$SA cget -highlightcolor] \
      -activeborderwidth 1 -font menu

   # attach it to the main window
   $W config -menu $M
   # Create more cascade menus
   foreach {m u} {File 0 Edit 0 Folders 1 Tracks 0 AutoPlaylists 0 Columns 2 Help 0} {
      switch -- $m {
         Folders {
            # Use existing menu
            $M add cascade -label $m -menu $::TREE.popup  -underline $u
         }
         Tracks {
            # Use existing menu
            $M add cascade -label $m -menu $::PLB.popup  -underline $u
         }
         AutoPlaylists {
            # Use existing menu
            $M add cascade -label $m -menu $W.mb.files.list -underline $u
         }
         Columns {
            set $m [menu $M.m$m -tearoff 1]
            $M add cascade -label $m -menu $M.m$m  -underline $u
            foreach {item} [lrange $::TA(cols) 1 end] {
               $Columns add checkbutton -label $item -variable ::snackAmpSettings(col,$item,show)
            }
         }
         Subfolders {
            set $m [menu $M.m$m]
            $M add cascade -label $m -menu $M.m$m  -underline $u
            foreach {subfolders} {0 1 2 3 4 5 6 7 8} {
               $Subfolders add radiobutton -label "Show [expr {$subfolders+1}] levels" -value $subfolders -variable ::snackAmpSettings(treeDepth)
            }
         }
         History {
            # Use existing menu
            $M add cascade -label $m -menu $W.mb.prevHistory.list  -underline $u
         }
         default {
            set $m [menu $M.m$m]
            $M add cascade -label $m -menu $M.m$m  -underline $u
         }
      }
   } ;#end foreach
   ##
   # Here is how to add a menu to the "system" menu
   # Use it to put "pinned" command
   #
   ontop::pinOnSystemMenu $M .wFC
   ##
   # File Menu
   #
   bindings::addCommand $File "Manage Media Library..."  {catalog::manageMediaFolders} 0
   bindings::addCommand $File "Manage Portable Player Library..."  {catalog::managePlayerFolders} 7
   bindings::addCommand $File "Index Media Library" {catalog::IndexAskFirst} 1
   bindings::addCommand $File "Check for Media Changes" {APL::rebuildIfNewMedia} 1
   $File add separator
IFDEF RMS {
   bindings::addCommand $File "Catalog Removable Media" catalog::CatalogMedia
   bindings::addCommand $File "Mount Removable Media"  {catalog::MountMedia "Removable"}
   $File add separator
   bindings::addCommand $File "Manage Removable Folders..." {
      package require saSettings
      folderManager::launch RemovableFolders
   }
   $File add separator
}
   bindings::addCommand $File "Save Auto-Playlist" {saveCurrentAPL 0} 0
   bindings::addCommand $File "Save Auto-Playlist As..." {SaveFolderFile}
   #bindings::addCommand $File "Update Auto-Playlist" {applyCurrentAPL 0}
   bindings::addCommand $File "Apply Auto-Playlist" {applyCurrentAPL 1} 0
   $File add separator
   bindings::addCommand $File "Sync to Playing" {SelectFile 1} 1
   bindings::addCommand $File "Sync to Playlist Selected" {SelectFile 0}
   $File add separator
   bindings::addCommand $File "Close" {APL::close} 0
   # Edit menu
   bindings::addCommand $Edit "Copy Selection" {APL::cut} 0
   bindings::addCommand $Edit "Paste" {APL::PasteFolderToTree} 0
   $Edit add separator
   $Edit add checkbutton -label "Edit Cell Mode" -variable trackList::editCellCheckState \
      -command {trackList::changeEdit}
   $Help add command -label "About" -command \
   {
      tk_messageBox -type ok -icon info -title "Help" -message \
"The Autoplaylist Manager allow allows you to set/save/restore preferred folders.
You can also rename, remove and move files. With the mouse you can:
- Select a folder to include (checked) or exclude (unchecked)
- Control Click: Check selected and all subfolders
- Alt Click: Uncheck selected and all subfolders
- Double Click: Toggle expand/collapse selected folder
- Right MB Click: Menu allowing you do above actions, apply and/or
  save the settings.
   - Rescan folder, sync internal list with any changes that has taken place
     on the disk.
   - Add folder to playlist, copy all of this folders files to the main
     playlist.
   - Move files copied to clipboard to highlighted folder
   - Create a new folder or rename and existing folder
   - Remove a folder and all of its subfolders and files.
- Using drag and drop you can move an entire tree limb from one
  folder to another.

From the menu and menu bar you can:
- Apply the settings to the playlist (does not save to file)
- Save, applies and saves the settings to the current Auto-playlist file
- Save as, save the settings to a new file name and make it current.
- File List checkbox, toggle a view of the files contained within the
  selected folder at the bottom.
- Close the window

You can drag highlighted files from the playlist to a folder
in the tree view to move them.
"
   }

}
proc APL::SelectedFolderRenameInPlace {{select 0}} {
   set W $::TREE
   # Note access into tree global here
   set tag [Tree::activetag $::TREE]
   if {[string length $tag] == 0} {
      return
   }
   set w  .canRndialog
   destroy $w
   toplevel $w -class SnackAmp
   set oldName [file tail $::TA(highlightPath)]
   set folder $::TA(highlightPath)
   set base [file dirname $folder]
   foreach {x y - -} [$W bbox $tag] {break}
   # Need to account for scrolled region
   # bu subtracting out canvas displayed 0 value
   set rx [winfo rootx $W]
   set ry [winfo rooty $W]
   set X [expr {$rx+$x-round([$W canvasx 0])}]
   set Y [expr {$ry+$y-round([$W canvasy 0])}]
   set font [$W itemcget $tag -font]
   set bg [$W cget -background]
   set fg $Tree::fg
   set ::TA(rename) $oldName
   set wid [expr {[string length $oldName]+5}]
   pack [entry $w.rn -bd 2 -relief groove -font $font -insertbackground $fg \
      -width $wid -bg $bg -fg $fg -textvariable ::TA(rename)]
   wm withdraw $w
   wm overrideredirect $w 1
   wm transient $w [winfo toplevel [focus]]
   if {$select} {
      $w.rn selection range 0 end
   }
   focus -force $w.rn
   # with it hiden, let the layout take place so we know its geometry
   update idletasks
   wm geometry $w "+$X+$Y"
   bind $w <FocusOut> {
      destroy .canRndialog
      set ::TA(rename) ""
   }
   bind $w <Escape> {
      destroy .canRndialog
      set ::TA(rename) ""
   }
   bind $w <Return> [list destroy .canRndialog]
   wm deiconify $w
   # Force on top if possible
   update idletasks
   if {$ontop::enabled && [ontop::state $W]} {
      ontop::control $w 1
   }
   ##
   # Wait for window to close
   #
   tkwait window $w
   ##
   # Back from window closure
   #
   if {[string length $::TA(rename)]>0 && $::TA(rename) != $oldName} {
      if {[catch {
         set newFullName [file join $base $::TA(rename)]
         catalog::moveFolder $folder $newFullName
         APL::treeSelect $newFullName
         APL::viewHighlighted $::TREE
         catalog::FlagDirtyCache $newFullName
         set ::TA(highlightPath) $newFullName
         set res 1
      } result] \
      } then {
         tk_messageBox -type ok -icon warning -message "Could not rename folder \n$folder to \n$newFullName\n$result"
         set res 0
      }
   }
}
#------------------------------------------------------------------------------
# Function   : APL::treeBind
# Description: Add bindings to the autoplay list tree
# Author     : Tom Wilkason
# Date       : 12/23/2001
#------------------------------------------------------------------------------
proc APL::treeBind {args} {
   Trace
   ##
   # Pop-up menu, save some data for use by callbacks
   #
   bind $::TREE <ButtonPress-3> {
      set lbl [Tree::labelat %W %x %y]
      Tree::setselection %W $lbl
      if {$lbl != ""} {
         if {[catch {APL::GetTreeFolder %W $lbl} folderName]} {
            return
         }
         APL::insertUnique $folderName
         set TA(highlightPath) $folderName
         #set TA(highlightEntry) $lbl
         #set TA(treeID) %W
         tk_popup %W.popup %X %Y
      }
      catch {unset lbl}
   }
   set ::TA(aplDND) 0
   ##
   # macros::Highlight the current folder, save away the selected folder
   #
   APL_TreeNodeAssign <ButtonPress-1> {
      catch {
         focus [winfo parent %W]
         set lbl [Tree::labelat %W %x %y]
         Tree::setselection %W $lbl
         if {![catch {APL::GetTreeFolder %W $lbl} folderName]} {
            set ::TA(highlightPath) $folderName
            set ::TA(aplDNDSource) $folderName
            set ::TA(aplDragTarget) $folderName
            APL::insertUnique $folderName
            APL::setFilter 0
# This interfears with the tcl only DND, I need to change completely over to tkdnd
# Once I am sure it runs on linux.
            after idle [list aplLB::UpdateList $folderName [winfo ismapped $::PLB]]
         }
         # Update file listbox if it exists.
      }
   }
   ##
   # During a drag, highlight the current folder under the cursor
   #
   APL_TreeNodeAssign <B1-Motion> {
      set lbl [Tree::labelat %W %x %y]
      if {![catch {APL::GetTreeFolder %W $lbl} folderName]} {
         if {![string match $folderName $::TA(aplDragTarget)]} {
            Tree::setselection %W $lbl
            set ::TA(aplDragTarget) $folderName
         }
      }
      $::TREE config -cursor hand2
      set ::TA(aplDND) 1
      catch {unset lbl}
   }
   ##
   # Drop action coming from the Tree
   #
   APL_TreeNodeAssign <ButtonRelease-1> {
      APL::busyCursor 0
      set lbl [Tree::labelat %W %x %y]
      set targetWin [winfo containing %X %Y]
      if {[winfo exists $targetWin] && $::TA(aplDND)} {
         APL::busyCursor 1
         if {$targetWin == $::TREE} {   ;# move the folder
            APL::cbDndTreeToTree %W $lbl
         } elseif {$targetWin == [LB::body]} {
            APL::cbDndTreeToClipboard %X %Y   ;# get to clipboard
            APL::cbDndPLBToLB %X %Y    ;# now works like PLB to LB
         } elseif {[winfo toplevel $targetWin] == [winfo toplevel $::LB]} {
            # Improve to append to bottom here
            APL::cbDndTreeToClipboard %X %Y   ;# get to clipboard
            APL::cbDndPLBToLB 0 0    ;# now works like PLB to LB
         }
         APL::busyCursor 0
      }
      catch {unset lbl}
   }
   ##
   # Paste Files/folders (not on canvas)
   #
   APL_TreeAssign <<Paste>> {APL::PasteFolderToTree}
   ##
   # Append selected folder to clipboard
   #
   APL_TreeAssign <<Copy>> {APL::popWatch;clipboard append "[list $TA(highlightPath)] "}
   APL_TreeAssign <<Cut>> {APL::popWatch;clipboard append "[list $TA(highlightPath)] "}
   ##
   # Cursor movement keys
   #
   APL_TreeAssign <Left> {APL::positionFolder left}
   APL_TreeAssign <Right> {APL::positionFolder right}
   APL_TreeAssign <Up>   {APL::positionFolder prev}
   APL_TreeAssign <Down> {APL::positionFolder next}
   ##
   # Assign hotkey to rename
   #
   APL_TreeAssign <F2> {APL::SelectedFolderRenameInPlace}
   APL_TreeAssign <Delete> {::catalog::removeFolder $::TA(highlightPath)}
   # Toggle open/close on Return key
   APL_TreeAssign <Return> {Tree::toggleExpand $::TREE [APL::pathToTreeName $::TA(highlightPath)]}
   APL_ListAssign <Escape> {
      if {$trackList::editCellState} {
         trackList::changeEdit
      }
   }
   #APL_TreeAssign <F2> {::catalog::renameFolder $::TA(highlightPath)}
   ##
   # Set this and all child checkboxes
   #
#    APL_TreeNodeAssign <Control-1> {
#       set lbl [Tree::labelat %W %x %y]
#       Tree::setselection %W $lbl
#       Tree::setDirty $::TREE
#       if {![catch {APL::GetTreeFolder %W $lbl} folderName]} {
#          APL::setSomeFolders $folderName 1
#       }
#       APL::insertUnique $folderName
#    }
   ##
   # Clear this and all child checkboxes
   #
#    APL_TreeNodeAssign <Alt-1> {
#       set lbl [Tree::labelat %W %x %y]
#       Tree::setDirty $::TREE
#       Tree::setselection %W $lbl
#       if {![catch {APL::GetTreeFolder %W $lbl} folderName]} {
#          APL::setSomeFolders $folderName 0
#       }
#       APL::insertUnique $folderName
#    }
   ##
   # Recurse expand on double click
   #
   APL_TreeNodeAssign <Double-1> {
      APL::busyCursor 1
      Tree::toggleExpand %W [Tree::labelat %W %x %y] 10
      APL::busyCursor 0
   }
   ##
   # Now build the pop-up menu for the tree
   #
   set m $::TREE.popup
   menu $m
   set P [winfo parent $::TREE]
   bindings::init $P menuBindings
   menu $m.folder
   $m add cascade -label "Folder" -menu $m.folder -underline 0
   ##
   # Folder submenu
   #
   bindings::addCommand $m.folder "Create" {APL::createFolder $TA(highlightPath)}
   bindings::addCommand $m.folder "Rename" {APL::SelectedFolderRenameInPlace} 0
   bindings::addCommand $m.folder "Remove" {::catalog::removeFolder $::TA(highlightPath)}
   $m.folder add separator
   bindings::addCommand $m.folder "Expand Children"  {APL::expandChildren 1} 0
   bindings::addCommand $m.folder "Collapse Children" {APL::expandChildren 0}
   $m.folder add separator
   bindings::addCommand $m.folder "Move Clipboard Files Here" {APL::PasteFolderToTree}
   bindings::addCommand $m.folder "Copy Folder to Clipboard" \
      {APL::popWatch;clipboard append "[list $TA(highlightPath)] "}
   $m.folder add separator
   bindings::addCommand $m.folder "Create Stream Shortcut Here" {
      if {[playurl::checkuskfolder]} {
         playurl::addPromptUrl $TA(highlightPath)
         aplLB::UpdateList $::TA(highlightPath) 1
      }
   }
   ##
   # Selected submenu
   #
   menu $m.selected
   $m add cascade -label "Viewed Tracks" -menu $m.selected -underline 0
   bindings::addCommand $m.selected "Correct File Names" {fixplus::correctCase [aplLB::getAllNames 1]}
   ##
   # Media Players
   #
   if {[llength $::snackAmpSettings(mp3Player)]} {
      $m.selected add separator
      $m.selected add cascade -label "Copy selected to Media Player" -menu [menu $m.selected.medplay] -underline 0
      foreach {player} $::snackAmpSettings(mp3Player) {
         bindings::addCommand $m.selected.medplay "->$player" [list macros::copyToMediaPlayer $player {aplLB::getAllNames 1}]
      }
   }

   ##
   # Menus for ratings
   #
   menu $m.rating -tearoff 1
   $m add cascade -label "Ratings" -menu $m.rating -underline 0
   bindings::addCommand $m.rating "+ Folder Rating" {::db::bumpFolderRating $::TA(highlightPath) 0 1}
   bindings::addCommand $m.rating "- Folder Rating" {::db::bumpFolderRating $::TA(highlightPath) 0 -1}
   $m.rating add separator
   bindings::addCommand $m.rating "+ Viewed Tracks Rating" {::db::bumpFolderRating $::TA(highlightPath) $snackAmpSettings(treeDepth) 1}
   bindings::addCommand $m.rating "- Viewed Tracks Rating" {::db::bumpFolderRating $::TA(highlightPath) $snackAmpSettings(treeDepth) -1}
   $m.rating add separator
   bindings::addCommand $m.rating "+ Tree Rating" {::db::bumpFolderRating $::TA(highlightPath) 50 1}
   bindings::addCommand $m.rating "- Tree Rating" {::db::bumpFolderRating $::TA(highlightPath) 50 -1}
   ##
   # Catalog submenu
   #
   menu $m.catalog
   $m add cascade -label "Media Library" -menu $m.catalog -underline 0
      bindings::addCommand $m.catalog "Index Folder and Children" {
         APL::indexFolderChildren $TA(highlightPath)
      } 0
      bindings::addCommand $m.catalog "Catalog Folder and Children" {
         db::catalog $TA(highlightPath) 1 100
         waitWindow::close
      } 8
      bindings::addCommand $m.catalog "Catalog Viewed Tracks" {
         APL::catalogFolder $TA(highlightPath)
      } 8

   ##
   # Playlist
   #
   menu $m.playlist -tearoff 1
   $m add cascade -label "Playlist" -menu $m.playlist -underline 0
   bindings::addCommand $m.playlist "Append Viewed Tracks" {APL::AddToPlayList $TA(highlightPath)} 0
   bindings::addCommand $m.playlist "Queue Viewed Tracks" {APL::InsertToPlayList $TA(highlightPath)} 0
   bindings::addCommand $m.playlist "Queue then Play Viewed Tracks" {aplLB::playSelectedFiles}
   ##
   # AutoPlaylist
   #
   menu $m.apl
   $m add cascade -label "Auto-Playlist" -menu $m.apl -underline 0
   bindings::addCommand $m.apl "Set This and Children"  {APL::setSomeFolders $TA(highlightPath) 1} 0
   bindings::addCommand $m.apl "Clear This and Children" {APL::setSomeFolders $TA(highlightPath) 0} 1
   $m.apl add separator
   bindings::addCommand $m.apl "Apply" {
      macros::filterMusic replace $::TA(matchAgainstAll)
   }
   bindings::addCommand $m.apl "Save" {
      macros::saveCurrentAPL 0
   }
   ##
   # ID3
   #
   menu $m.id3
   $m add cascade -label "ID3" -menu $m.id3 -underline 0
#   bindings::addCommand $m.id3 "ID3 Group Editor"  {::APL::EditId3}
   bindings::addCommand $m.id3 "ID3 Table Editor"  {::APL::treeNameToTag 0} 0
   bindings::addCommand $m.id3 "ID3 Tag To Name" {::APL::treeTagToName}
   ##
   # Auto-Level
   #
   menu $m.gain
   $m add cascade -label "Auto-Level" -menu $m.gain -underline 0
   # Highlighted Folder Tracks
   menu $m.gain.folder
   $m.gain add cascade -label "Selected Folder Tracks" -menu $m.gain.folder -underline 0
   bindings::addCommand $m.gain.folder "Compute Track Gain" {gainControl::computeSome [db::filesInFolder $::TA(highlightPath)]}
   bindings::addCommand $m.gain.folder "Compute Album Gain"   {gainControl::averageSome [db::filesInFolder $::TA(highlightPath)]}
   bindings::addCommand $m.gain.folder "Reset Gain" {gainControl::resetSome [db::filesInFolder $::TA(highlightPath)]}
   if {[string length $::snackAmpSettings(mp3gainLoc)]} {
      $m.gain.folder add separator
      bindings::addCommand $m.gain.folder "mp3gain Tracks" {mp3gain::mp3gain 0 "/s s /r /d 3.0"}
      bindings::addCommand $m.gain.folder "mp3gain Albums" {mp3gain::mp3gain 0 "/s s /a /d 3.0"}
   }

   # Viewed Tracks
   menu $m.gain.viewed
   $m.gain add cascade -label "Viewed Tracks" -menu $m.gain.viewed -underline 0
   bindings::addCommand $m.gain.viewed "Compute Track Gain" {gainControl::computeSome [aplLB::getAllNames 1]}
   bindings::addCommand $m.gain.viewed "Compute Album Gain" {gainControl::averageSome [aplLB::getAllNames 1]}
   bindings::addCommand $m.gain.viewed "Reset Gain" {gainControl::resetSome [aplLB::getAllNames 1]}
   $m.gain.viewed add separator
   bindings::addCommand $m.gain.viewed "Increase Gain by +0.1" {gainControl::adjustSome [aplLB::getAllNames 1] .1}
   bindings::addCommand $m.gain.viewed "Decrease Gain by -0.1" {gainControl::adjustSome [aplLB::getAllNames 1] -.1}

   # Highlighted Folder and subfolders Tracks
   menu $m.gain.sub
   $m.gain add cascade -label "Selected Folder and Subfolders Tracks" -menu $m.gain.sub -underline 0
   bindings::addCommand $m.gain.sub "Compute Track Gain" {gainControl::computeSome [db::filesMatchingPattern [file join $::TA(highlightPath) *]]}
   bindings::addCommand $m.gain.sub "Compute Album Gain"   {gainControl::averageAlbums [aplFolder::foldersMatchingPattern [file join $::TA(highlightPath) *] 1]}
   bindings::addCommand $m.gain.sub "Reset Gain" {gainControl::resetSome [db::filesMatchingPattern [file join $::TA(highlightPath) *]]}
   if {[string length $::snackAmpSettings(mp3gainLoc)]} {
   $m.gain.sub add separator
      bindings::addCommand $m.gain.sub "mp3gain Tracks" {mp3gain::mp3gain 1 "/s s /r /d 3.0"}
      bindings::addCommand $m.gain.sub "mp3gain Album" {mp3gain::mp3gain 1 "/s s /a /d 3.0"}
   }

   ##
   # Misc
   #
   $m add separator
   bindings::addCommand $m "Track Info..." {
      APL::treeProperties $::TA(highlightPath)
   }

}
#------------------------------------------------------------------------------
# Function: APL::cut
#
# Copy the files contained in the selected path to the clipboard
# for later moving.
#------------------------------------------------------------------------------
proc APL::cut {args} {
   set here [focus]
   if {$here == [winfo parent $::TREE]} {
      APL::popWatch
      clipboard append "[list $::TA(highlightPath)] "
   } elseif {[winfo parent $here] == $::PLB} {
      APL::popWatch
      clipboardCopy $::PLB 1
   } else {
      tk_messageBox -type ok -icon warning \
         -message "You must have a selection in either the tree of file list area"
   }
}
#------------------------------------------------------------------------------
# Function   : APL::treeProperties
# Description: Show the stats of the tree under the folder
# Author     : Tom Wilkason
# Date       : 6/22/2002
#------------------------------------------------------------------------------
proc APL::treeProperties {selectedFolder} {
   ##
   # Total up stats for all files within a tree
   #
   set txt [db::statsText "" $selectedFolder]
   tk_messageBox -type ok -icon info -title "Properties for $selectedFolder" \
      -message "Information for '$selectedFolder'\n\n$txt"
   return
}
#------------------------------------------------------------------------------
# Function   : APL::PasteFolderToTree
# Description: Move the files in the clipboard to the selected folder
# Author     : Tom Wilkason
# Date       : 6/18/2002
#------------------------------------------------------------------------------
proc APL::PasteFolderToTree {{delete 1}} {
   APL::popWatch 100
   set folderName $::TA(highlightPath)
   if {$delete} {
      if {[catalog::isMediaPlayer $folderName]} {
         if {[catalog::copyClipboardFiles $folderName] > 0} {
            aplLB::UpdateList $folderName [winfo ismapped $::PLB]
            APL::treeSelect $folderName
         }
      } else {
         if {[catalog::moveClipboardFiles $folderName] > 0} {
            aplLB::UpdateList $folderName [winfo ismapped $::PLB]
            APL::treeSelect $folderName
         }
      }
   }
}
#------------------------------------------------------------------------------
# Function   : APL::createFolder
# Description: Prompt then create a subfolder under some level
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc APL::createFolder {folder} {
   set newName "                               "
   set res 0

   ##
   # TFW: ToDo: Need to create a new folder called "New Folder"
   # highlight it
   # bring up the rename dialog on it  SelectedFolderRenameInPlace
   # If it didn't change then remove it

   if {[getEntryDialog "Enter Folder Name" newName "Create Folder"]=="ok"} {
      if {[catch {
         set newName [file join $folder [string trim $newName]]
         file mkdir $newName
         aplFolder::Set $newName [aplFolder::Get [file dirname $newName]]
         newItem $newName
         APL::treeOpen [file dirname $newName]
         APL::treeSelect [file dirname $newName]
         update idletasks
         APL::viewHighlighted $::TREE
         set res 1
      } result] \
      } then {
         tk_messageBox -type ok -icon warning -message "Could not create folder $newName\n$result"
         set res 0
      }
   }
   return $res
}
#------------------------------------------------------------------------------
# Function   : APL::chooseIcon
# Description: Select the appropropiate icon based on file name
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc APL::chooseIcon {Name} {
   variable icons
   variable rmName
   if {[catalog::isFixedFolder $Name]} {
      set icon [lindex $icons 0]
   } elseif {[catalog::isFolderMediaPlayer $Name]} {
      set icon [lindex $icons 2]
   } elseif {[catalog::isARemMedia $Name]} {
      set icon [lindex $icons 1]
   } else {
      set icon [lindex $icons 0]
   }
   return $icon
}
#------------------------------------------------------------------------------
# Function   : APL::treeSelect
# Description: macros::Highlight a node
# Author     : Tom Wilkason
# Date       : 3/4/2002
#------------------------------------------------------------------------------
proc APL::treeSelect {folder} {
   variable folderList
   Tree::setselection $::TREE [APL::pathToTreeName $folder]
}
#------------------------------------------------------------------------------
# Function   : APL::treeOpen
# Description: Open a node
# Author     : Tom Wilkason
# Date       : 3/4/2002
#------------------------------------------------------------------------------
proc APL::treeOpen {folder {recurse 0}} {
   Tree::open $::TREE [APL::pathToTreeName $folder] $recurse
}
#------------------------------------------------------------------------------
# Return whether a node is open or not
#------------------------------------------------------------------------------
proc APL::isOpen {folder} {
   return [Tree::isOpen $::TREE [APL::pathToTreeName $folder]]
}
#------------------------------------------------------------------------------
# Function   : APL::busyCursor
# Description: Turn on/off a busy cursor
# Author     : Tom Wilkason
# Date       : 3/9/2002
#------------------------------------------------------------------------------
proc APL::busyCursor {busy} {
   if {$busy} {
      .wFC config -cursor watch
      $::TREE config -cursor watch
   } else {
      .wFC config -cursor $::TA(arrow)
      $::TREE config -cursor $::TA(arrow)
   }

}
#------------------------------------------------------------------------------
# Expand or contract the children of the selected node, typically bound
# to a key
#------------------------------------------------------------------------------
proc APL::expandChildren {how {levels 10}} {
   if {$how} {
      Tree::close $::TREE [APL::pathToTreeName $::TA(highlightPath)] $levels
      Tree::open $::TREE [APL::pathToTreeName $::TA(highlightPath)] $levels
   } else {
      Tree::close $::TREE [APL::pathToTreeName $::TA(highlightPath)] $levels
   }
}
#------------------------------------------------------------------------------
# Function   : APL::indexFolderChildren
# Description: Catalog a folder tree, called after it has been modified.
#              Assign to hotkeys
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc APL::indexFolderChildren {folder} {
   set opened [APL::isOpen $folder]
   catalog::Index $folder
   # Need to reselect the folder, make a dummy file so the folder is found in APL::SelectFile
   # Visible jump here due to update before this
   if {$opened} {
      APL::treeOpen $folder
   }
   APL::SelectFile [file join $folder xx]  ;# need to get file name selected in listbox, not the folder name
}
# Needed for hotkey compatability
proc APL::catalogFolderChildren {folder} {
   APL::indexFolderChildren $folder
}
#------------------------------------------------------------------------------
# Function   : APL::catalogFolder
# Description: Catalog a single folder, called after it has been modified.
#              Assign to hotkeys
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc APL::catalogFolder {folder} {
   waitWindow::show "Cataloging\n$folder"
   Wait 1
   # Nice to refresh the levels that are showing
   db::cacheFolderDetails $folder $::snackAmpSettings(treeDepth)
   aplLB::UpdateList $folder [winfo ismapped $::PLB]
   waitWindow::close
   Wait 0
}
#------------------------------------------------------------------------------
# Function   : APL::GetTreeFolder
# Description: Return the folder name corresponding to the label
#              Throws an error if can find corresponding folder name
# Author     : Tom Wilkason
# Date       : 12/27/2001
#------------------------------------------------------------------------------
proc APL::GetTreeFolder {W lbl} {
   upvar #0 $W myTree
   regexp -- {^.+?\((.+?)\)$} $myTree($lbl:variable) x Var
   return $Var
}
#----------------------------------------------------------------------------
# Check the media folders and rebuild tree if one as added or removed.
#----------------------------------------------------------------------------
proc APL::rebuildIfNewMedia {args} {
   global TA
   set oldlen [llength $TA(mediaDirs)]
   catalog::validateMediaPaths
   if {$oldlen!=[llength $TA(mediaDirs)]} {
      APL::reBuildTree
      tk_messageBox -type ok -icon info -title "Media Changed" \
         -message "Your available media folders are:\n[join $TA(mediaDirs) \n]"
   }
}
#----------------------------------------------------------------------------
# Rebuild the entire tree from scratch
#----------------------------------------------------------------------------
proc APL::reBuildTree {args} {
   APL::buildAPL $::TREE
   APL::treeBind
   Tree::open $::TREE / 1
}
#------------------------------------------------------------------------------
# Function   : APL::buildAPL
# Description: Rebuild the Auto Play List tree with new data.
# Author     : Tom Wilkason
# Date       : 11/4/2001
#------------------------------------------------------------------------------
proc APL::buildAPL {tree} {
   variable rmName
   global TA snackAmpSettings
   set base [join [lrange [split $tree .] 0 end-1] .]
   destroy $tree
   set ::Tree::font treeFont
   if {$snackAmpSettings(useAPLColors)} {
      set Tree::fg $snackAmpSettings(fg)
      set Tree::bg $snackAmpSettings(bg)
      settings::regColor $tree \
         -background ::snackAmpSettings(bg) \
         -foreground ::snackAmpSettings(fg) \
         -highlight  ::snackAmpSettings(highlight)
   } else {
      set Tree::fg black
      set Tree::bg white
   }
   set Tree::highlight $snackAmpSettings(highlight)
   # To work
   # 1. Text Tree selection outline box color can't be black (saTree ln 544)
   # 2. Text on tablelist for highlighted item can't be white
   Tree::create $tree -bg $Tree::bg -width 300 -height 400 -yscrollcommand [list $base.sb set]
   Tree::configure $tree -rippleSet $::snackAmpSettings(rippleSet)
   Tree::configure $tree -treeLine $::snackAmpSettings(treeLine)
   set i 0
   ##
   # Build up list of media paths including the RMC removable virtual path
   #
   set TA(mediaFolders) [list]
   foreach {Path} [concat $TA(mediaDirs) $::RMT] {
      # Folders won't have a / on end, so we don't want our pattern match to either
      lappend TA(mediaFolders) $Path [string trimright $Path /]
      # Title for tree just has the leading slash, nothing else
      set TA(Title,$Path) "/[string trim [string map {: {} / { }} $Path]]"
   }
   # Replace /RMC with A little better name
   set TA(Title,$::RMT) $rmName
   ##
   # Map each real folder name into one based from each media folder
   # to show it in the tree. Each media folder is considered one
   # level regardless of how may real directory levels it is on the disk.
   #
   # F:/Toms/Stuff     -> /F Toms/Stuff
   # //Toms/Toms/Stuff -> /Toms Toms/Stuff
   # Set the two types of icons, then need to have .open/.close types for each
   if {[catch {
      foreach folder [aplFolder::foldersMatchingPattern * 1] {
         newItem $folder
      }
   } result]} {
      debug $result
   }
   Tree::setDirty $::TREE 0
   pack $tree -side left -fill both -expand 1 -padx 0 -pady 0
}
#----------------------------------------------------------------------------
# For tkdnd, during a drop outside here just return the name of the
# highlighted folder
#----------------------------------------------------------------------------
proc APL::dndSrcTreeDrag {args} {
   Trace
   if {$::tcl_platform(platform) == "unix"} {
      return [file nativename $::TA(highlightPath)]
   } else {
      return [list [file nativename $::TA(highlightPath)]]
   }
}
#----------------------------------------------------------------------------
# Dragging from the checked label
#----------------------------------------------------------------------------
proc APL::dndSrcCheckedTreeDrag {args} {
   Trace
   if {[aplFolder::isSet $::TA(highlightPath)]} {
      lappend folders [file nativename $::TA(highlightPath)]
   } else {
      set folders [list]
   }
   foreach folder [aplFolder::foldersMatchingPattern [file join $::TA(highlightPath) *]] {
      if {[aplFolder::isSet $folder]} {
         lappend folders [file nativename $folder]
      }
   }
   if {[llength $folders]>0} {
      return [dndFileList $folders]
   } else {
      return -code break
   }
   #############
   set tracks [list]
   foreach {track} [db::filesMatchingPattern [file join $::TA(highlightPath) *]] {
      if {[aplFolder::isSet [file dirname $track]]} {
         lappend tracks $track
      }
   }
   if {[llength $tracks]>0} {
      return $tracks
   } else {
      return -code break
   }
}
#----------------------------------------------------------------------------
# Dragging from the total label
#----------------------------------------------------------------------------
proc APL::dndSrcTotalTreeDrag {args} {
   Trace
   lappend folders [file nativename $::TA(highlightPath)]
   foreach folder [aplFolder::foldersMatchingPattern [file join $::TA(highlightPath) *]] {
      lappend folders [file nativename $folder]
   }
   if {[llength $folders]>0} {
      return [dndFileList $folders]
   } else {
      return -code break
   }
   #############


   set tracks [list]
   foreach {track} [db::filesMatchingPattern [file join $::TA(highlightPath) *]] {
      lappend tracks [file nativename $track]
   }
   if {[llength $tracks]>0} {
      return $tracks
   } else {
      return -code break
   }
}
#----------------------------------------------------------------------------
# After a drag operation, grab the selected files from the apl listbox and
# return them.
#----------------------------------------------------------------------------
proc APL::dndSrcLBDrag {src} {
   Trace
   set useFiles [list]
   set tryFiles [getSelectedTracks $::PLB]
   foreach {file} $tryFiles {
      if {[file exists $file]} {
         lappend useFiles [file nativename $file]
      }
   }
   return [dndFileList $useFiles]
}
#------------------------------------------------------------------------------
# Function   : APL::delItem
# Description: Remove a new node from the tree
# Author     : Tom Wilkason
# Date       : 5/8/2002
#------------------------------------------------------------------------------
proc APL::delItem {folder} {
   Tree::delitem $::TREE [APL::pathToTreeName $folder]
}
#------------------------------------------------------------------------------
# Function   : APL::newItem
# Description: Add a new node to the tree
# Author     : Tom Wilkason
# Date       : 5/8/2002
#------------------------------------------------------------------------------
proc APL::newItem {folder} {
   set name [APL::pathToTreeName $folder]
   if {[string length $name] > 0} {
      # TFW: would be nice to find if it already exists
      if {[catch {
         Tree::newitem $::TREE $name \
            -variable ::teFolders($folder) \
            -image [APL::chooseIcon $folder]
      } result] } then {
         saLog $::errorInfo
      }
   } else {
      #bug: teFolders will contain unmounted drives
      #saLog "No translation for $folder"
   }
}
#----------------------------------------------------------------------------
# Add a new item and set it up all the way to its root
#----------------------------------------------------------------------------
proc APL::addToRoot {folder how} {
   global teFolders
   set hits 0
   set nodes [list]
   while {![info exists teFolders($folder)]} {
      incr hits
      set teFolders($folder) $how
      lappend nodes $folder
      # Backup one
      set root [file dirname $folder]
      if {[string equal $folder $root]} {
         break
      } else {
         set folder $root
      }
   }
   # Add nodes in the proper order
   foreach {node} [lsort -dictionary $nodes] {
      APL::newItem $node
   }
   return $hits
}

#------------------------------------------------------------------------------
# Function   : APL::EditId3, NameToTag, TagToName
# Description: Operate on ID3 tags for multiple files
# Author     : Tom Wilkason
# Date       : 10/30/2001
#------------------------------------------------------------------------------
proc APL::EditId3 {args} {
   package require id3Editor
   Wait 1
   id3Editor::id3Editor [aplLB::nameSelected]
   Wait 0
}
proc APL::NameToTag {smart} {
   package require id3Editor
   Wait 1
   id3Editor::id3NameToTag [aplLB::nameSelected] $smart
   Wait 0
}
proc APL::treeNameToTag {smart} {
   package require id3Editor
   # Limit to 200 files
   Wait 1
   set files [trackList::getFiles [aplFolder::getFolders $::TA(highlightPath) $::snackAmpSettings(treeDepth)]]
   if {[llength $files] > $::snackAmpSettings(ID3editLimit)} {
      tk_messageBox -type ok -icon warning \
         -message "Limiting operation to first $::snackAmpSettings(ID3editLimit) files"
   }
   id3Editor::id3NameToTag [lrange $files 0 $::snackAmpSettings(ID3editLimit)]
   Wait 0
}
proc APL::treeEditId3 {args} {
   package require id3Editor
   # Limit to 200 files
   Wait 1
   set files [trackList::getFiles [aplFolder::getFolders $::TA(highlightPath) $::snackAmpSettings(treeDepth)]]
   if {[llength $files]>$::snackAmpSettings(ID3editLimit)} {
      tk_messageBox -type ok -icon warning \
         -message "Limiting operation to first $::snackAmpSettings(ID3editLimit) files"
   }
   id3Editor::id3Editor [lrange $files 0 $::snackAmpSettings(ID3editLimit)]
   Wait 0
}
proc APL::TagToName {args} {
   package require id3Editor
   Wait 1
   id3Editor::id3TagToName [aplLB::nameSelected]
   Wait 0
}
proc APL::treeTagToName {args} {
   package require id3Editor
   # Limit to 200 files
   Wait 1
   set files [trackList::getFiles [aplFolder::getFolders $::TA(highlightPath) $::snackAmpSettings(treeDepth)]]
   if {[llength $files]>$::snackAmpSettings(ID3editLimit)} {
      tk_messageBox -type ok -icon warning \
         -message "Limiting operation to first $::snackAmpSettings(ID3editLimit) files"
   }
   id3Editor::id3TagToName [lrange $files 0 $::snackAmpSettings(ID3editLimit)]
   Wait 0
}
proc APL::removeTag {args} {
   id3Tag::removeTag [aplLB::nameSelected]
}
#------------------------------------------------------------------------------
# Rename files by splitting name and moving into subfolders
# so each folder contains one album.
# Note: Only handles Artist - Album - xxx formatted files
#        Adding Album - xxx format
# No error checking for shorter formats.
#------------------------------------------------------------------------------
proc APL::SplitFile {args} {
   set dirName ""
   set Artist ""
   set Album ""
   set sep $::snackAmpSettings(splitChar)
   set files [aplLB::nameSelected]
   foreach file $files {
      if {[string length $dirName] == 0} {
         set dirName [file dirname $file]
      }
      set parts [split [file tail $file] $sep]
      # Skip short names
      if {[llength $parts] < 2} {
         # This doesn't skip everything
         continue
      }
      # Assume Artist - Album format
      foreach {Artist Album} $parts {
         set Artist [string trim $Artist]
         set Album [string trim $Album]
         set F(Artist,$Artist) 1
         set F(Album,$Album) 1
         break
      }
   }
   # Param 1 check for uniqueness
   if {[llength [array names F Artist,*]] > 1} {
      tk_messageBox -message  "Artist Names must be unique"
      return
   }
   ##
   # Create Artist folder and create from there
   #
   if {![string equal -nocase $Artist [file tail $dirName]]} {
      # If Artist Folder doesnt already exist, need to create it
      set base [file join [file dirname $file] $Artist]
   } else {
      # Artist folder already exists, so use it
      set base [file dirname $file]
   }
   # Param 2 check for uniqueness
   # Assume format of Artist/Album - xxx
   if {[llength [array names F Album,*]] > 1} {
      set index 1
      set message  "Do you want to split [llength $files] files into folder '$base'?"
   } else {
      set index 2
      set message "Do you want to split [llength $files] files for '$Artist' into folder '[file join $base $Album]'?"
   }
   ##
   # Move the files
   #
   set result [tk_messageBox -type yesno -icon question -message $message]
   if {$result == "yes"} {
      foreach file $files {
         switch -- $index {
            1 {set newFile [file join $base [string trim [join [lrange [split $file $sep] $index end] $sep]]]}
            2 {set newFile [file join $base $Album [string trim [join [lrange [split $file $sep] $index end] $sep]]]}
            default {}
         }
         if {[catch {catalog::RenameFile $file $newFile} result]} {
            tk_messageBox -icon warning -message "Could not rename\n$file to \n$newFile\n$result"
         } else {
            catalog::RenameEntry $file $newFile
         }
      }
      # Catalog this tree to see the new folders
      APL::indexFolderChildren $::TA(highlightPath)
   }
}
#------------------------------------------------------------------------------
# Rename the files based on the folder names beneath them
#------------------------------------------------------------------------------
proc APL::CombineFile {level} {
   set sep $::snackAmpSettings(splitChar)
   set files [aplLB::nameSelected]
   foreach file $files {
      set folder [file dirname $file]
      set tail [file tail $file]
      set newTail [list]
      for {set i $level} {$i > 0} {incr i -1} {
         set newTail [linsert $newTail 0 [file tail $folder]]
         set folder [file dirname $folder]
      }
      set newTail [join $newTail " $sep "]
      set tail "$newTail $sep $tail"
      #set newFile [file join $folder $tail]
      set newFile [file join [file dirname $file] $tail] ;# leave in same folder
      if {[catch {catalog::RenameFile $file $newFile} result]} {
         tk_messageBox -icon warning -message "Could not rename\n$file to \n$newFile\n$result"
      } else {
         catalog::RenameEntry $file $newFile
      }
   }
   # Catalog this tree to see the new folders
   APL::indexFolderChildren $::TA(highlightPath)
}
#------------------------------------------------------------------------------
# Function   : APL::DeleteFiles
# Description: Remove some files from the auto playlist list box
# Author     : Tom Wilkason
# Date       : 10/30/2001
#------------------------------------------------------------------------------
proc APL::DeleteFiles {args} {
   set toDelete [aplLB::nameSelected 0]
   if {[llength $toDelete] > 0} {
      if {$::snackAmpSettings(dontMoveFiles)} {
         tk_messageBox -icon warning -type ok \
            -message "You don't have permission to remove files"
         return
      }
      set result [tk_messageBox -type yesno -title "Please Answer" \
         -icon question -message "Are you sure you want to delete:\n[join $toDelete \n]"]
      if {$result == "yes"} {
         foreach {File} $toDelete {
            catalog::DeleteFile $File
         }
         # Schedule a rescan when all is said and done
         # to reflect the new configuration
         trackList::populateHighighted 1 1
         gui::updateFileDisplay
      }
   }
}
#------------------------------------------------------------------------------
# Function   : APL::AddToPlayList
# Description: Add the entire set of songs in this folder to the playlist
# Author     : Tom Wilkason
# Date       : 10/28/2001
#------------------------------------------------------------------------------
proc APL::AddToPlayList {args} {
   ##
   # Now list the files out
   #
   set Cursor [LB::idxEnd]
   LB::AppendList [aplLB::getAllNames 1]
   LB::activate $Cursor
   gui::updateFileDisplay
}
#------------------------------------------------------------------------------
# Function   : APL::InsertToPlayList
# Description: Insert the entire set of songs in this folder to the playlist
# Author     : Tom Wilkason
# Date       : 10/28/2001
#------------------------------------------------------------------------------
proc APL::InsertToPlayList {args} {
   LB::userInsertList [aplLB::getAllNames 1]
}
#------------------------------------------------------------------------------
# Function   : APL::SelectFile
# Description: Given some file, select its folder in the tree if possible
# Author     : Tom Wilkason
# Date       : 6/6/2001
#------------------------------------------------------------------------------
proc APL::SelectFile {Path {pushPath 1}} {
   Trace
   variable folderList
   set W $::TREE
   if {[winfo exists $W] && ([wm state .wFC] == "normal")} {
      if {[aplFolder::Exists $Path 1]} {
         set file {}
         set Dir $Path
      } else {
         set file $Path
         set Dir [file dirname $Path]
      }
      # Save the path in the navagiation history
      if {$pushPath} {
         insertUnique $Dir
      }
      set showDir $Dir
      set ::TA(highlightPath) $Dir
      set Item [APL::pathToTreeName $Dir]
      set Dir [file dirname $Dir]
      # Expand containing folder and all parents too
      while {[file dirname $Dir] != $Dir} {
         Tree::open $W [APL::pathToTreeName $Dir]
         set Dir [file dirname $Dir]
      }
      # Extract the index number from the Tree variable
      Tree::setselection $W $Item
      upvar #0 $W myTree
      if {[catch {lindex [split $myTree($Item:variable) "()"] 1} index]} {
         return
      }
      ##
      # Update file listbox if it exists.
      # Note, must be drawn
      if {[winfo ismapped $::PLB]} {
         APL::setFilter 0
         aplLB::UpdateList $showDir
         ##
         # macros::Highlight the item in the listbox
         #
         if {[string length $file] > 0} {
            set contents [aplLB::getAllNames]
            set index [lsearch -exact $contents [file tail $Path]]
            aplLB::setSelection $index
         }
      }
      ##
      # Scroll the canvas so the highlighted item is visible
      # (if not all of data shown)
      APL::viewHighlighted $W
      # Give tree the focus to allow pasting to work
      focus [winfo parent $W]
   }
}
#------------------------------------------------------------------------------
# Function   : APL::viewHighlighted
# Description: Scroll the canvas so the highlighted item is visible
# Author     : Tom Wilkason
# Date       : 3/5/2002
#------------------------------------------------------------------------------
proc APL::viewHighlighted {W} {
   upvar #0 $W myTree
   update idletasks
   set OnScreenHeight [winfo height $W]
   set EntireHeight  [lindex [$W cget -scrollregion] 3]
   if {$EntireHeight > $OnScreenHeight} {
      set CursorY [lindex [$W bbox $myTree(selidx)] 3]
      if {[string length $CursorY] > 0} {
         set MoveTo [expr {double($CursorY-$OnScreenHeight+50)/$EntireHeight}]
         $W yview moveto $MoveTo
      }
   }
}

#------------------------------------------------------------------------------
# Function   : APL::pathToTreeName
# Description: Convert a File Path into a Tree Name
# Author     : Tom Wilkason
# Date       : 6/6/2001
# TODO: Can this be sped up?
#------------------------------------------------------------------------------
proc APL::pathToTreeName {folder} {
   global TA
   foreach Path $TA(mediaDirs) {
      # folder names won't end in a / (except drive vols) so our match pattern can't either
      # The titles are formed to not have path separators within them
      # Note: () or [] in media names need to be escaped for the regsub to work
      if {[string match ${Path}* $folder]} {
         set plen [string length $Path]
         set name [file join $TA(Title,$Path) [string trimleft [string range $folder $plen end] /]]
         return $name
      }
   }
   # Didn't find one mounted
   return {}
}
#------------------------------------------------------------------------------
# Function   : APL::setSomeFolders
# Description: Set/clear folders after base to some value
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc APL::setSomeFolders {base value} {
   aplFolder::SetSome "$base*" $value
   APL::refreshTree
}
#------------------------------------------------------------------------------
# Function   : APL::refreshTree
# Description: Refresh the tree such that any variables updated are relfected
#              in the checkboxes
# Author     : Tom Wilkason
# Date       : 1/11/2001
#------------------------------------------------------------------------------
proc APL::refreshTree {args} {
   global snackAmpSettings TA
   set W $::TREE
   # Save scroll % and restore it when done
   set scroll [lindex [$W yview] 0]
   foreach {Path} $TA(mediaDirs) {
      Tree::close $W $TA(Title,$Path)
      Tree::open $W $TA(Title,$Path)
   }
   $W yview moveto $scroll
}
#------------------------------------------------------------------------------
# Function   : APL::setFileListToShow
# Description: Show or hide the filelist
# Author     : Tom Wilkason
# Date       : 3/9/2001
#------------------------------------------------------------------------------
proc APL::setFileListToShow {what widget} {
   variable paneOrient
   if {$::snackAmpSettings(hideFileList)} {
      pane $what -orient $paneOrient -dynamic 0 \
         -size $::snackAmpSettings(FileListSize)
      $widget configure -relief sunken
      set ::snackAmpSettings(hideFileList) 0
   } else {
      $widget configure -relief flat   ;# button
      set ::snackAmpSettings(hideFileList) 1
      set curSize [pane $what -size]
      if {$curSize != {}} {
         set ::snackAmpSettings(FileListSize) $curSize
         pane forget [winfo parent $what] $what
      }
   }
}
#------------------------------------------------------------------------------
# Function   : aplLB Namespace
# Description: Create bindings for the APL tree
# Author     : Tom Wilkason
# Date       : 3/9/2001
#------------------------------------------------------------------------------
namespace eval aplLB {
   proc commonBindings {args} {
      # Toplevel bindings
      bind .wFC <F5>       {SelectFile 0}
      bind .wFC <Control-F5> {SelectFile 1}
      bind .wFC <Alt-F4> {APL::close}
      # Track List
      APL_ListAssign <F2> {::APL::RenameFile}
      APL_ListAssign <Control-F2> ::APL::RenameFile
      APL_ListAssign <F3> ::APL::EditId3
      APL_ListAssign <F4> {::APL::NameToTag 0}
#      APL_TreeAssign <F4> {::APL::treeNameToTag 0}
      APL_TreeAssign <F3> {::APL::treeEditId3}
      APL_ListAssign <F6> {::APL::TagToName}
      APL_TreeAssign <F6> {::APL::treeTagToName}
      # todo: replace with tkdnd
      APL_ListAssign <Delete> ::APL::DeleteFiles
      APL_ListAssign <B1-Motion> {
         # Prevent the select-drag action
         # It would mess up the DND
         $::PLB config -cursor hand2
         break
      }
      APL_ListAssign <ButtonPress-3> {
         tk_popup $::PLB.popup %X %Y
      }
      APL_ListAssign <ButtonPress-1> {
         # Allow key bindings to work
         focus %W
         # Process any special columns
         trackList::colClick %W %x %y %X %Y
      }
      APL_ListAssign <<Copy>> {APL::popWatch;clipboardCopy $::PLB 1;break}
      ##
      # DND leave, grap to the clipboard
      #
      # todo: replace with tkdnd
      APL_ListAssign <B1-Leave> {
         # Leaving the listbox, snapshot the files
         clipboardCopy $::PLB 0
         break   ;# Prevent the drag select
      }
      ##
      # Drop action coming from the APL list
      #
      APL_ListAssign <ButtonRelease-1> {
         # See if we are over a node on the tree
         set target [winfo containing %X %Y]
         if {[winfo exists $target]} {
            if {[string equal $target $::TREE]} {
               APL::cbDndPLBToTree %X %Y
            } elseif {[string equal $target [LB::body]]} {
               APL::cbDndPLBToLB %X %Y
            } elseif {[winfo toplevel $target] == [winfo toplevel $::LB]} {
               APL::cbDndPLBToLB 0 0
            }
         }
         $::PLB config -cursor $::TA(arrow)
         catch {unset target}
      }
   }
   proc appendSelectedFiles {args} {
      #if {[array size ::trackList::Index2File] > 0} {
         LB::AppendList [aplLB::nameSelected 0]
      #}
   }
   proc insertSelectedFiles {args} {
      if {[winfo parent $::TREE] == [focus]} {
         LB::userInsertList [aplLB::getAllNames 1]
      } else {
         #if {[array size ::trackList::Index2File] > 0} {
            LB::userInsertList [aplLB::nameSelected 0]
         #}
      }
   }
   ##
   # Adjust the user rating of the track(s)
   #
   proc bumpRating {path depth amount} {
      if {[winfo parent $::TREE] == [focus]} {
         # Folder
         db::bumpFolderRating $path $depth $amount
      } else {
         # Selected Files
         set selected [aplLB::nameSelected 0]
         db::bumpTrackRating $selected $amount
      }
   }
   ##
   # Play the highlighted files or folder
   #
   proc playSelectedFiles {args} {
      if {[winfo parent $::TREE] == [focus]} {
         # Folder
         set end [LB::userInsert]
         LB::userInsertList [aplLB::getAllNames 1]
         LB::activate $end
         Play active
      } else {
         # Selected Files
         set selected [aplLB::nameSelected 0]
         if {[llength $selected] == 1} {
            LB::InsertPlay [lindex $selected 0]
         } else {
            set end [LB::userInsert]
            if {[LB::userInsertList $selected] > 0} {
               LB::inverseVideoFile xx 0
               LB::activate $end
               Play active
            }
         }
      }
   }
   ##
   # Menus bound to listbox
   #
   proc menus {args} {
      ##
      # Rating popup
      #
      set m $::PLB.plbRating
      menu $m -postcommand [list trackList::ratingPopup $m]
      ##
      # File list box bindings
      #
      set plb [$::PLB bodypath]
      bindings::init $plb menuBindings
      set m $::PLB.popup
      menu $m
      ##
      # Copy the files to the playlist
      #
      bindings::addCommand $m "Play Track(s)" {aplLB::playSelectedFiles} 0
      bindings::addCommand $m "Queue Tracks(s) to Play Next" {aplLB::insertSelectedFiles} 0
      bindings::addCommand $m "Append Tracks(s) to Playlist" {aplLB::appendSelectedFiles}
      $m add separator
      bindings::addCommand $m "Increase Rating" {aplLB::bumpRating $::TA(highlightPath) 100 1}
      bindings::addCommand $m "Decreasing Rating" {aplLB::bumpRating $::TA(highlightPath) 100 -1}
      $m add separator
      ##
      # Rename the active selection
      #
      menu $m.edit
      $m add cascade -label "Edit" -menu $m.edit -underline 0
      bindings::addCommand $m.edit "Delete Tracks(s)"  {::APL::DeleteFiles} 0
      bindings::addCommand $m.edit "Cut Tracks(s) to clipboard" {APL::popWatch;clipboardCopy $::PLB 1} 0
      # clipboardPaste doesn't work with APL table
      bindings::addCommand $m.edit "Copy Clipboard Tracks Here" {APL::popWatch;APL::PasteFolderToTree} 0
      bindings::addCommand $m.edit "Move Clipboard Tracks Here" {APL::popWatch;APL::PasteFolderToTree} 0
#      bindings::addCommand $m.edit "Autolevel Tracks" {} 0

      menu $m.tn
      $m add cascade -label "Track Names" -menu $m.tn -underline 0
      bindings::addCommand $m.tn "Rename Track"  {::APL::RenameFile} 0
      bindings::addCommand $m.tn "Combine Name with Folder (1)" {::APL::CombineFile 1}
      bindings::addCommand $m.tn "Combine Name with Folders (2)" {::APL::CombineFile 2}
      bindings::addCommand $m.tn "Split Name Into Folders" {::APL::SplitFile}
      bindings::addCommand $m.tn "Correct Track Name(s)" {fixplus::correctCase [aplLB::nameSelected]}

      menu $m.plf
      $m add cascade -label "Playlist File" -menu $m.plf -underline 0
      bindings::addCommand $m.plf "Save Selected to File"  {macros::SavePlaylistFile $::PLB replace selected}
      bindings::addCommand $m.plf "Save All to File"  {macros::SavePlaylistFile $::PLB replace all}
      ##
      # Media Players
      #
      if {[llength $::snackAmpSettings(mp3Player)]} {
         $m add cascade -label "Copy selected to Media Player" -menu [menu $m.medplay] -underline 0
         foreach {player} $::snackAmpSettings(mp3Player) {
            bindings::addCommand $m.medplay "->$player" [list macros::copyToMediaPlayer $player {selectedTracks $::PLB}]
         }
      }
      $m.plf add separator
      bindings::addCommand $m.plf "Append Selected to File"  {macros::SavePlaylistFile $::PLB append selected}
      bindings::addCommand $m.plf "Append All to File"  {macros::SavePlaylistFile $::PLB append all}

      menu $m.id3
      $m add cascade -label "ID3 Tags" -menu $m.id3 -underline 0
      bindings::addCommand $m.id3 "Edit Track(s) Info"  {::APL::EditId3} 0
      bindings::addCommand $m.id3 "Table Editor"  {::APL::NameToTag 0}
      bindings::addCommand $m.id3 "Tag To Name Editor" {::APL::TagToName}

      menu $m.gain
      $m add cascade -label "Auto-Level" -menu $m.gain -underline 0
      bindings::addCommand $m.gain "Compute Track(s) Gain" {gainControl::computeSome [aplLB::nameSelected]}
      bindings::addCommand $m.gain "Average Track(s) Gain" {gainControl::averageSome [aplLB::nameSelected]}
      bindings::addCommand $m.gain "Reset Track(s) Gain" {gainControl::resetSome [aplLB::nameSelected]}
      $m add separator
      bindings::addCommand $m "Properties" {catalog::properties [aplLB::nameSelected]}

#      $m add separator
#      bindings::addCommand $m "Create Stream Shortcut Here" {
#         if {[playurl::checkuskfolder]} {
#            playurl::addPromptUrl $TA(highlightPath)
#            aplLB::UpdateList $::TA(highlightPath) 1
#         }
#      }
   }
   proc selected {args} {
      return [$::PLB curselection]
   }
   proc setSelection {index} {
      $::PLB selection clear 0 end
      $::PLB selection set $index
      $::PLB activate $index
      $::PLB see $index
   }
   ##
   # Table Stuff
   #
   set APL::paneOrient "horizontal"
   package require Tablelist
   #----------------------------------------------------------------------------
   # Called from button, catalog and rescan the folder and appropriate subfolders.
   #----------------------------------------------------------------------------
   proc RefreshFolderTree {path} {
      APL::indexFolderChildren $path
      update
      UpdateList $path 1
   }
   ##
   # Update the listbox based on incoming folder
   # Generally called when a new folder is selected
   proc UpdateList {index {insert 1}} {
      Trace
      ##
      # New Table
      #
      if {$insert} {
         if {[catch {
            trackList::populate $index $::PLB 1 1
         } result] } then {
            saLog $::errorInfo
         }
      }
      ##
      # Update the selected counts
      #
      set checked 0
      set unchecked 0
      set pattern "[string trimright $index /]/*"
      # Show folders checked instead of files (faster)
      foreach {dirName} [aplFolder::foldersMatchingPattern $pattern] {
         if {[aplFolder::isSet $dirName 1]} {
            incr checked
         } else {
            incr unchecked
         }
      }
      if {($checked+$unchecked) >= 0} {
         set ::TA(ck,selected)   "Checked $checked"
         set ::TA(ck,unselected) "Unchecked $unchecked"
         set ::TA(ck,total)      "Total [expr {$checked+$unchecked}]"
         if {[string length [info command ntFreeSpace]]} {
            if {[catch {
               set fr [ntFreeSpace [file nativename [fileRoot $index]]]
               if {[llength $fr]==2} {
                  foreach {tot free} $fr {break}
                  set ::TA(ck,free)   "Free [fileSize $free]"
               }
            } result] } then {
               set ::TA(ck,free)   "Free unKnown"
               saLog $result
            }
         }
      }
      # Signal done
   }
   proc rowIndex2file {index} {
      return [trackList::rowIndex2file $index]
   }
   proc getAllNames {{long 0}} {
      set files [list]
      # Get file name for each row
      set end [$::PLB index end]
      for {set i 0} {$i < $end } {incr i} {
         if {$long} {
            lappend files [catalog::TranslateMount [trackList::rowIndex2file $i]]
         }  else {
            lappend files [file tail [trackList::rowIndex2file $i]]
         }
      }
      return $files
   }
   proc getActiveFullName {args} {
      set active  [$::PLB index active]
      return [catalog::TranslateMount [trackList::rowIndex2file $active]]
   }
   proc nameSelected {{translate 1}} {
      set fullNames [list]
      if {$translate} {
         foreach {index} [selected] {
            lappend fullNames [catalog::TranslateMount [trackList::rowIndex2file $index]]
         }
      } else {
         foreach {index} [selected] {
            lappend fullNames [trackList::rowIndex2file $index]
         }
      }
      return $fullNames
   }
   proc RemoveFile {File} {
      if {[winfo exists $::PLB]} {
         set lnFile [file tail $File]
         set index [lsearch -exact [getAllNames] $lnFile]
         if {$index > -1} {
            $::PLB delete $index
         }
      }
   }
   ##
   # Key bindings for table
   #
   proc bindings {args} {
      # Common Keys
      commonBindings
      # Couple hot keys

      APL_ListAssign <Double-ButtonPress-1> {
         #if {[array size ::trackList::Index2File] > 0} {
            if {[catch {trackList::row2file %W %y} File] } then {
               saLog $File
            } else {
               LB::InsertPlay $File
            }
         #}
      }
   }
}

#------------------------------------------------------------------------------
# Function   : RenameFile
# Description: Rename a file directly from the listbox
# Author     : Tom Wilkason
# Date       : 6/19/2002
#------------------------------------------------------------------------------
proc APL::RenameFile {args} {
   set W $::PLB
   if {[catch {aplLB::getActiveFullName} fullName]} {
      saLog "APL::RenameFile $fullName\n$::errorInfo"
      return
   }
   set oldName [file tail $fullName]
   set dirName [file dirname $fullName]
   set newName $oldName
   ##
   # Block on dialog
   #
   set name [catalog::renameInPlaceDialog $W 1 $oldName]
   ##
   # Back from window closure
   #
   if {[string length $::TA(rename)]>0 && $::TA(rename) != $oldName} {
      set newFullName [file join $dirName $::TA(rename)]
      if {[catch {catalog::RenameFile $fullName $newFullName} result]} {
         tk_messageBox -icon warning -message "Could not rename\n$fullName to \n$newFullName\n$result"
      } else {
         catalog::RenameEntry $fullName $newFullName
      }
   }
}
namespace eval APLcover {
   variable B .aplCoverart
}
#------------------------------------------------------------------------------
# Toggle the display of album art
#------------------------------------------------------------------------------
proc APLcover::ToggleArt {args} {
   global TA
   variable B
   if {![winfo exists $B]} {
      APLcover::coverArtViewer
      wm deiconify $B;update idletasks
      after idle [list APLcover::packCoverArt $::TA(highlightPath)]
      $TA(aplArtButton) configure -relief sunken
   } else {
      if {[wm state $B]=="withdrawn"} {
         wm deiconify $B
         after idle [list APLcover::packCoverArt $::TA(highlightPath)]
         $TA(aplArtButton) configure -relief sunken
      } else {
         wm withdraw $B
         $TA(aplArtButton) configure -relief flat
      }
   }
}
#------------------------------------------------------------------------------
# Create a view window for coverart
#------------------------------------------------------------------------------
proc APLcover::coverArtViewer {args} {
   variable B
   global snackAmpSettings
   if {![winfo exists $B]} {
      toplevel $B
      wm title $B "Folder Art"
      wm withdraw $B
      wm protocol $B WM_DELETE_WINDOW APLcover::ToggleArt
      bind $B <Double-ButtonPress-1> {
         APLcover::searchCoverArt
      }

      button $B.drop -text "Query Art" -relief flat -padx 0 -command {APLcover::searchCoverArt}
      balloon::button $B.drop "Press to bring up a dialog to query google for album art"
      image create photo aplCoverArtPhoto -data $gui::saLogo
      label $B.ca -image aplCoverArtPhoto -highlightthickness 0 -border 2 -relief sunken
      balloon::define $B.ca "Double click to query cover art for this album. Drag desired image from web browser to here to save it."
      pack $B.ca -expand true -fill both
      ontop::pinOnSystemMenu {} $B
      if {![catch {package require tkdnd} result] } then {
         dnd bindtarget $B text/uri-list <Drop> {APLcover::dndDropCoverArt %D}
         # bug: unix, toplevel or frame targets don't work
         dnd bindtarget $B.ca text/uri-list <Drop> {APLcover::dndDropCoverArt %D}
      }
      wm geometry $B {}
      after 100 [list myIcon $B]
   }
}
#----------------------------------------------------------------------------
# Search for cover art for the currently playing track
#----------------------------------------------------------------------------
proc APLcover::searchCoverArt {args} {
   set parts [file split $::TA(highlightPath)]
   set data(Album) [lindex $parts end]
   set data(Artist) [lindex $parts end-1]
   browser::coverart data
}

#------------------------------------------------------------------------------
# Insert the coverart for the currently playing track
# state is normal, iconic, or withdrawn
#------------------------------------------------------------------------------
proc APLcover::packCoverArt {folder args} {
   Trace
   variable B
   if {[winfo exists $B] && [winfo ismapped $B]} {
      # If processing the same folder, then use existing cover image
      set ca [coverart::covername $folder]
      if {[llength $ca]} {
         catch {image delete aplCoverArtPhoto}
         if {[catch {image create photo aplCoverArtPhoto -file $ca} result] } then {
            saLog $result
            pack $B.drop -expand true -fill x -before $B.ca
            image create photo aplCoverArtPhoto -data $gui::saLogo
         } else {
            gui::clampImageSize aplCoverArtPhoto \
               $::snackAmpSettings(minArtSize) $::snackAmpSettings(maxArtSize)  ;# clamp image size 100 min 500 max
            pack forget $B.drop
         }
      } else {
         pack $B.drop -expand true -fill x -before $B.ca
         image create photo aplCoverArtPhoto -data $gui::saLogo
      }
      wm title $B "$folder"
      wm geometry $B {}
   }
}
#------------------------------------------------------------------------------
#  Handler for dropping cover art on the album viewer
#------------------------------------------------------------------------------
proc APLcover::dndDropCoverArt {files} {
   global TA
   set flag 0
   if {[llength $files]} {
      # Hack, if first file is a graphic, just replace the
      # folder image
      set ext [string tolower [file extension [lindex $files 0]]]
      switch -- $ext {
         .jpg -
         .bmp -
         .png {
            copyImageToFolder [lindex $files 0] $::TA(highlightPath)
            APLcover::packCoverArt $::TA(highlightPath)
         }
      }
   }
   return 1
}
#------------------------------------------------------------------------------
# trackList namespace: Used for table widget
#------------------------------------------------------------------------------
namespace eval trackList {
   variable col2Use [list]
   variable lastFolder ""
   variable editCellButton
   variable editCellState 0
   variable editCellCheckState 0
   variable TABLEDATA [list]
   variable butrel
   array set butrel {0 sunken 1 flat}
}
#------------------------------------------------------------------------------
# Schedule a rescan when all is said and done
# to reflect the new configuration
#------------------------------------------------------------------------------
proc trackList::populateHighighted {sort insert {delay "idle"}} {
   afterCancel trackList::populate
   after $delay [list trackList::populate $::TA(highlightPath) $::PLB $sort $insert]
}
#------------------------------------------------------------------------------
# Return a file name based on row index
# Col 0 has a hidden column with original key values
# which are maintained with the table is sorted.
# use the value to lookup the correct file name
#------------------------------------------------------------------------------
proc trackList::rowIndex2file {index} {
   variable TABLEDATA
   return [db::fileByPosition [lindex $TABLEDATA [list $index 0]]]
   #return [db::fileByPosition [lindex [lindex $TABLEDATA $index] 0]]
}
#------------------------------------------------------------------------------
# Map selected row to full file name
#------------------------------------------------------------------------------
proc trackList::row2file {W y} {
   set File {}
   set y [expr {$y + [winfo y $W]}]
   set index [$::PLB nearest $y]
   return [rowIndex2file $index]
}
proc trackList::colClick {W x y X Y} {
   variable activeFile
   variable ROW
   variable COL
   trackList::pushClick $W $x $y
   set activeFile [row2file $W $y]
   set col [$::PLB columncget [$::PLB nearestcolumn [expr {$x + [winfo x $W]}]] -title]
   switch -- $col {
      "Rating" {
         # call the pop-up which uses the ratingPopup postcommand callback below
         tk_popup $::PLB.plbRating $X $Y
      }
      default {}
   }
}
##
# Pop-up menu to adjust track rating
# This is called from a post command
#
proc trackList::ratingPopup {menu} {
   variable activeFile
   variable trackRating
   variable TABLEDATA
   variable ROW
   variable COL
   $menu delete 0 end
   #  activeFile set during click
   set me $activeFile
   set rating [db::getValues $me Rating]
   set trackRating $rating
   foreach {n} $db::ratingOptions {
      $menu add radiobutton -label $n -value $n \
      -variable trackList::trackRating \
      -command "[list db::setVal $me Rating $n]\;[list trackList::setVal $trackList::ROW $trackList::COL $n]"
   }
}
#----------------------------------------------------------------------------
# trackList::pushClick
# Save the row/cell index where the last click was
#----------------------------------------------------------------------------
proc trackList::pushClick {W x y} {
   variable ROW
   variable COL
   # W is PLB.body
   foreach {ROW COL} [split [$::PLB nearestcell [expr {$x + [winfo x $W]}] [expr {$y + [winfo y $W]}]] ,] {break}
}
#----------------------------------------------------------------------------
#  trackList::setVal
#
#  Set some row/col value in the displayed data, faster than refreshing
#----------------------------------------------------------------------------
proc trackList::setVal {row col newVal} {
   variable TABLEDATA
   lset TABLEDATA $row $col $newVal
}
#------------------------------------------------------------------------------
# trackList::editCellStart
#
# When a cell is first edited, this callback fires
#
#------------------------------------------------------------------------------
proc trackList::editCellStart {W row col newVal} {
   Trace
   variable oCellValue
   set oCellValue $newVal
   return $newVal
}
#------------------------------------------------------------------------------
# trackList::editCellEnd
#
# When a cell is edited, this callback fires
#------------------------------------------------------------------------------
proc trackList::editCellEnd {W row col newVal} {
   Trace
   variable TABLEDATA
#   variable Index2File
   variable oCellValue
   #$W rejectinput ;# To do nothing leaving editing open
   #$W cancelediting ;# To do nothing leaving editing open
   #.wFC.sv.lb.tbl cancelediting
   set col [$W columncget $col -title]
   #set file $Index2File([lindex [lindex $TABLEDATA $row] 0])

   set file [trackList::rowIndex2file $row]
   switch -exact -- $col {
      Folder {
         # This can be dangerous, disable Folder from editing
         set newFullName [file join $newVal [file tail $file]]
         if {[catch {catalog::RenameFile $file $newFullName} result]} {
            tk_messageBox -icon warning -message "Could not rename\n$file to \n$newFullName\n$result"
            return $oCellValue
         } else {
            catalog::RenameEntry $file $newFullName
            # Also need to change the tracklist key data, renamng file changed its row number
            # in the database making the database different than the TABLEDATA row info.
            # find the database row of this new track and replace the TABLEDATA instance
            # lset TABLEDATA [lindex $row 0]
            trackList::setVal $row 0 [db::whereInDb $newFullName]
         }
      }
      File {
         set newFullName [file join [file dirname $file] $newVal]
         if {[catch {catalog::RenameFile $file $newFullName} result]} {
            tk_messageBox -icon warning -message "Could not rename\n$file to \n$newFullName\n$result"
            return $oCellValue
         } else {
            catalog::RenameEntry $file $newFullName
            # Also need to change the tracklist key data
            trackList::setVal $row 0 [db::whereInDb $newFullName]
         }
      }
      Title -
      Artist -
      Album -
      Track -
      Media -
      Year -
      Genre -
      Note -
      Genre -
      Comp -
      Desc -
      Comp {
         if {$::snackAmpSettings(ID3ReadOnly)==0} {
            if {[catch {
               #id3Tag::id3Label $file fnameData V1
               set fnameData($col) $newVal
               id3Tag::id3Modify $file fnameData
            } result] } then {
               debug $result
               return $oCellValue
            }
         } else {
            return $oCellValue
         }
      }
      Gain {
         # This needs a conversion first
         gainControl::add $file [lindex $newVal 0]
      }
      Rating {
         db::setVal $file Rating $newVal
      }
      Played {
         db::setVal $file Played $newVal
      }
      default {saLog "Unknown option '$col'"}
   }
   return $newVal
}

#------------------------------------------------------------------------------
# trackList::changeEdit
#
# Change the editing mode of a table
#------------------------------------------------------------------------------
proc trackList::changeEdit {args} {
   variable editCellButton
   variable editCellCheckState
   variable editCellState
   variable butrel
   catch {tablelist::canceleditingSubCmd $::PLB}
   ##
   # set or clear the editable columns
   #
   set i 0
   $editCellButton configure -relief $butrel($editCellState)
   foreach {xx key xx} [$::PLB cget -columns] {
      if {$editCellState} {
         $::PLB columnconfigure $i -editable 0
      } else {
         $::PLB columnconfigure $i -editable $::snackAmpSettings(col,$key,edit)
      }
      incr i
   }
   set editCellState [expr {!$editCellState}]
   set editCellCheckState $editCellState
}

#------------------------------------------------------------------------------
# trackList::buildTable
#
# Builds the widget w as a tablelist widget
# contained in a newly created top-level widget.  Returns the name of the
# tablelist widget.
#------------------------------------------------------------------------------
proc trackList::buildTable {top _config} {
   variable col2Use
   variable TABLEDATA
   upvar $_config config
   # List of possible columns
   set col2Use [list]
   set i 0
   set j 0

   set sortMode [list]
   # Build up col info
   # TFW: Need to do this in saved order
   foreach key $::TA(cols) {
      if {$config(col,$key,show) || $key=="Key"} {
         lappend col2Use $config(col,$key,width) $key $config(col,$key,align)
         lappend sortMode $j $config(col,$key,type)
         incr j
      }
   }
   #
   # Create a scrolled tablelist widget with 5 dynamic-width
   # columns and interactive sort capability within the top-level
   #
   set tf [frame $top]     ;# User should pack in frame
   set tbl $tf.tbl
   set vsb $tf.vsb
   set hsb $tf.hsb
   tablelist::tablelist $tbl \
         -columns $col2Use \
         -labelcommand tablelist::sortByColumn \
         -xscrollcommand [list scrollCmd $hsb  [list grid $hsb -row 1 -column 0 -sticky ew]] \
         -yscrollcommand [list scrollCmd $vsb  [list place $vsb -relx 1.0 -rely 0.0 -anchor ne -relheight 1.0]] \
         -foreground       $Tree::fg \
         -background       $Tree::bg   \
         -selectforeground $::snackAmpSettings(fg) \
         -selectbackground $::snackAmpSettings(highlight) \
         -showseparators   no \
         -borderwidth      1 \
         -selectmode       extended \
         -setgrid          no \
         -listvariable   trackList::TABLEDATA \
         -editstartcommand {trackList::editCellStart} \
         -editendcommand {trackList::editCellEnd} \
         -movablecolumns 1
   # $PLB cget -columns will return current cols in order (width tag alignment)
   # Define editiable columns
   ##
   # Set the options
   #
   settings::regColor $tbl \
      -foreground ::snackAmpSettings(fg) \
      -background ::snackAmpSettings(bg) \
      -selectforeground ::snackAmpSettings(fg) \
      -selectbackground ::snackAmpSettings(highlight)

   if {$::snackAmpSettings(useTableStripes)} {
      settings::regColor $tbl -stripebackground ::snackAmpSettings(tsbg)
      $tbl config -stripebackground $::snackAmpSettings(tsbg)
   }
   if {$::snackAmpSettings(useTableBars)} {
      $tbl config -showseparators yes
   }
   if {$::snackAmpSettings(autoTableStretch)} {
      $tbl config -stretch all
   }
   # Define how each col is sorted.
         # -yscrollcommand [list scrollCmd $vsb  [list grid $vsb -row 0 -column 1 -sticky ns]]
   if {[catch {
      foreach {col mode} $sortMode {
         switch -- $mode {
            "commify" {
               $tbl columnconfigure $col -sortmode integer -formatcommand commify
            }
            "timify" {
               $tbl columnconfigure $col -sortmode real -formatcommand timify
            }
            default {
               $tbl columnconfigure $col -sortmode $mode
            }
         }
      }
   } result]} {
      debug $result
   }
   # Hide column used for the key
   $tbl columnconfigure 0 -hide 1
   scrollbar $vsb -orient vertical   -command [list $tbl yview]
   scrollbar $hsb -orient horizontal -command [list $tbl xview]
   #
   # Manage the widgets
   #
   grid $tbl -row 0 -column 0 -sticky news
   #grid $vsb -row 0 -column 1 -sticky ns
   place $vsb -relx 1.0 -rely 0.0 -anchor ne -relheight 1.0
   grid $hsb -row 1 -column 0 -sticky ew
   grid rowconfigure    $tf 0 -weight 1
   grid columnconfigure $tf 0 -weight 1
   # Below has hard reference to table .body
   mscroll $tbl.body
   return $tbl
}
#------------------------------------------------------------------------------
# trackList::saveColumnWidths
#
# Save the current column widths for later reuse
#
#
#------------------------------------------------------------------------------
proc trackList::saveColumnWidths {} {
   variable lastFolder
   variable TABLEDATA
   global snackAmpSettings
   Trace
   # Save displayed col widths and order
   set colOrder [list]
   foreach {width name just} [$::PLB cget -columns] {
      lappend colOrder $name
      set snackAmpSettings(col,$name,width) $width
   }
   # TFW: Need to work on this
   set ::snackAmpSettings(colorder) $colOrder
   set ::snackAmpSettings(table,sortcol) [$::PLB sortcolumn]
   set ::snackAmpSettings(table,sortorder) [$::PLB sortorder]
}
#------------------------------------------------------------------------------
# trackList::populate
#
# Populate the table widget with track information
# if 'insert' is not true, the cache files are built
# but no attempt is made to populate the listbox
#------------------------------------------------------------------------------
proc trackList::populate {baseIndex tbl sort insert} {
#   variable Index2File
   variable lastFolder
   variable TABLEDATA
   set llen [llength $TABLEDATA]
   set col2Use [list]
   set isDirty 0
   #APL::busyCursor 1
   ##################update idletasks
   # --------------------------------------------------------------
   # Only use the columns that are currently used
   foreach {xx col xx} [$tbl cget -columns] {
      lappend col2Use $col
   }
   catch {tablelist::canceleditingSubCmd $::PLB}
   # Init some values
   set totalSize 0
   set totalTime 0
   set tableData [list]
   ##
   # Return a nested list of table data for this folder and treeDepth subfolders
   # Cache any data as necessary
   set filterHow $::snackAmpSettings(FilterWhat)
   if {$::TA(useSAFilter)} {
      set filter $::TA(SAFilter)
      set depth 100
      # Special case, match multiple fields
      if {$filterHow == "Any"} {
         set filterHow [list Title Artist Album File]
      }
   } else {
      set filter ""
      set depth $::snackAmpSettings(treeDepth)
   }
   #set checkStat 1 ;# Stat each file for updates
   set checkStat 0 ;# Stat each file for updates
   set isDirty [db::tableTreeData $baseIndex $col2Use $checkStat $depth $filter $filterHow totalTime totalSize tableData _Index2File]

   #-----------------------------------------
   # Insert the list configSet into the tablelist widget
   # Only if data changed or files added/removed  && ($isDirty || $llen != [llength $tableData] || ![string equal $baseIndex $lastFolder])
   if {$insert} {
      if {![string equal $baseIndex $lastFolder]} {
         $::PLB selection clear 0 end
      }
      # TABLEDATA is linked such that changing it changes
      # the displayed data.
      set trackList::TABLEDATA $tableData
      #array unset Index2File
      #array set Index2File [array get _Index2File]
      set ::TA(ck,totalSize) "Size: [commify $totalSize] KB"
      set ::TA(ck,totalFolders) "Files: [commify [llength $tableData]]"
      set days [expr {int($totalTime)/86400}]
      if {$days == 0} {
         set days ""
      }
      set ::TA(ck,totalTime) "Time:$days [timify $totalTime]"
      if {$sort} {
         ##
         # Sort by last used or by col 1 by default (col 0 is hidden)
         #
         set lastSort [$::PLB sortcolumn]
         set lastSortOrder [$::PLB sortorder]
         if {$lastSort >= 0} {
            $tbl sortbycolumn $lastSort -$lastSortOrder
         } elseif {$::snackAmpSettings(table,sortcol) >= 0} {
            if {[catch {
               $tbl sortbycolumn $::snackAmpSettings(table,sortcol) \
                  -$::snackAmpSettings(table,sortorder)
            } result] } then {
               debug $result
            }
         }
      }
   }
   #APL::busyCursor 0
   waitWindow::close
   set lastFolder $baseIndex
}

#------------------------------------------------------------------------------
# Function: trackList::getFiles
#
# Return a list of files within some folders
#------------------------------------------------------------------------------
proc trackList::getFiles {folders} {
   set result [list]
   foreach folder $folders {
      foreach candidate [db::filesInFolder $folder] {
         lappend result $candidate
      }
   }
   return $result
}
package provide autoPlayList 1.0
