#------------------------------------------------------------------------------
# snackAmp - graburl package 1.0 for snackAmp
# capture live streams to a file, supports N number of simultaneous streams.
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2003 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#------------------------------------------------------------------------------
namespace eval graburl {
   package require snack
   package require http 2.4
   variable linkedInfo
   variable top .gburlui
   variable status
}
#------------------------------------------------------------------------------
# Status user interface while downloading streams
#------------------------------------------------------------------------------
proc graburl::ui {sock} {
   variable top
   variable status
   set i 0
   while {[winfo exists $top.${i}url]} {
      incr i
   }
   set mytop $top.${i}url
   if {![winfo exists $top]} {
      toplevel $mytop
      wm title $mytop "URL Downloader Status"
      grid \
      [label $mytop.but -label "Cancel"] \
      [label $mytop.url -label $file] \
      [label $mytop.status -label "Received"]
   }
   set status($sock,rec) 0
   set status($sock,index) $i
   grid \
   [button $mytop.but -label "Cancel" -command graburl::cancel $i] \
   [label $mytop.url -label $file -relief sunken] \
   [label $mytop.status -textvariable graburl::status($i,rec) -relief sunken]
   return $i
}
#------------------------------------------------------------------------------
# Cancel a download, pass in index denoting slot
#------------------------------------------------------------------------------
proc graburl::cancel {sock} {
   variable status
   variable top
   if {[winfo exists $top.${i}but]} {
      grid forget $top.${i}but $top.${i}url $top.${i}status
      close $sock
      close $status($sock,fid)
      foreach {index} [array names status $sock*] {
         unset status($index)
      }
   }
}
#------------------------------------------------------------------------------
# Download a shortcut file, grab info from it for display
#------------------------------------------------------------------------------
proc graburl::downloadusk {file} {
   variable status
   Trace
   set url [graburl::readurl $file]
   if {[catch {graburl::grab $url} er] } then {
      return -code error $er
   } else {
      set status($sock,file) $file
      id3Tag::id3Label $file Data V1
      return [list $Data(Title) $Data(Album)]
   }
}
#------------------------------------------------------------------------------
# Play a url, after the headers the grabcb will file
# where we connect the socket to snack
# Note: -command xxx is used so the geturl
#------------------------------------------------------------------------------
proc graburl::grab {url} {
   Trace
   variable status
   if {[catch {
      http::geturl $url -handler {graburl::grabcb} \    "need to pass in file name"
                        -binary true \
                        -command graburl::done \
                        -blocksize 100000
   } result] } then {
      return -code error $result
   }
   return 1
}
#------------------------------------------------------------------------------
# Background callback, should only be called once
# since we disable fileevents. We highjack the socket
# from http and feed it directly into snack.
# This routine is first called right after the header
# is read and before any data has been read.
#
# Transfer to snack based on a nifty trick by Kristofer Lawson
#------------------------------------------------------------------------------
proc graburl::grabcb {sock token} {
   Trace
   variable linkedInfo
   variable status
   ##
   # If first time being called, get metadata info
   #
   if {![info exists status($sock)]} {
      set status($sock) 1
      ##
      # Configure the socket as binary, give it a big buffer
      #
      fconfigure $sock -translation {binary binary} -encoding binary \
         -blocking true -buffering full -buffersize 200000
      # remove "Connecting ..." from front of string
      set status($sock,Title) [string range $mylink(Title) 15 end]
      set status($sock,fDate) [clock format [clock seconds] -format "%x %X"]
      # Save any icycast data for use displaying on the front GUI
      if {[info exists stateVar(meta)]} {
         #parray stateVar
         array set icydata $stateVar(meta)
         foreach {icy ar} {
            icy-br          rate
            icy-genre       Genre
            icy-name        Title
            ice-description Album
            ice-genre       Genre
            ice-name        Title
            icy-url         Album
            icy-artist      Artist
            Content-Length  Size
            Last-Modified   Date
         } {
            if {[info exists icydata($icy)]} {
               set status($sock,icy,$ar) $icydata($icy)
            }
         }
      }
   }
   return 1
}
#------------------------------------------------------------------------------
# Validate a URL (streams don't typically have HEAD features, so not useful)
#------------------------------------------------------------------------------
proc graburl::validate {url} {
   Trace
   if {[catch {
      http::geturl $url -validate 1
   } token]} then {
      return -code error $token
   }
   upvar $token stateVar
   set code [lindex [http::code $token] 1]
   if {[string equal $code 200]} {
      return 1
   } else {
      return 0
   }
}
#------------------------------------------------------------------------------
# Proxy required callback, called before every get,
# Determines if the proxy is needed (local or remote).
#------------------------------------------------------------------------------
proc graburl::proxycb {url} {
   global snackAmpSettings
   if {[iptools::localUrl $url]==0 && $snackAmpSettings(useproxy)} {
      return [list $snackAmpSettings(proxyhost) $snackAmpSettings(proxyport)]
   } else {
      return [list]
   }
}
#------------------------------------------------------------------------------
# Callback when stream is done, not sure what to do here since snack
# is handling the stream. Don't think this is ever called
# since we cancel the file events
# Will be called on ::http::9 $token
#------------------------------------------------------------------------------
proc graburl::done {token} {
   variable linkedInfo
   Trace
   upvar $linkedInfo mylink
   set mylink(Title) "Done"
   soundControl::stop
   # Fire the next sound command
}
#------------------------------------------------------------------------------
# Cleanup routine, called from Stop button
#------------------------------------------------------------------------------
proc graburl::cleanupLast {sock} {
   Trace
   close $sock
}
#------------------------------------------------------------------------------
# Create a new url as a file using the
# ID3 tag info to hold
#------------------------------------------------------------------------------
proc graburl::addPromptUrl {folder} {
   set tl .apurl
   destroy $tl
   toplevel $tl
   wm title $tl "URL Shortcut in $folder"
   variable desc
   variable url
   variable ok
   set ok 0
   set i 0
   set fr [frame $tl.[incr i]]
   pack $fr -side top -fill x
   grid [label $fr.[incr i] -text "Enter Description"] \
        [entry $fr.[incr i] -textvariable graburl::desc -width 45]
   grid [label $fr.[incr i] -text "Enter http url"] \
        [entry $fr.[incr i] -textvariable graburl::url -width 45]

   # Line
   pack [frame $tl.[incr i] -height 2 -borderwidth 2 -relief sunken] -expand true -fill x

   # Button frame
   set bf [frame $tl.[incr i]]
   pack $bf -side bottom -fill none -expand false ;# This will center it
   #  ----------------------------------------------------------------------------
   #  Buttons at the bottom
   #  ----------------------------------------------------------------------------
   # Apply and Close
   set okBut [button $bf.[incr i] -text "OK" -width 8 -borderwidth 1 -command {
      set ::graburl::ok 1
      destroy .apurl
   } -relief raised]
   bind $tl <Return> [list $okBut invoke]

   # Cancel, restore settings
   set canBut [button $bf.[incr i] -text "Cancel" -width 8 -borderwidth 1 -command {
      set ::graburl::ok 0
      destroy .apurl
   } -relief raised]
   bind $tl <Escape> [list $canBut invoke]
   pack $okBut $canBut -side left -padx 20 -pady 5
   focus $fr.3
   centerWindow $tl
   myIcon $tl
   # Wait for user to close window
   grab $tl
   tkwait visibility $tl
   raise $tl
#  ----------------------------------------------------------------------------
#  Blocking
#  ----------------------------------------------------------------------------
   vwait ::graburl::ok
   if {$graburl::ok} {
      # Create the file
      graburl::addUrl $folder $graburl::desc $graburl::url
   }
}
#------------------------------------------------------------------------------
# Create a new url shortcut file, store info within
# the file itself.
#------------------------------------------------------------------------------
proc graburl::addUrl {folder description url} {
   set file [string trim [string map {/ - ? " " * " " : " " . _} $description]]
   set file [file join $folder $file.usk]
   if {[file exists $file]} {
      #saLog "graburl::addUrl: Overwriting $file"
   }
   if {[catch {open $file w} fid] } then {
      saLog "graburl::addUrl: $fid"
      return "Error with $file"
   } else {
      writeurl $url $fid
      close $fid
   }
   # Also put the url into the Album/Note fields to see in the APL table
   set url [HttpD::Decode $url]
   set parts [split $url ?]
   # TFW: Can clean up the Album/Artist part a little better
   set Data(Album) [lindex $parts 0]
   if {[llength $parts]>1} {
      set Data(Artist)  [lindex $parts 1]
   }
   set Data(Title) $description
   id3Tag::id3Modify $file Data V1  ;# no V2 for url shortcuts
   return $file
}
#------------------------------------------------------------------------------
# Grab the contents of some URL and return it
# Used to grab a pls/m3u file for use
# in decoding.
#------------------------------------------------------------------------------
proc graburl::getUrlBody {url} {
   #direct to contents
   package require http
   #set url "http://www.shoutcast.com/sbin/shoutcast-playlist.pls?m=1457&file=filename.pls"
   set token [::http::geturl $url]
   upvar #0 $token state
   return [string trim $state(body)]
}
#------------------------------------------------------------------------------
# Determine if a file is a url shortcut file
#------------------------------------------------------------------------------
proc graburl::isusk {file} {
   if {[string equal -nocase [file extension $file] ".usk"]} {
      return 1
   } else {
      return 0
   }
}
#------------------------------------------------------------------------------
# Read a url from a url shortcut
#------------------------------------------------------------------------------
proc graburl::readurl {file} {
   set fid [open $file r]
   set url [string trim [read $fid 256]]
   close $fid
   return $url
}
#------------------------------------------------------------------------------
# Write a url to a url shortcut
#------------------------------------------------------------------------------
proc graburl::writeurl {url fid} {
   set ulen [string length $url]
   if {[string length $url]} {
      set blank "$url[string repeat " " [expr {512-$ulen}]]"
      seek $fid 0 start
      puts -nonewline $fid $blank
   }

}
#------------------------------------------------------------------------------
# Verify we have a location to save the http shortcuts in
#------------------------------------------------------------------------------
proc graburl::checkuskfolder {args} {
   if {![file exists $::snackAmpSettings(uskDir)]} {
      set deffolder [file join [lindex $::TA(mediaDirs) 0] http]
      if {![file exists $deffolder] && [catch {file mkdir $deffolder} result]} {
         set deffolder [lindex $::TA(mediaDirs) 0]
      }
      tk_messageBox -title "Choose Folder" -type ok \
         -message "Please select a folder to save your URL shortcuts in.\n$deffolder is the default"
      set deffolder [chooseDirectory $deffolder [focus]]
      if {![file exists $deffolder]} {
         return 0
      } else {
         set ::snackAmpSettings(uskDir) $deffolder
         return [file exists $::snackAmpSettings(uskDir)]
      }
   } else {
      return 1
   }
}
package provide graburl 1.0
# Set up proxy callbacks
http::config -proxyfilter graburl::proxycb  -useragent "SnackAmp"
