#------------------------------------------------------------------------------
# id3Tag 1.0
#
# ID3 Lookup editor windows for snackAmp Player in Tcl/Tk
# Uses id3Tag package
# Copyright (C) 2003 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#
#------------------------------------------------------------------------------
namespace eval id3Editor {
#   option add *ID3Editors*Entry.background $::snackAmpSettings(bg) interactive ;# high priority
#   option add *ID3Editors*Entry.foreground $::snackAmpSettings(fg) interactive ;# high priority
}

#------------------------------------------------------------------------------
# Function   : id3Editor
# Description: Window to edit the ID3 info for a file
# Author     : Tom Wilkason
# Date       : 2/8/2001
#------------------------------------------------------------------------------
proc id3Editor::id3Editor {fnames} {
   set i 0
   option add *ID3Editors*Label.font entry interactive ;# high priority

   foreach fname $fnames {
      if {![file exists $fname]} {
         saLog "'$fname' Does not exist"
      }
   }
   switch -- [llength $fnames] {
      0 {
         tk_messageBox -type ok -icon warning -message "No files specified"
         return
      }
   }
   set here [focus]
   ##
   # Get a unique window, note $W is also a global variable
   # holding stuff for this instance of the GUI
   #
   while {[winfo exists .ide$i]} {incr i}
   set W .ide$i
   upvar #0 $W data
   upvar #0 $W.vars valueList
   toplevel $W -cursor $::TA(arrow)  -class ID3Editors
   wm protocol $W WM_DELETE_WINDOW [list id3Editor::closeWindow $W [me]]
   wm iconify $W
   ##
   # Determine if one or multiple files are being edited, behaviour
   # is different for multi-edits
   #
   if {[llength $fnames] == 1} {
      set multiEdit 0
      wm title $W $fname
   } else {
      set multiEdit 1
      wm title $W "Editing [llength $fnames] tracks"
   }
   set useFnames [list]
   ##
   # Process tags for each file, if a particular tag is different between
   # the files than set it to the special text $id3Tag::mmMatch to indicate
   # that field should not be re-applyed in common to all files upon an apply.
   #
   foreach {primTagType altTagType} $::snackAmpSettings(ID3ReadOrder) {break}
   set tagType "Unknown"
   foreach fname $fnames {
      Unset fnameData
      set fnameData(Tag) "$primTagType"
      #set hasTag [id3Tag::id3Label $fname fnameData $::snackAmpSettings(ID3ReadOrder)]  ;# use primary data
      set hasTag [id3Tag::id3Label $fname fnameData $primTagType]
      if {$hasTag==0} {
         set hasTag [id3Tag::id3Label $fname fnameData $altTagType]
      }
      if {$::snackAmpSettings(ID3OKCreate)==0 && ([llength $fnames] > 1 && $hasTag == 0)} {
         # Remove file name from consideration
         continue
      } else {
         set tagType "ID3$fnameData(Tag)"
         lappend useFnames $fname
      }
      # Note: TFW, files with no tag data don't set the multi-match field!! fnameData is blank!!
      # Bug, if all but one are blank then the non-blank field will be used instead of multi-match
      foreach {retField} $id3Tag::v1Parts {
         # If previously found one, see if it is different
         set fTag [id3Tag::blankForNull fnameData $retField]
         lappendUnqiue valueList($retField) $fTag
         if {[info exists data($retField)]} {
            # Set to multi match tag if all the fields are not the same.
            if {![string equal $data($retField) $fTag]} {
               set data($retField) $id3Tag::mmMatch
            }
         } else {
            # First time check, use actual data
            set data($retField) $fTag
         }
      }
   }
   # Get V2/ogg unique data so user can copy it over (if one file selected)
   if {! $multiEdit} {
      set data(fname) $fnames
      # This is a bug AFAIAK, file commands don't work if item is a single list length
      set ftest [lindex $fnames 0]
      if {[id3Tag::isogg $fname]} {
         id3Tag::id3Label $fname id3altData OGG
         set altTag "OGG"
      } else {
         id3Tag::id3Label $fname id3altData $altTagType
         if {[info exists id3altData(Tag)]} {
            set altTag "ID3$id3altData(Tag)"
         } else {
            set altTag "No Alternate"
         }
      }
   } else {
      set altTag "Not Shown"
      ##
      # Only edit files with existing tags unless
      # no files have tags, then use them.
      #
      if {[llength $useFnames] == 0} {
         set data(fname) $fnames
      } else {
         set data(fname) $useFnames
      }
   }
   ##
   # Build the GUI
   #
   set i 0
   set span 5
   set row 0
   # header
   grid [label $W.tag -text "Field"] [label $W.v2 -text "$altTag Tag"] x [label $W.v1 -text "$tagType Tag"]
   # TFW: Mod to only show the V1 labels/entries if multi-edit in process
   # Rows of all editable fields
   foreach {Index} $id3Tag::v1Parts {
      if {![info exists id3altData($Index)]} {
         set id3altData($Index) {}
      }
      label $W.v2lb$Index -text $id3altData($Index) \
         -anchor w -relief sunken -width 40
      balloon::define $W.v2lb$Index "$altTag Data for $Index"

      button $W.v2bt$Index -bd 0 -image saImages::NavForward \
         -command [list set ::${W}($Index) $id3altData($Index)]
      balloon::define $W.v2bt$Index "Copy $altTag Data '$id3altData($Index)' to V1 $Index field"

      label $W.lb$Index -text $Index -anchor e
      if {$Index=="Genre"} {
         if {![info exists data($Index)] || $data($Index)==""} {
            set data($Index) "Not Set"
         }
         # Make a cascaded menu like font names (or reduce options)
         menubutton $W.en$row -relief raised -indicatoron 1 -pady 0 \
            -textvariable ::${W}($Index) -menu $W.en$row.menu -direction flush
         balloon::define $W.en$row "Click to select a genre"
         widgetWatch::setWatch $W.en$row ::${W}($Index)
         menu $W.en$row.menu \
            -postcommand [list id3Editor::postGenreNames $W.en$row.menu ::${W}($Index)]
      } else {
         if {[llength $valueList($Index)]>1} {
            ComboBox $W.en$row \
                    -borderwidth 1 \
                    -width 50 \
                    -height 20 \
                    -hottrack true \
                    -textvariable ::${W}($Index) \
                    -editable true \
                    -values $valueList($Index)

            widgetWatch::setWatch $W.en$row.e ::${W}($Index)  ;# not .e is hardcoded for ComboBox
         } else {
            entry $W.en$row -textvariable ::${W}($Index) -width 50
         }
         balloon::define $W.en$row "$tagType Data for $Index"
      }
      ##
      # Bind up/down keys to move in column
      #
      bind $W.en$row <Up> [list Focus $W.en[expr {$row-1}]]
      bind $W.en$row <Down> [list Focus $W.en[expr {$row+1}]]
      ##
      # Links to allmusic site for certain fields
      #
      if {[lsearch -exact [list Artist Album Title] $Index] > -1} {
         button $W.http$Index -bd 0 -image saImages::web -command [list browser::allMusic ::${W}($Index) $Index]
         balloon::define $W.http$Index "Search the All Music Guide web site for $Index information"
         grid $W.lb$Index $W.v2lb$Index $W.v2bt$Index $W.en$row $W.http$Index -sticky ew
      } else {
         grid $W.lb$Index $W.v2lb$Index $W.v2bt$Index $W.en$row -sticky ew
      }
      incr row
   }
   ##
   # Rename file based on tag
   #
   if {! $multiEdit} {
      button $W.lbTag2File -text "Rename File" \
         -relief raised -command [list id3Editor::id3Rename ::${W}(fname) $W.enTag2File]
      balloon::define $W.lbTag2File "Push this to Rename the file to the name on the right"
      set data(newName) [file rootname [file tail $data(fname)]]
      entry $W.enTag2File -textvariable ::${W}(newName)
      widgetWatch::setWatch $W.enTag2File ::${W}(newName)

      grid $W.lbTag2File $W.enTag2File - - -sticky ew

      ##
      # If a URL shortcut, put that info in also
      #
      if {[playurl::isusk $ftest]} {
         set data(URL) [playurl::readurl $ftest]
         label $W.lbURL -text "Change URL"
         balloon::define $W.lbURL "Change the URL this URL shortcut refers to"
         set data(newURL) $data(URL)
         entry $W.enlbURL -textvariable ::${W}(newURL)
         widgetWatch::setWatch $W.enlbURL ::${W}(newURL)
         grid $W.lbURL $W.enlbURL - - -sticky ew
      }
   }

   grid columnconfigure $W 1 -weight 1
   grid columnconfigure $W 3 -weight 2
   # Line
   grid [horzLine $W.l[incr i]] -pady 3 -columnspan 5 -sticky nsew

   ##
   # Frame for OK/Cancel/Apply buttons
   #
   grid [frame $W.but] -columnspan 5

   button $W.but.lbcddb -text "Freedb Query" -default normal \
      -command [list browser::cddb ::$W] \
      -borderwidth 1  -relief raised
   balloon::define $W.but.lbcddb "Push this to guess a file name based on the tag"
   button $W.but.coverart -text "Art Query" -default normal \
      -command [list browser::coverart ::$W] \
      -borderwidth 1  -relief raised
   balloon::define $W.but.coverart "Push this to search google for cover art images"

   if {! $multiEdit} {
      ##
      # Tag file based on name
      #
      button $W.but.lbMakeFname -text "Guess Name" -default normal \
         -command [list id3Editor::id3Editor_guessFileName ::$W newName] \
         -borderwidth 1  -relief raised
      balloon::define $W.but.lbMakeFname "Push this to guess a file name based on the tag"

      button $W.but.lbFile2Tag -text "Guess Tag" -default normal \
         -command [list id3Editor::cbSmartTagName ::${W}(fname) ::$W] \
         -borderwidth 1  -relief raised
      balloon::define $W.but.lbFile2Tag "Push this to create a tag based on the file name"
   }
   #fix
   set fix [button $W.but.[incr i] -width 8 -default normal -text "Fix Case" -relief raised \
      -command [list id3Editor::id3Editor_SingleTagToNameFix ::$W] -borderwidth 1  -relief raised]
   balloon::define $fix "Push this to automatically correct file name spacing and capitalization"

   # Apply and close
   set ok [button $W.but.[incr i] -width 8 -default active -text "OK" \
      -command [list id3Editor::id3Editor_cbModify ::${W}(fname) $W [me] 1] \
      -borderwidth 1  -relief raised]
   balloon::define $ok \
   "Push this to apply and save the current settings then close this dialog\n[join [set ::${W}(fname)]\n]"

   # Apply
   set app [button $W.but.[incr i] -width 8 -default normal -text "Apply" \
      -command [list id3Editor::id3Editor_cbModify ::${W}(fname) $W [me] 0] \
      -borderwidth 1  -relief raised]
   balloon::define $app "Push this apply and save the current settings\n[join [set ::${W}(fname)]\n]"

   # Cancel
   set can [button $W.but.[incr i] -width 8 -default normal -text "Cancel" \
      -command [list id3Editor::closeWindow $W [me]] \
      -borderwidth 1 -relief raised]
   balloon::define $can "Push this to ignore changes made and restore the previous settings"
   if {! $multiEdit} {
      grid $W.but.coverart $W.but.lbcddb $W.but.lbMakeFname $W.but.lbFile2Tag $fix $ok $app $can -padx [extraPad 1] -pady 1
   } else {
      grid $W.but.coverart $W.but.lbcddb $fix $ok $app $can -padx [extraPad 1] -pady 1
   }
   ##
   # Key bindings
   #
   $ok configure -default active
   bind $W <Return> [list $ok invoke]
   bind $W <Escape> [list $can invoke]
   settings::restoreWidth $W [me]
   wm deiconify $W
   ontop::pinOnSystemMenu {} $W
   myIcon $W
   focus $W
}

#------------------------------------------------------------------------------
# Function   : id3Editor::id3Editor_SingleTagToNameFix
# Description: Clean the file name
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc id3Editor::id3Editor_SingleTagToNameFix {_W {patt *}} {
   upvar #0 $_W W
   global fnameData ;# needs to be global
   foreach {element} [lsort [array names W $patt]] {
      set fname $W($element)
      # Only fix certain fields
      if {[lsearch -exact [list Title Artist Album newName] $element]>= 0} {
         set fn [fixplus::cleanName $fname]
      } else {
         set fn $fname
      }
      if {[string length $fn] > 0 && ![string equal $fn $fname]} {
         # Use orig name, traced variable
         set ${_W}($element) $fn
      }
   }
}

#------------------------------------------------------------------------------
# Function   : id3Editor_cbModify
# Description: Callback to apply the ID3 changes
# Author     : Tom Wilkason
# Date       : 2/7/2001
#------------------------------------------------------------------------------
proc id3Editor::id3Editor_cbModify {_fnames W winTag {close 0}} {
   upvar $_fnames fnames
   upvar $W data
   set todo [llength $fnames]
   set i 0
   if {$::snackAmpSettings(ID3ReadOnly)==0} {
      foreach {fname} $fnames {
         incr i
         waitWindow::percent $i $todo "Setting Tag for\n[file tail $fname]" 1
         if {[waitWindow::ifbreak]} {
            break
         }
         # Make a copy of the common array since it is modified in place
         array unset localArray
         array set localArray [array get ::$W]
         # Write the data
         id3Tag::id3Modify $fname localArray
         ##
         # Update the URL data at the beginning of the data if required.
         #
         if {[info exists data(URL)] && [info exists data(newURL)] && ![string equal $data(URL) $data(newURL)]} {
            if {[catch {
               set fid [open $fname a]
               seek $fid 0 start
               playurl::writeurl $data(newURL) $fid
               close $fid
            } result] } then {
               debug $result
               catch {close $fid}
            }
         }
         ##
         # If this song is the currently linked song then
         # update the array that is linked to the displayed
         # fields.
         if {[string equal [soundControl::fileName] $fname]} {
            array set ::songArray [array get localArray]
         }
      }
      waitWindow::close
   }
   # Can't close window till done, need data from it above
   if {$close} {
      id3Editor::closeWindow $W $winTag
   }
}

#------------------------------------------------------------------------------
# Function   : id3Editor::id3Editor_guessFileName
# Description: Guess a file name
# Author     : Tom Wilkason
# Date       : 3/19/2002
#------------------------------------------------------------------------------
proc id3Editor::id3Editor_guessFileName {_W index} {
   upvar $_W Data
   set Data($index) [smartFileName $_W]
}

#------------------------------------------------------------------------------
# Function   : id3NameToTag
# Description: Window to edit the proposed tag names based on file name data
# Author     : Tom Wilkason
# Date       : 2/8/2001
#------------------------------------------------------------------------------
proc id3Editor::id3NameToTag {fnames {smart 0}} {
   set i 0
   set tr 0
   if {[llength $fnames] == 0} {
      tk_messageBox -type ok -icon warning -message "No files specified"
      return
   }
   foreach fname $fnames {
      if {![file exists $fname]} {
         saLog "'$fname' Does not exist"
      }
   }
   set here [focus]
   # Get a unique window, note $W is also a global variable
   # holding stuff for this instance of the GUI
   while {[winfo exists .ntIdent$i]} {incr i}
   set W .ntIdent$i
   upvar $W MyVal
   toplevel $W -cursor $::TA(arrow)  -class ID3Editors
   if {$smart} {
      wm title $W "Name To Tag Editor"
   } else {
      wm title $W "ID3 Table Editor"
   }
   wm iconify $W
   wm protocol $W WM_DELETE_WINDOW [list id3Editor::closeWindow $W [me]]
   # get current data and line to global with the same name
   # as this widget (to allow multiple editors at once)
   # Save current file name
   # Note: The ${W}($fname,tag) represent what goes into the GUI fields
   ##
   # Build the GUI
   #
   # Decide when to use a scroll bar and how big to make the screen,
   # don't want to make any bigger than necessary.
   set needHeight [expr {round([llength $fnames] * [font metrics entry -linespace]*1.5)}]
   set maxHeight [winfo pixels . $::snackAmpSettings(scrolledCanvasHeight)]
   if {$maxHeight > $needHeight} {
      set maxHeight $needHeight
   }
   set sw [ScrolledWindow $W.data -relief flat -borderwidth 1]
   set sf [ScrollableFrame $sw.f -constrainedwidth true -height $maxHeight]
   $sw setwidget $sf
   set hull [$sf getframe]
   set span 6
   # Genre menubutton
   # Artist and Albmum menu buttons
   foreach {Type} [list Artist Album Genre] {
      set ::${W}($Type) "$Type"
      set mb [menubutton $hull.lb$Type -relief raised -indicatoron 1 -pady 0 \
          -text [set ::${W}($Type)] -menu $hull.lb$Type.menu -direction flush -width 10]
      balloon::define $mb "Click to select a common $Type for all files"
      if {$Type=="Genre"} {
         menu $mb.menu \
            -postcommand [list id3Editor::postGenreNames $mb.menu ::${W}($Type)]
      } else {
         menu $mb.menu \
            -postcommand [list id3Editor::id3NameToTag_postTypeNames $mb.menu ::${W} $Type]
      }
      trace variable ::${W}($Type) w id3Editor::id3NameToTag_commonType
      # When variable is updated during postcommand, call id3NameToTag_commonType to set others to same
   }

   # Grid out the labels
   grid \
        [label $hull.lb[incr i] -text "Guess Tag"] \
        [label $hull.lb[incr i] -text "Change"]  \
        $hull.lbArtist \
        $hull.lbAlbum  \
        [label $hull.lb[incr i] -text "Title"] \
        [label $hull.lb[incr i] -text "Track"] \
        $hull.lbGenre -sticky ew
   set i 0
   ##
   # Build one row for each file
   #
   set tag 0
   # Data
   set ent 1
   foreach {primTagType altTagType} $::snackAmpSettings(ID3ReadOrder) {break}
   foreach fname $fnames {
      set toGrid [list]
      # Get existing tag data
      # Determine data from parsed name
      # Only updates entries where
      # the file name provides it
      foreach {field} {Artist Album Title Track Genre} {
         set fnameData($field) ""
      }
      # Build smart tag if desired
      if {$smart} {
         cbSmartTagName fname fnameData
      }
      # Overlay with existing ID3 data
      set hasTag [id3Tag::id3Label $fname fnameData $::snackAmpSettings(ID3ReadOrder)]
      #set hasTag [id3Tag::id3Label $fname fnameData]
      set ::${W}($tag,oldFullName) $fname
      if {$hasTag} {
         foreach {field} {Artist Album Title Track Genre} {
            set ::${W}($tag,$field) $fnameData($field)
         }
      }
      # Default to true
      set ::${W}($tag,ok) 1
      lappend toGrid [button $hull.bt[incr i] -relief raised -highlightthickness 0 -pady 0 -text "Guess" \
         -command [list id3Editor::id3NameToTag_cbOneSmartTagName $fname $tag ::${W}]]
      balloon::define $hull.bt$i "Press this to guess the tags for \n$fname"

      lappend toGrid [checkbutton $hull.en[incr i] -variable ::${W}($tag,ok) -font {Helvetica 5}]
      balloon::define $hull.en$i "Check this to have OK/Apply modify the tag in \n$fname"
      foreach {field width} {Artist 20 Album 20 Title 30 Track 3} {
         set item [entry $hull.en$field$ent -width $width -textvariable ::${W}($tag,$field)]
         widgetWatch::setWatch $item ::${W}($tag,$field)
         if {$field=="Track"} {
            $item configure -justify right
         }
         ##
         # Bind up/down keys to go up down in column
         #
         bind $item <Up> [list Focus $hull.en$field[expr {$ent-1}]]
         bind $item <Down> [list Focus $hull.en$field[expr {$ent+1}]]
         lappend toGrid $item
         balloon::define $item "Modify $field data then press OK or Apply to make the change. Ctrl-N to auto number."
         if {$field=="Track"} {
              # Bind Control N to track to auto number those below
              bind $item <Control-n> [list id3Editor::id3NameToTag_NumTrack $hull $ent]
         } elseif {[lsearch -exact [list Title Artist Album] $field]>= 0} {
              bind $item <Control-f> [list id3Editor::id3NameToTag_NumCase $hull [list $field]]
         }
      }
      incr ent
      # Make a cascaded menu of genres
      set mb [menubutton $hull.en[incr i] -relief raised -indicatoron 1 -pady 0 \
         -textvariable ::${W}($tag,Genre) -menu $hull.en$i.menu -direction flush -width 10]
      balloon::define $mb "Click to select a genre"
      menu $mb.menu \
         -postcommand [list id3Editor::postGenreNames $mb.menu ::${W}($tag,Genre)]
      widgetWatch::setWatch $mb ::${W}($tag,Genre)
      lappend toGrid $mb
      eval grid $toGrid -sticky ew
      incr tag
   }
   # Ensure only the data columns expand
   grid columnconfigure $hull 2 -weight 1
   grid columnconfigure $hull 3 -weight 1
   grid columnconfigure $hull 4 -weight 1
   ##
   # Frame for OK/Cancel/Apply buttons
   #
   pack [frame $W.but] -fill x -expand false -side bottom
   # Line
   pack [horzLine $W.l[incr i]] \
      -fill x -expand false -side bottom
   # Data
   pack $W.data -side top -expand true -fill both
   set wid 9
   # Check All
   set checkAll [button $W.but.[incr i] -width $wid -default normal -text "Check All" \
      -command [list id3Editor::cbSetOK ::$W 1] \
      -borderwidth 1 -relief raised]
   balloon::define $checkAll "Push this to check all checkboxes"

   # Check None
   set checkNone [button $W.but.[incr i] -width $wid -default normal -text "Check None" \
      -command [list id3Editor::cbSetOK ::$W 0] \
      -borderwidth 1 -relief raised]
   balloon::define $checkNone "Push this to un-check all checkboxes"

   # Guess Tags
   set guessTags [button $W.but.[incr i] -width $wid -default normal -text "Guess Tags" \
      -command [list id3Editor::id3NameToTag_cbManySmartTagName ::$W] \
      -borderwidth 1  -relief raised]
   balloon::define $checkNone "Push this to guess tags based on the file name"
   # Guess Tags

   set v2tag [button $W.but.[incr i] -width $wid -default normal -text "Alt Tag" \
      -command [list id3Editor::id3NameToTag_cbManygetV2data $altTagType ::$W] \
      -borderwidth 1  -relief raised]
   balloon::define $checkNone "Push this to copy in ID3V2 data (if it exists)"

   # Guess Case
   set guessCase [button $W.but.[incr i] -width $wid -default normal -text "Fix Case" \
      -command [list id3Editor::id3NameToTag_NumCase $hull {Artist Album Title}] \
      -borderwidth 1  -relief raised]
   balloon::define $checkNone "Push this to correct the text case in the tags"

   # Apply and close
   set ok [button $W.but.[incr i] -width $wid -default active -text "OK" \
      -command [list id3Editor::id3NameToTag_apply ::$W [me] 1] -borderwidth 1  -relief raised]
   balloon::define $ok "Push this to modify the ID3 tags for checked entries, then close this dialog"

   # Apply
   set app [button $W.but.[incr i] -width $wid -default normal -text "Apply" \
      -command [list id3Editor::id3NameToTag_apply ::$W [me] 0] \
      -borderwidth 1  -relief raised]
   balloon::define $app "Push this to modify the ID3 tags for checked entries"

   # Cancel
   set can [button $W.but.[incr i] -width $wid -default normal -text "Cancel" \
      -command [list id3Editor::closeWindow $W [me]] \
      -borderwidth 1 -relief raised]
   balloon::define $can "Push this to ignore changes made and restore the previous settings"
   grid $checkAll $checkNone $v2tag $guessTags $guessCase $ok $app $can  -padx [extraPad 1] -pady 2
   ##
   # Key bindings
   #
   $ok configure -default active
   bind $W <Return> [list $ok invoke]
   bind $W <Escape> [list $can invoke]

   settings::restoreWidth $W [me]
   wm deiconify $W
   ontop::pinOnSystemMenu {} $W
   myIcon $W
   focus $W
}
#------------------------------------------------------------------------------
# Function   : id3Editor::id3NameToTag_NumTrack
# Description: Auto number tracks below based on this one
#
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc id3Editor::id3NameToTag_NumTrack {hull index} {
   set len [string length $hull.enTrack]
   set start 1
   foreach {entry} [lsort -dictionary [winfo children $hull]] {
      if {[string match $hull.enTrack* $entry]} {
         set enIndex [string range $entry $len end]
         set gvalue [$entry cget -textvariable]
         if {$enIndex == $index} {
            set start [set $gvalue]
            if {![string is integer -strict $start]} {
               set start 1
               set $gvalue $start
            }
         } elseif {$enIndex >= $index} {
            incr start
            set $gvalue $start
         }
      }
   }
}

#------------------------------------------------------------------------------
# Function   : id3Editor::id3NameToTag_NumCase
# Description: Correct the case for each item in this column (field)
#
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc id3Editor::id3NameToTag_NumCase {hull fields} {
   foreach {entry} [lsort -dictionary [winfo children $hull]] {
      foreach field $fields {
         if {[string match $hull.en${field}* $entry]} {
            # Need $gvalue due to trace on it
            set gvalue [$entry cget -textvariable]
            set $gvalue [stringTitle [set $gvalue]]
         }
      }
   }
}

#------------------------------------------------------------------------------
# Function   : id3Editor::id3NameToTag_commonType
# Description: Callback to change all genres for block editor
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc id3Editor::id3NameToTag_commonType {_Glob Element args} {
   upvar $_Glob Glob
   set value $Glob($Element)
   foreach {item} [array names Glob "*,$Element"] {
      # Traced, use full global name
      set ${_Glob}($item) $value
   }
}

#------------------------------------------------------------------------------
# Function   : id3Editor::id3NameToTag_apply
# Description: Retag a set of files
#
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc id3Editor::id3NameToTag_apply {_W winTag {close 0}} {
   upvar #0 $_W W
   # Process each file that is checked
   set indicies [lsort [array names W *,ok]]
   set i 0
   foreach {element} $indicies {

      # If checkbox checked, try to rename it
      if {$W($element)} {
         set tag [lindex [split $element ,] 0]
         set fname $W($tag,oldFullName)
         # Blank then Get existing data
         array unset fnameData
         #id3Tag::id3Label $fname fnameData
         set changed 0
         foreach {field} {Artist Album Title Track Genre} {
            # Copy in only existing non-blank data
            if {[info exists W($tag,$field)] && [string length $W($tag,$field)] > 0} {
               set fnameData($field) $W($tag,$field)
               incr changed
            }
         }
         # Don't bother to update if the data hasn't changed
         if {$changed} {
            waitWindow::percent $i [llength $indicies] "Setting Tag for\n[file tail $fname]"
            id3Tag::id3Modify $fname fnameData
         }
      }
      incr i
   }
   waitWindow::close
   # If close widget when done indicated.
   if {$close} {
      id3Editor::closeWindow $_W $winTag
   }
}

#------------------------------------------------------------------------------
# Function   : id3NameToTag_cbOneSmartTagName
# Description: Move data from a single smart tag into an indexed array
#              used by the block editor.
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc id3Editor::id3NameToTag_cbOneSmartTagName {fname index _Array} {
   upvar $_Array Data
   cbSmartTagName fname oneData
   foreach {item} [array names oneData] {
      if {![string equal $Data($index,$item) $oneData($item)]} {
         # Traced, use full global name
         set ${_Array}($index,$item) $oneData($item)
         set ${_Array}($index,ok) 1
      }
   }
}

#------------------------------------------------------------------------------
# Function   : id3NameToTag_cbManySmartTagName
# Description: Create a number of smart tag to name mappings
#
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc id3Editor::id3NameToTag_cbManySmartTagName {_Array} {
   upvar $_Array Data
   foreach {item} [array names Data *,oldFullName] {
      set fname $Data($item)
      set index [lindex [split $item ,] 0]
      Unset oneData
      cbSmartTagName fname oneData
      foreach {item} [array names oneData] {
         if {![string equal $Data($index,$item) $oneData($item)]} {
            # Traced, use full global name
            set ${_Array}($index,$item) $oneData($item)
            set ${_Array}($index,ok) 1
         }
      }
   }
}

#------------------------------------------------------------------------------
# Function   : id3NameToTag_cbManygetV2data
# Description: Read the ID3V2/ogg data and if it exists use
#              it instead
#              used by the block editor.
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc id3Editor::id3NameToTag_cbManygetV2data {altTagType _Array} {
   upvar $_Array Data
   foreach {item} [array names Data *,oldFullName] {
      set fname $Data($item)
      set index [lindex [split $item ,] 0]
      Unset oneData
      if {[id3Tag::isogg $fname]} {
         set type "OGG"
      } else {
         set type $altTagType
      }
      if {[id3Tag::id3Label $fname oneData $type]} {
         foreach field {Title Artist Album Track} {
            if {[info exists oneData($field)]} {
               if {$field=="Track"} {
                  # Just want the base number
                  set val [lindex [split $oneData($field) /] 0]
               } else {
                  set val $oneData($field)
               }
               set ${_Array}($index,$field) $val
            }
         }
         set ${_Array}($index,ok) 1
      }
   }
}

#------------------------------------------------------------------------------
# Function   : id3NameToTag_postTypeNames
# Description: On pop-up, post the common types
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc id3Editor::id3NameToTag_postTypeNames {Menu _Array Element} {
   upvar #0 $_Array Array
   ##
   # Get list of current entries
   #
   set last [$Menu index end]
   set added [list]
   if {![string match "none" $last]} {
      for {set i 0} {$i <= $last} {incr i} {
         lappend added [$Menu entrycget $i -label]
      }
   }
   ##
   # Add any unique items in the array to the menu
   #
   foreach {item} [lsort [array names Array "*,$Element"]] {
      set value $Array($item)
      if {[string length $value] > 0 && [lsearch $added $value]==-1} {
         # TFW: What we really want is to set te $item fields directly
         # not the Element which is also a label
         $Menu add command -label "$value" \
            -command [list set ${_Array}($Element) $value]
         lappend added $value
      }
   }
}

#------------------------------------------------------------------------------
# Function   : id3TagToName
# Description: Window to edit the proposed file names based on tag data
# Author     : Tom Wilkason
# Date       : 2/8/2001
#------------------------------------------------------------------------------
proc id3Editor::id3TagToName {fnames} {
   set i 0
   global fnameData ;# Note, due to upvar #0, TFW Fix later
   if {[llength $fnames] == 0} {
      tk_messageBox -type ok -icon warning -message "No files specified"
      return
   }
   foreach fname $fnames {
      if {![file exists $fname]} {
         saLog "'$fname' Does not exist"
      }
   }
   set here [focus]
   # Get a unique window, note $W is also a global variable
   # holding stuff for this instance of the GUI
   while {[winfo exists .tnIdent$i]} {incr i}
   set W .tnIdent$i
   toplevel $W -cursor $::TA(arrow)  -class ID3Editors
   wm title $W "Tag To Name Editor"
   wm protocol $W WM_DELETE_WINDOW [list id3Editor::closeWindow $W [me]]
   wm iconify $W
   upvar #0 $W myData
   ##
   # Build the GUI
   #
   set i 0
   set span 3
   set header [frame $W.header]
   ##
   # Frame to hold the Name guessing
   #
   set hull [labeledFrame $header.rep -text "Name Template (for Guessing)"]
   set ::TA(smartName) $::snackAmpSettings(smartName)
   grid [button $hull.snl -highlightthickness 0 -pady 0 -text "Guess" -relief raised -command [list id3Editor::id3TagToName_Guess ::$W]] \
        [entry $hull.sne -textvariable ::snackAmpSettings(smartName)] \
        -sticky ew
         widgetWatch::setWatch $hull.sne ::snackAmpSettings(smartName)
   balloon::define $hull.sne \
      "Enter naming scheme for converting tags to file names:\n%%A - Artist, %%C - Album, %%N - Track Number, %%T - Title"
   balloon::define $hull.snl "Push this to guess the correct file names based on tags"
   ##
   # Frame to hold the replacement text
   #
   set hull [labeledFrame $header.entry -text "Replace Text"]
   # header
   grid [button $hull.rlbh -text "Replace" -relief raised -command [list id3Editor::id3TagToName_Substitute ::$W]] \
        [entry $hull.rlbf -textvariable ::${W}(fromText) -width 30] \
        [entry $hull.rlbt -textvariable ::${W}(toText) -width 30] \
        -sticky ew
   # Expand entry boxes
   grid columnconfigure $hull 1 -weight 1
   grid columnconfigure $hull 2 -weight 1

   balloon::define $hull.rlbh "Push this to replace text in names (Change field must be checked)"
   balloon::define $hull.rlbf "Enter Text to Replace here"
   balloon::define $hull.rlbt "Enter Replacement Text here"
   pack $header.rep -side left
   pack $header.entry -side left -fill x -expand true
   pack $header -side top -anchor w -fill x -expand false

   # Line
   pack [horzLine $W.l[incr i]] \
      -expand false -fill x -side top

   # Decide when to use a scroll bar and how big to make the screen,
   # don't want to make any bigger than necessary.
   set needHeight [expr {round([llength $fnames] * [font metrics entry -linespace]*1.5)}]
   set maxHeight [winfo pixels . $::snackAmpSettings(scrolledCanvasHeight)]
   if {$maxHeight > $needHeight} {
      set maxHeight $needHeight
   }
   set sw [ScrolledWindow $W.data -relief flat -borderwidth 1]
   set sf [ScrollableFrame $sw.f -constrainedwidth true -height $maxHeight]
   $sw setwidget $sf
   set hull [$sf getframe]
   # header
   grid [label $hull.lb[incr i] -text "Change"] \
        [label $hull.lb[incr i] -text "Old Name"] \
        [label $hull.lb[incr i] -text "New Name"] \
        -sticky ew
   ##
   # Build one row for each file
   #
   set tag 0
   foreach fname $fnames {
      # Get existing tag data
      # Propose new file name
      set myData($tag,oldFullName) $fname
      set fn [file rootname [file tail $fname]]
      set myData($tag,oldName) $fn
      set myData($tag,newName) $fn
      set myData($tag,ok) 1
      # Note using a very small font allows the checkbuttons to pack tighter
      grid [checkbutton $hull.cb$i -variable ::${W}($tag,ok) -font {Helvetica 5}] \
         [label $hull.ln$i -width 50 -textvariable ::${W}($tag,oldName) -relief sunken -anchor w] \
         [entry $hull.en$i -width 50 -textvariable ::${W}($tag,newName)] \
         -sticky ew
         widgetWatch::setWatch $hull.en$i ::${W}($tag,newName)
         ##
         # Bind up/down keys to go up down in column
         #
         bind $hull.en$i <Up> [list Focus $hull.en[expr {$i-1}]]
         bind $hull.en$i <Down> [list Focus $hull.en[expr {$i+1}]]

      incr tag
      incr i
   }
   grid columnconfigure $hull 1 -weight 1
   grid columnconfigure $hull 2 -weight 1
   ##
   # Frame for OK/Cancel/Apply buttons
   #
   pack [frame $W.but] -fill x -expand false -side bottom
   # Line
   pack [horzLine $W.l[incr i]] \
      -expand false -fill x -side bottom
   # Data
   pack $W.data -side top -expand true -fill both
   # Check All
   set checkAll [button $W.but.[incr i] -default normal -text "Check All" \
      -command [list id3Editor::cbSetOK ::$W 1] \
      -borderwidth 1  -relief raised]
   balloon::define $checkAll "Push this to check all checkboxes"

   # Check None
   set checkNone [button $W.but.[incr i] -default normal -text "Check None" \
      -command [list id3Editor::cbSetOK ::$W 0] \
      -borderwidth 1  -relief raised]
   balloon::define $checkNone "Push this to un-check all checkboxes"

   #fix
   set fix [button $W.but.[incr i] -width 8 -default normal -text "Fix Case" -relief raised \
      -command [list id3Editor::id3TagToName_Fix ::$W] -borderwidth 1  -relief raised]
   balloon::define $fix "Push this to automatically correct file name spacing and capitalization"

   # Apply and close
   set ok [button $W.but.[incr i] -width 8 -default active -text "OK" \
      -command [list id3Editor::cbTagToName ::$W [me] 1] -borderwidth 1  -relief raised]
   balloon::define $ok "Push this to rename the changed files that are checked, then close this dialog"
   # Apply
   set app [button $W.but.[incr i] -width 8 -default normal -text "Apply" \
      -command [list id3Editor::cbTagToName ::$W [me] 0] \
      -borderwidth 1  -relief raised]
   balloon::define $app "Push this rename the changed files that are checked"

   # Cancel
   set can [button $W.but.[incr i] -width 8 -default normal -text "Cancel" \
      -command [list id3Editor::closeWindow $W [me]] \
      -borderwidth 1 -relief raised]
   balloon::define $can "Push this to ignore changes made and restore the previous settings"
   grid $checkAll $checkNone $fix $ok $app $can  -padx [extraPad 1] -pady 2
   ##
   # Key bindings
   #
   $ok configure -default active
   bind $W <Return> [list $ok invoke]
   bind $W <Escape> [list $can invoke]
   settings::restoreWidth $W [me]
   wm deiconify $W
   ontop::pinOnSystemMenu {} $W
   myIcon $W
   focus $W
}
#------------------------------------------------------------------------------
# Function   : id3Editor::id3TagToName_Substitute
# Description: Replace some text in enabled file names
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc id3Editor::id3TagToName_Substitute {_W} {
   upvar #0 $_W W
   foreach {element} [lsort [array names W *,ok]] {
      foreach {key xx} [split $element ,] {break}
      if {$W($element)} {
         # _W is traced variable, use full name when setting
         #regsub  -- $W(fromText) $W($key,oldName) $W(toText) ${_W}($key,newName)
         #regsub -- $W(fromText) $W($key,newName) $W(toText) ${_W}($key,newName)
         set ${_W}($key,newName) [string map [list $W(fromText) $W(toText)] [set ${_W}($key,newName)]]
      }
   }
}
#------------------------------------------------------------------------------
# Function   : id3Editor::id3TagToName_Guess
# Description: Guess the appropriate name based on a template
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc id3Editor::id3TagToName_Guess {_W} {
   upvar #0 $_W W
   global fnameData ;# needs to be global
   foreach {element} [lsort [array names W *,ok]] {
      foreach {tag xx} [split $element ,] {break}
      Unset fnameData
      set fname $W($tag,oldFullName)
      id3Tag::id3Label $fname fnameData $::snackAmpSettings(ID3ReadOrder)
      set fn [file rootname [file tail $fname]]
      set sfn [smartFileName fnameData]
      if {[string length $sfn] > 0 && ![string equal $sfn $fn]} {
         # Use orig name, traced variable
         set ${_W}($tag,newName) $sfn
         set ${_W}($tag,ok) 1
      }
   }
}
#------------------------------------------------------------------------------
# Function   : id3Editor::id3TagToName_Fix
# Description: Clean the file name
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc id3Editor::id3TagToName_Fix {_W {patt *,ok}} {
   upvar #0 $_W W
   global fnameData ;# needs to be global
   foreach {element} [lsort [array names W $patt]] {
      foreach {tag xx} [split $element ,] {break}
      set fname $W($tag,newName)
      set fn [fixplus::cleanName $fname]
      if {[string length $fn] > 0 && ![string equal $fn $fname]} {
         # Use orig name, ${_W} is a traced variable
         set ${_W}($tag,newName) $fn
         set ${_W}($tag,ok) 1
      }
   }
}
#------------------------------------------------------------------------------
# Function   : id3Editor::cbTagToName
# Description: Rename a set of files
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc id3Editor::cbTagToName {_W winTag {close 0}} {
   upvar #0 $_W W
   foreach {element} [lsort [array names W *,ok]] {
      # If checkbox checked, try to rename it
      if {$W($element)} {
         set tag [lindex [split $element ,] 0]
         set oldFullName $W($tag,oldFullName)
         set newBase     $W($tag,newName)
         set newFullName [file join [file dirname $oldFullName] "$W($tag,newName)[file extension $oldFullName]"]
         waitWindow::show "Renaming [file tail $oldFullName]..."
         if {![string equal $oldFullName $newFullName]} {
            if {[catch {catalog::RenameFile $oldFullName $newFullName} result] } then {
               tk_messageBox -icon warning -message "Could not rename\n$oldFullName to \n$newFullName\n$result"
            } else {
               catalog::RenameEntry $oldFullName $newFullName
               # Update the fields
               set W($tag,oldFullName) $newFullName
               set W($tag,oldName)     [file rootname [file tail $newFullName]]
            }
         }
      }
   }
   waitWindow::close
   # If close widget when done indicated.
   if {$close} {
      id3Editor::closeWindow $_W $winTag
   }
}
#----------------------------------------------------------------------------
# Common Functions between editors
#----------------------------------------------------------------------------
#------------------------------------------------------------------------------
# Callback to set/clear checkboxes
#------------------------------------------------------------------------------
proc id3Editor::cbSetOK {_W how} {
   upvar #0 $_W W
   foreach {field} [lsort [array names W *,ok]] {
      set W($field) $how
   }
}
#------------------------------------------------------------------------------
# Function   : id3Editor::id3Rename
# Description: Rename a file based on the tag name
# Author     : Tom Wilkason
# Date       : 3/18/2002
#------------------------------------------------------------------------------
proc id3Editor::id3Rename {_fname widget} {
   upvar $_fname fnames
   # Note: TFW: Can only rename one file now, can make more intellegent
   # with substitutions and such for multiple files later.
   set fname [lindex $fnames 0]
   set newName [$widget get]
   set newfname [file join [file dirname $fname] "$newName[file extension $fname]"]
   if {[string equal $newfname $fname] || [string length newName] < 1} {
      return $fname
   }
   if {[catch {
      catalog::RenameFile $fname $newfname
      catalog::RenameEntry $fname $newfname
      set fnames [list $newfname]
   } result] \
   } then {
     tk_messageBox -icon warning -message "Could not rename\n$fname to \n$newfname\n$result\n$::errorInfo"
   }
   return $newfname
}
#------------------------------------------------------------------------------
# Callback used to save the width of each of the editors then close them
# Also removes global data associated with the window
#------------------------------------------------------------------------------
proc id3Editor::closeWindow {W tag} {
   set W [string trimleft $W :]
   settings::pushWidth $W $tag
   destroy $W
   unset -nocomplain ::$W
   unset -nocomplain ::$W.vars
}
#------------------------------------------------------------------------------
# Function   : postGenreNames
# Description: On pop-up, post the genres, if already built then post menu
#              Menu is shared for all buttons to save memory
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc id3Editor::postGenreNames {Menu _Element} {
   if {[$Menu index end] == "none" || [$Menu index end] == 0} {
      set genres [concat $id3Tag::v1Genres $::snackAmpSettings(customGenres)]
      foreach {genre} $genres {
         set first [string index $genre 0]
         lappend genreNames($first) $genre
      }
      $Menu delete 0 end
      foreach {Value} [lsort [array names genreNames]] {
         $Menu add cascade -label $Value -menu $Menu.f$Value
         menu $Menu.f$Value
         foreach {item} $genreNames($Value) {
           $Menu.f$Value add command -label "$item" \
            -command [list set $_Element $item]
         }
      }
   }
}

#------------------------------------------------------------------------------
# Function   : smartFileName
# Description: Return a smart file name based on available tag data
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc id3Editor::smartFileName {_Array} {
   upvar #0 $_Array Data
   set name ""
   set sp [string trim $::snackAmpSettings(splitChar)]
   set fname $::snackAmpSettings(smartName)
   #TFW: Substitute the fields with appropriate data
   # %A - %C - %N - %T for example
   foreach {field} [list Artist Album Track Title] {code} [list %A %C %N %T] {
      if {[hasData Data $field] && $Data($field) != "0"} {
         # Substitute tag with data if it exists
         if {$field=="Track"} {
            # Format tracks using two digits
            set Data($field) [id3Tag::cleanTrack $Data($field)]
         }
         set fname [string map "[list $code] [list $Data($field)]" $fname]
      } else {
         # Remove tag
      }
   }
   # Remove unused codes
   regsub -all  -- {%A|%C|%N|%T} $fname {} fname
   # Collapse spaces and other bad juju to single space
   while {[regsub -all  -- {  |/|\\|\?|:|\*} $fname { } fname]} {}
   set fname [string map {\" '} $fname]     ;#"
   # Collapse holes for unused params
   while {[regsub -all -- "( $sp)( $sp)+" $fname {\1} fname]} {}
   set fname [string trim $fname " $sp"]
   return $fname
}
#------------------------------------------------------------------------------
# Function   : cbSmartTagName
# Description: Fill in tag data based on a file name
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc id3Editor::cbSmartTagName {_fname _Array} {
   upvar $_Array Data
   upvar $_fname fname

   # TFW: Can be smarter about which entry is the track number
   set sp [string trim $::snackAmpSettings(splitChar)]
   set parts [split [file rootname [file tail [fixplus::cleanName $fname]]] $sp]
   switch -- [llength $parts] {
      1 {
         set Data(Title) [lindex [string trim $parts] 0]
      }
      2 {
         # Artist - Title
         # Track - Title
         foreach {First Title} $parts {
            if {[string is double $First]} {
               set Data(Track) [string trim $First]
            } else {
               set Data(Artist) [string trim $First]
            }
            set Data(Title)  [string trim $Title]
         }
      }
      3 {
         # Artist - Album - Title
         # Album - Track - Title
         # Track - Artist - Title
         foreach {Artist Album Title} $parts {
            if {[string is double $Album]} {
               set Data(Album) [string trim $Artist]
               set Data(Track) [string trim $Album]
               set Data(Title) [string trim $Title]
            } else {
               if {[string is double $Artist]} {
                  set Data(Track)  [string trim $Artist]
                  set Data(Artist) [string trim $Album]
                  set Data(Title)  [string trim $Title]
               } else {
                  # Default
                  set Data(Artist) [string trim $Artist]
                  set Data(Album)  [string trim $Album]
                  set Data(Title)  [string trim $Title]
               }
            }
         }
      }
      4 {
         foreach {Artist Album Track Title} $parts {
            set Data(Artist) [string trim $Artist]
            set Data(Album)  [string trim $Album]
            set Data(Title)  [string trim $Title]
            set Data(Track)  [string trim $Track]
         }
      }
      5 {
         # TBD: handle Artist AlbA - AlbB - Track - Title
         foreach {Artist Album xx Track Title} $parts {
            set Data(Artist) [string trim $Artist]
            set Data(Album)  "[string trim $Album], [string trim $xx]"
            set Data(Title)  [string trim $Title]
            set Data(Track)  [string trim $Track]
         }
      }
      default {}
   }
   if {[info exists Data(Track)]} {
      if {[string is double $Data(Track)]} {
         # Handle octal interpretation
         set Data(Track) [scan $Data(Track) %d]
      } else {
         set Data(Track) ""
      }
   }
   if {[info exists Data(Genre)] && $Data(Genre) == "Not Set" && [info exists lastGenre]} {
      set Data(Genre) $lastGenre
   }
   return
}
package provide id3Editor 1.0
