#------------------------------------------------------------------------------
# macros 1.0
#
# High level commands for SnackAmp
#
# Copyright (C) 2003 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#
#------------------------------------------------------------------------------
set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/macros.tcl,v 1.13 2004/07/12 23:43:30 wilkason Exp $
}
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# Macros Namespace
#/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
namespace eval macros {
   namespace export *
}
#------------------------------------------------------------------------------
# Function:  macros::PlayOffset
#
# Play a music file at some offset, likely bound to slider control
#------------------------------------------------------------------------------
proc macros::PlayOffset {{manual 0}} {
   Trace
   global TA LB snackAmpSettings
   # Make sure we have an open file
   set soundLength [soundControl::length]
   if {$soundLength > 0} {
      # Some files are fast and some are slow to respond to -start = 0
      if {$manual == 0} {
         set offset [expr {round(double($gui::trackPosition)/$gui::songDuration*$soundLength)}]
      } else {
         set offset $manual
      }
      if {[catch {
         soundControl::stop
         ;# Inverse active track
         LB::inverseVideoFile $TA(currentMF) 1
         # Workaround for bug on linux/ogg where file plays from last pos, not from -start
         if {$::tcl_platform(platform) == "unix"} {
            soundControl::sound [soundControl::fileName]
         }
         soundControl::play -start $offset -command [soundControl::getNextCommand $::snackAmpSettings(trackGap)]
      } result]} then {
         saLog "PlayOffset: $result"
         return
      }
      soundControl::Timer
   }
}
#------------------------------------------------------------------------------
# Function   : bumpScaling
# Description: Adjust the current scaling for the playing song for some amount
# Author     : Tom Wilkason
# Date       : 1/9/2002
#------------------------------------------------------------------------------
proc macros::bumpScaling {percent} {
   global TA
   set gain [snack::audio scaling]
   gainControl::add $TA(currentMF) [expr {$gain*(1.0+$percent)}]
   snack::audio scaling [expr {$gain*(1.0+$percent)}]
}
#------------------------------------------------------------------------------
# Function   : Assign
# Description: Routine to assign global hotkeys, called from hotkeys file
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::Assign {key Command} {
   if {[catch {bind .saTop $key $Command} result]} {
      saLog "FilterAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
   }
}
#------------------------------------------------------------------------------
# Function   : FilterAssign
# Description: Routine to assign filter box hotkeys, called from hotkeys file
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::FilterAssign {key Command} {
   if {[catch {bind $::FI $key $Command} result]} {
      saLog "FilterAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
   }
}
#------------------------------------------------------------------------------
# Function   : PlaylistAssign
# Description: Routine to assign play list hotkeys, called from hotkeys file
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::PlaylistAssign {key Command} {
   if {[catch {bind [LB::body] $key $Command} result]} {
      saLog "PlaylistAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
   }
}
#------------------------------------------------------------------------------
# Function   : APL_ListAssign
# Description: Routine to assign APL play list hotkeys, called from hotkeys file
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::APL_ListAssign {key Command} {
   if {[APLExists]} {
      set plb [$::PLB bodypath]
      if {[catch {bind $plb $key $Command} result]} {
         saLog "APL_ListAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
      }
   }
}
#------------------------------------------------------------------------------
# Function   : APL_TreeAssign
# Description: Routine to assign APL tree hotkeys, called from hotkeys file
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::APL_TreeAssign {key Command} {
   if {[APLExists]} {
      if {[catch {bind [winfo parent $::TREE] $key $Command} result]} {
         saLog "APL_TreeAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
      }
   }
}
#------------------------------------------------------------------------------
# Function   : APL_TreeAssign
# Description: Routine to assign APL tree hotkeys, called from hotkeys file
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::APL_TreeNodeAssign {key Command} {
   if {[APLExists]} {
      if {[catch {$::TREE bind x $key $Command} result]} {
         saLog "APL_TreeNodeAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
      }
   }
}
#------------------------------------------------------------------------------
# Function   : APLFilterAssign
# Description: Routine to assign APL filter box hotkeys, called from hotkeys file
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::APLFilterAssign {key Command} {
   if {[APLExists]} {
      if {[catch {bind $::SAFind $key $Command} result]} {
         saLog "APLFilterAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
      }
   }
}

#------------------------------------------------------------------------------
# Function   : editPLID3
# Description: Edit the highligted files ID3 info
# Author     : Tom Wilkason
# Date       : 3/18/2002
#------------------------------------------------------------------------------
proc macros::SelectFile {{playing 0}} {
   if {!([winfo exists .wFC] && [wm state .wFC] == "normal")} {
      AutoPlaylist
   }
   if {$playing} {
      APL::SelectFile [soundControl::fileName]
   } else {
      APL::SelectFile [LB::nameActive]
   }
}

#------------------------------------------------------------------------------
# Function: macros::EditPLID3
#
# Invoke the id3 editor (in group mode) on the selected playlist file
#------------------------------------------------------------------------------
proc macros::EditPLID3 {args} {
   package require id3Editor
   set files [LB::nameSelected 1]
   if {[llength $files] > $::snackAmpSettings(ID3editLimit)} {
      tk_messageBox -type ok -icon warning \
         -message "Limiting operation to first $::snackAmpSettings(ID3editLimit) files"
   }
   id3Editor::id3Editor [lrange $files 0 $::snackAmpSettings(ID3editLimit)]
}

#------------------------------------------------------------------------------
# Function: macros::PLNameToTag
#
# Invoke the name to tag editor on the playlist selected files
#------------------------------------------------------------------------------
proc macros::PLNameToTag {smart} {
   package require id3Editor
   id3Editor::id3NameToTag [LB::nameSelected 1] $smart
}

#------------------------------------------------------------------------------
# Function: macros::PLTagToName
#
# Invoke the tag to name editor on the playlist selected files
#------------------------------------------------------------------------------
proc macros::PLTagToName {args} {
   package require id3Editor
   id3Editor::id3TagToName [LB::nameSelected 1]
}

#------------------------------------------------------------------------------
# Function: macros::removeTag
#
# Remove an ID3 tag (not currently implemented)
#------------------------------------------------------------------------------
proc macros::removeTag {args} {
   id3Tag::removeTag [LB::nameSelected 1]
}

#------------------------------------------------------------------------------
# Function: macros::EditPlaying
#
# Edit the tag info on the currently playing track
#------------------------------------------------------------------------------
proc macros::EditPlaying {args} {
   package require id3Editor
   id3Editor::id3Editor [list $::TA(currentMF)]
}
#------------------------------------------------------------------------------
# Function   : macros::Highlight
# Description: macros::Highlight one and only one selection
# Author     : Tom Wilkason
# Date       : 12/27/2000
#------------------------------------------------------------------------------
proc macros::Highlight {i {inverse 0}} {
   global TA
   LB::highlightIndex $i $inverse
   if {[APLExists] && $::snackAmpSettings(syncAPL)} {
      set TA(entryName) [LB::nameFromIdx $i 0]
      APL::SelectFile $TA(entryName)
   }
}
##
# Move the highlight up or down
#
proc macros::MoveTo {how} {
   global TA
   set i [lindex [LB::idxSelected] 0]
   if {![string is integer -strict $i]} {
      set i 0
   }
   switch -- $how {
      "down" {
         incr i
      }
      "up" {
         incr i -1
      }
      default {}
   }
   macros::Highlight $i
}
#------------------------------------------------------------------------------
# Function   : Swap
# Description: Move a selection up/down the list (bind keys to this)
# Author     : Tom Wilkason
# Date       : 10/29/2001
#------------------------------------------------------------------------------
proc macros::Swap {how} {
   global LB
   if {[llength [LB::idxSelected]] > 0} {
      set selected [lsort -integer [LB::idxSelected]]
      set first [lindex $selected 0]
      set last [lindex $selected end]
      if {$how=="up"} {
         ##
         # Move the item above the selection below the selection
         #
         set moveIndex [expr {$first-1}]
         if {$moveIndex >= 0} {
            LB::moveByIndex $moveIndex $last
         }
      } else {
         ##
         # Move the item below the selection above the selection
         #
         set moveIndex [expr {$last+1}]
         if {$moveIndex < [LB::idxEnd]} {
            LB::moveByIndex $moveIndex $first
         }
      }
   }
}
#------------------------------------------------------------------------------
# Function   : PlayGain/getPlayGain
# Description: Provide x2 gain control to better match audio experience
# Author     : Tom Wilkason
# Date       : 10/19/2001
#------------------------------------------------------------------------------
proc macros::PlayGain {value} {
   set newValue [expr {round(pow($value,2)/40.0)}]
   set ::TA(line,$::snackAmpSettings(lineSelect)) $newValue
}

#------------------------------------------------------------------------------
# Function   : macros::getPlayGain
# Description: Return the current gain setting log scaled
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::getPlayGain {args} {
   if {[catch {
      set audioGain $::TA(line,$::snackAmpSettings(lineSelect))
      set gain [expr {pow($audioGain*40.,0.5)}]
   } result] } then {
      saLog $result
      set gain 1.0
   }
   return $gain
}
#------------------------------------------------------------------------------
# Function   : bumpLevel
# Description: Adjust the current play level by some amount
# Author     : Tom Wilkason
# Date       : 11/9/2001
#------------------------------------------------------------------------------
proc macros::bumpLevel {factor} {
   set ::TA(gain) [expr {$::TA(gain)+$factor}]
   PlayGain $::TA(gain)
}
#------------------------------------------------------------------------------
# Function   : Console
# Description: Toggle the console for debugging purposes (WIN32 only)
# Author     : Tom Wilkason
# Date       : 12/27/2000
#------------------------------------------------------------------------------
proc macros::Console {args} {
   global TA LB
   if {$TA(Console)==0} {
      console hide
   } else {
      console show
   }
}
#------------------------------------------------------------------------------
# Function   : applyCurrentAPL
# Description: Apply the latest Auto Playlist settings, ignoring any filtering
#              in effect.
# Author     : Tom Wilkason
# Date       : 11/8/2001
#------------------------------------------------------------------------------
proc macros::applyCurrentAPL {{resort 1}} {
   Trace
   # Re-sort if required, otherwise use the existing sorting order
   soundControl::latency long
   Wait 1
   if {$resort} {
      # Cut currently playing entry
      filtering::createMasterList
      Update
      # position it at top
   }
   ##
   # Change this to a specific function to apply an APL
   #
   waitWindow::show "Applying Auto Playlist..."
   catalog::applyAPL ;# right now, a sort is happening in poplist
   waitWindow::close
   Wait 0
   soundControl::latency
}
#------------------------------------------------------------------------------
# Function   : RenameFileIP
# Description: Rename a file directly from the listbox
# Author     : Tom Wilkason
# Date       : 6/19/2002
#------------------------------------------------------------------------------
proc macros::RenameFile {{fullName {}}} {
   set W $::LB
   if {[string length $fullName] == 0} {
      set fullName [LB::nameActive]    ;# lookup long name
   }
   if {[string length $fullName]==0} {
      return
   } elseif {![file exists $fullName]} {
      tk_messageBox -icon warning -message "Could not rename '$fullName'"
   }
   set oldName [file tail $fullName]
   set dirName [file dirname $fullName]
   set newName $oldName
   ##
   # Block on dialog
   #
   set name [catalog::renameInPlaceDialog $W 0 $oldName]
   ##
   # Back from window closure
   #
   if {[string length $::TA(rename)]>0 && $::TA(rename) != $oldName} {
      set newFullName [file join $dirName $::TA(rename)]
      if {[catch {catalog::RenameFile $fullName $newFullName} result]} {
         tk_messageBox -icon warning -message "Could not rename\n$fullName to \n$newFullName\n$result"
      } else {
         catalog::RenameEntry $fullName $newFullName
      }
   }
}
#------------------------------------------------------------------------------
# Function   : macros::RenameFile
# Description: Rename a file using a dialog, called from key binding or menu
# Author     : Tom Wilkason
# Date       : 10/30/2001
#------------------------------------------------------------------------------
proc macros::RenameFileDialog {{fullName {}}} {
   if {[string length $fullName] == 0} {
      set fullName [LB::nameActive]    ;# lookup long name
   }
   set oldName [file tail $fullName]
   set dirName [file dirname $fullName]
   set newName $oldName

   set answer [getEntryDialog "$oldName" newName "Rename File"]
   if {$answer == "ok" && $newName != $oldName} {

      set newFullName [file join $dirName $newName]
      if {[catch {catalog::RenameFile $fullName $newFullName} result]} {
         tk_messageBox -icon warning -message "Could not rename\n$fullName to \n$newFullName\n$result"
      } else {
         catalog::RenameEntry $fullName $newFullName   ;# update full sorted list
      }
   }
}

#------------------------------------------------------------------------------
# Function   : SaveFolderFile
# Description: Save the current folder settings to a file
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::SaveFolderFile {{fName ""}} {
   global TA LB snackAmpSettings
   ##
   # If a list for GlobData is passed in, the second item should be the name to save as (if different)
   #
   if {$fName == ""} {
      set types [list [list "Auto-Playlists Files" .tafs] [list "All Files" "*"]]
      set fName [tk_getSaveFile -title "Save Auto-Playlists File" -filetypes $types \
         -initialdir $::HOME -defaultextension .tafs]
      if { $fName == "" } {
         return  ""
      }
   }
   # Force extension in case dialog lost it (as on linux)
   set fName "[file rootname $fName].tafs"
   if {[catch {open $fName w} InFile]} {
      set result [tk_messageBox -type ok -title "SnackAmp - Please Respond" -icon warning \
         -message "Save Error: Unable to open '$fName' for writing"]
      return ""
   }
   fconfigure $InFile -buffersize 100000

   set snackAmpSettings(lastFolderFile) $fName
   set TA(AutoPlayListSelection) [file rootname [file tail $fName]]
   if {$InFile!=0} {
      foreach {File} [aplFolder::foldersMatchingPattern * 1] {
         puts $InFile [list $File [aplFolder::Get $File]]
      }
      close $InFile
      if {[APLExists]} {
         Tree::setDirty $::TREE 0
      }
      set snackAmpSettings(lastFolderFile) $fName
      return $fName
   } else {
      return -1
   }
}

#------------------------------------------------------------------------------
# Function   : removeAPLFile
# Description: Remove an APL file, called from menu callback
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::removeAPLFile {fName} {
   set result [tk_messageBox -type yesno -title "Please Answer" \
      -icon question -message "Are you sure you want to remove APL setting '[file rootname [file tail $fName]]'?"]
   if {[string equal $result "yes"]} {
      if {[catch {file delete -force $fName} result] } then {
         tk_messageBox -type ok -title "Error" \
            -icon warning -message "Failed to remove '[file rootname [file tail $fName]]'\n$result"
      }

   }
}
#------------------------------------------------------------------------------
# Function   : removePLFile
# Description: Remove an playlist file, called from menu callback
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::removePLFile {fName} {
   set result [tk_messageBox -type yesno -title "Please Answer" \
      -icon question -message "Are you sure you want to playlist '[file rootname [file tail $fName]]'?"]
   if {[string equal $result "yes"]} {
      if {[catch {file delete -force $fName} result] } then {
         tk_messageBox -type ok -title "Error" \
            -icon warning -message "Failed to remove '[file rootname [file tail $fName]]'\n$result"
      }

   }
}
#------------------------------------------------------------------------------
# Function   : readAPLFile
# Description: Read a file and change the favorate folders
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::readAPLFile {fName {apply 0}} {
   Trace
   global TA snackAmpSettings
   Wait 1
   ;# Build default APL folders if not yet done
   # Save a corresponding playlist file
   set lf $::snackAmpSettings(lastFolderFile)
   if {[file exists $fName]} {
      if {[aplFolder::entries]==0} {
         aplFolder::createTeFolders 0
      }
      ;# Can we remove the option below?
      ##
      # TBD: Move this code below to aplFolder namespace
      #
      if {$::snackAmpSettings(savePlonApply) && [file exists $lf] && $lf!=$fName} {
         set plfile "[file rootname $lf].m3u"
         file delete -force $plfile
         SavePlaylistFile $::LB replace all $plfile
         Update
      }
      if {[catch {
         set fid [open $fName r]
      } result] \
      } then {
         tk_messageBox -icon warning -type ok -title "Open Error" -message $result
         return
      }
      fconfigure $fid -buffersize 100000
      set full [split [read $fid] \n]
      close $fid
      # Invalidate all current settings
      aplFolder::SetSome * -1
      # Only update values if an entry already exists from prior scan
      # This prevents garbage if directories were moved
      foreach line $full {
         foreach {Index Value} $line {
            aplFolder::SetIfExists $Index $Value
         }
      }
      Update
      ##
      # Some folders will still be tagged with -1 from
      # above invalidation, meaning they weren't in the new autoplaylist
      # file. If so, then set them to the value of their parent if possible.
      # Also, set any parents to a 2 if they have children with checks.
      #
      foreach {Index} [aplFolder::foldersMatchingPattern * 1] {
         aplFolder::setTeDefault $Index
      }
      if {[winfo exists .wFC]} {
         APL::refreshTree
      }
      ;# See if we need to apply it also
      if {$apply} {
         applyCurrentAPL 0
         Wait 1
      }
   } else {
      # If no default file create one with everything one
      ##
      # Create the default Folder list based on files
      # New ones are turned off by default.
      # We should already have the teFolders set to some value
      #
      aplFolder::SetSome * 1
      SaveFolderFile $fName
      applyCurrentAPL 0
      Wait 1
      waitWindow::close
      set result [tk_messageBox -type yesno -title "SnackAmp - Please Respond" -icon question \
         -message "All of your Auto-playlist folders containing music are set true \nby default, would you like to change them now?"]
      if {$result == "yes"} {
         AutoPlaylist
      }

   }
   set snackAmpSettings(lastFolderFile) $fName
   if {[APLExists]} {
      Tree::setDirty $::TREE 0
   }
   set TA(AutoPlayListSelection) [file rootname [file tail $fName]]
   Wait 0
}
# trace variable teFolders w ttttt
# proc ttttt {args} {
#    Trace
# }
#------------------------------------------------------------------------------
# Function   : saveCurrentAPL
# Description: Save and then apply the current Auto-playlist
# Author     : Tom Wilkason
# Date       : 11/8/2001
#------------------------------------------------------------------------------
proc macros::saveCurrentAPL {{apply 1}} {
   global snackAmpSettings
   SaveFolderFile $snackAmpSettings(lastFolderFile)
   if {$apply} {
      applyCurrentAPL 0
   }
   if {[APLExists]} {
      Tree::setDirty $::TREE 0
   }
}
#------------------------------------------------------------------------------
# Function   : Clear
# Description: Clear the playlist
# Author     : Tom Wilkason
# Date       : 12/27/2000
#------------------------------------------------------------------------------
proc macros::Clear {args} {
   if {[tk_messageBox -icon question -type yesno -title "Please Respond" \
      -message "Are you sure you want to clear the playlist"]=="yes"} {
      global TA LB
      LB::deleteAll
   }
}
#------------------------------------------------------------------------------
# Function   : OpenFile
# Description: Open a single file for playing, add it to play list and play it
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::OpenFile {args} {
   global TA LB
   if {[catch {
      # This isn't in earlier snack versions. Also needs
      # Tcl 8.43+ to work, therefore the catch.
      set files [snack::getOpenFile -format MP3 -multiple 1]
   } result] } then {
      set files [snack::getOpenFile -format MP3]
   }
   # Here is a funky way to tell if a list of files was returned
   # Look at the string rep.
   if {![string match "\{*" $files]} {
      LB::InsertPlay $files
   } elseif {[llength $files] > 0} {
      LB::userInsertList $files
   }
}
#------------------------------------------------------------------------------
# Open up a URL, save a shortcut and play it
#------------------------------------------------------------------------------
proc macros::OpenURL {args} {
   if {[playurl::checkuskfolder]} {
      set shortcut [playurl::addPromptUrl $::snackAmpSettings(uskDir)]
      LB::InsertPlay $shortcut
   }
}
#------------------------------------------------------------------------------
# Function   : macros::Append
# Description: Wrapper for LB::AppendFile for macros & registry
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::Append {file} {
   LB::AppendFile $file
}
#------------------------------------------------------------------------------
# Function   : macros::PlayResume
# Description: Play button binding, resume play if paused otherwise
#              play active
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::PlayResume {args} {
   switch -- [soundControl::playState] {
      "Playing" {
         # Restart
         Play
      }
      "Paused" {
         # Resume play
         Pause
      }
      "Stopped" {
         # Play
         Play
      }
   }
}
#------------------------------------------------------------------------------
# Function   : macros::skipPlay
# Description: Skip some number of seconds in the playing track
# Author     : Tom Wilkason
# Date       : 8/9/2003
#------------------------------------------------------------------------------
proc macros::skipPlay {amount} {
   set elapsed [soundControl::position]
   set ::TA(setdrag) 1
   set newTime [expr {$elapsed+$amount}]
   if {$newTime < 0} {
      set newTime 0
   }
   gui::setTime $newTime
   PlayOffset
}
#------------------------------------------------------------------------------
# Function   : Play
# Description: Play a list item. Bound to Play key but used elsewhere.
#              Updates the GUI with the song info. Allows an offset within
#              the song to be passed in.
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::Play {{What active} {offset -1} {pause 0}} {
   Trace
   global TA LB snackAmpSettings
   ##
   # Determine if the current active index
   # is a valid song. If so play it. Otherwise, flag it and jump to the next one.
   #
   if {$snackAmpSettings(NextAfterCurrnt)} {
      set i [LB::idxPlaying]
      if {$i < 0} {
         set i [LB::idxActive]
      }
   } else {
      set i [LB::idxActive]
   }
   ##
   # Based on prev, current or next position cursor to proper
   # file
   #
   switch -- $What {
      current {
         Stop
         if {$i == "" || $i<0} {
            set i 0
         }
      }
      active {
         # Always use active (e.g double click)
         Stop
         set i [LB::idxActive]
         if {$i == "" || $i<0} {
            set i 0
         }
      }
      next {
         # Only called, on automatic transition to next song
         # Can be called just before other song complete.
         # Don't issue a stop to keep the snack queue flowing.
         set soundControl::playState "Stopped"
         killAfter soundControl::Timer
         gui::setTime 0
         if {$i == "" || $i<0} {
            set i 0
         } else {
            incr i
         }
         # check for continuous play
         if {$i >= [LB::idxEnd]} {
            if {$snackAmpSettings(ContinuousPlay)} {
               set i 0
               # Resort on a wrap if some form of random play
               filtering::sortPlaylist
            } else {
               return
            }
         }
      }
      prev {
         if {$i == ""} {
            set i 0
         } elseif {$i==0} {
            set i [LB::idxEnd]
         } else {
            if {$i > 0} {
               incr i -1
            }
         }
      }
      default {
         saLog "Invalid subcommand :Play $What"
      }
   }
   ##
   # Get the track name at this index, translate if possible
   #
   set SelectionName [LB::nameFromIdx $i 1]
   set TA(currentMF) $SelectionName

   ##
   # If file does not exist then assume it needs to be mounted
   #
   if {[string length $SelectionName] == 0} {
      # macros::Highlight this file then skip to the next one
      # if we haven't been commanded to stop
      saLog "Play: Could not validate '[LB::nameFromIdx $i]' as valid file"
      set ::songArray(Title) "'[LB::nameFromIdx $i]' invalid"
      set ::songArray(Artist) "File not verifiable"
      set ::songArray(Album) "Error"
      LB::activate $i
      set LB::indexPlaying $i
      LB::invalidateIndex $i
      if {$::TA(STOP) == 0} {
         after 1000 Next
         # Next will clear background so delay
         after 1500 [list LB::invalidateIndex $i]
      }
      return
   }
   # File exists, inverse video it showing it is playing
   # large playlists are slow to inverse, so delay the inverse
   # till after track is playing
   set LB::indexPlaying $i
   after idle [list LB::inverseVideoIndex $i]
   ##
   # See if user changed the slider to adjust where to play
   # will reset any fading that happened
   #
   PlayGain $TA(gain)
   ##
   # If this is a url to play, then play it directly
   #
   if {[playurl::isusk $SelectionName]} {
      soundControl::reset url
      foreach {index} [array names ::songArray] {
         set ::songArray($index) ""
      }
      # TODO: should we read from the database instead?
      id3Tag::id3Label $SelectionName ::songArray $::snackAmpSettings(ID3ReadOrder)

      set ::songArray(Title) "Connecting ... $::songArray(Title)"
      set ::songArray(Album) "$SelectionName"
      if {[catch {playurl::playusk [soundControl::sound] $SelectionName [soundControl::getNextCommand] ::songArray} result] } then {
         saLog $::errorInfo
         set ::songArray(Title) "Error: $result"
         after 200 [list eval [soundControl::getNextCommand]]
         return
      } else {
         soundControl::pauseRelief 0
         set max [expr {3600*24}]      ;# 24 hours
         set gui::songDuration $max
         $gui::tsScale configure -to $max
         balloon::define $::SA.panel.stat.l.dtTitle $SelectionName
         if {[winfo exists .miniPlayer]} {
            balloon::define .miniPlayer.tt.title $SelectionName
         }
         catch {$miniPlayer::mpSS configure -to $max}
         set songTitle [string map {// /} [file rootname [join [lrange [file split $TA(currentMF)] end-1 end] /]]]
         wm title $::SA "$songTitle"
         set TA(songTitle) $songTitle
         soundControl::Timer
      }
      return
   }
   ##
   # Process a file
   #
   ##
   # It seems faster on play -start to have the short file name and cd to that directory
   # Also, UNC names are really slow (on Win2K anyway)!
   # Strangly, network drive mappings are also fast (if not using UNC)
   #
   soundControl::reset file
   cd [file dirname $SelectionName]
   if {[catch {soundControl::fileName $SelectionName} result]} {
      tk_messageBox -icon warning -type ok -title "sound error" \
         -message "Error creating sound object using $SelectionName\n$result\n$::errorInfo"
      return
   }
   ##
   # Check for an invalid header
   #
   foreach {sndSamples sndRate x x x x sndExt} [soundControl::sinfo] {break}
   if {$sndSamples == -1 || $sndRate==0 || [lsearch -exact $::TA(upcaseExtensions) $sndExt] < 0} {
      soundControl::fileName ""
      saLog "Invalid file: $SelectionName"
      set ::songArray(Title) "Invalid header:$SelectionName"
      update
      if {$snackAmpSettings(AutoDelete)} {
         Delete
      }
      after 1000 [list eval [soundControl::getNextCommand]]
      return
   }
   # #
   # Get ID3 and detailed song info and display it, if ID3 not available
   # Use file name as song title
   catalog::trackTextData $SelectionName ::songArray
   #catalog::trackFullData $SelectionName ::songArray
   if {$::songArray(ID3)=={}} {
      set ::songArray(Title) [stringTitle [file rootname [file tail $TA(currentMF)]]]
   }
   # This should be use when applying an APL
   if {$What == "next"} {
      # Check for minimum rating before play
      set rating [ratingToNum [db::getVal $SelectionName Rating]]
      if {$rating > 0 && ($rating < [ratingToNum $::snackAmpSettings(minTxtRating)])} {
         set ::songArray(Title) "(Rating) Skipping $::songArray(Title)"
         saLog "(Rating) Skipping $SelectionName"
         if {$::TA(STOP)==0} {
            after idle Next
         }
         return
      }
      ##
      # Verify minimum bit rate
      #
      set rate [lindex $::songArray(Rate) 0]
      if {$rate < $::TA(minRate)} {
         set ::songArray(Title) "(Bitrate) Skipping $::songArray(Title)"
         saLog "(Bitrate) Skipping $SelectionName"
         if {$::TA(STOP)==0} {
            after idle Next
         }
         return
      }
   }
   ##
   # Adjust gain if the song has an auto level setting (and it is enabled)
   #
   if {$snackAmpSettings(autoLevel)} {

      set value [gainControl::createIfNeeded $SelectionName]
      ##
      # Pre compute the next track in the playlist if possible
      # Wait about 20 seconds to do so
      if {$snackAmpSettings(autoLevelPreComute)} {
         set nextSelection [LB::nameFromIdx [expr {$i+1}] 1]
         if {[string length $nextSelection]} {
            after 20000 [list gainControl::createIfNeeded $nextSelection]
         }
      }
      # Clamp db Gain at runtime
      if {[catch {toDb $value} db] } then {
         set db 0.0
      } else {
         if {$db > $::snackAmpSettings(maxGain)} {
            set db $::snackAmpSettings(maxGain)
         } elseif {$db < $::snackAmpSettings(minGain)} {
            set db $::snackAmpSettings(minGain)
         }
      }
      snack::audio scaling [fromDb $db]
   } else {
      snack::audio scaling 1.0
   }
   # Fires Next command at end of this file
   if {$offset < 0} {
      set offset [gainControl::getStart [soundControl::fileName]]
   }
   after 200 {gui::packCoverArt [soundControl::fileName]}
   soundControl::play \
      -start $offset \
      -command [soundControl::getNextCommandBump [soundControl::fileName] $::snackAmpSettings(trackGap)]
   if {$pause} {
      Pause
   }
   # Only log this if going to a file
   if {$::snackAmpSettings(logFile)} {
      saLog "Playing $TA(currentMF)"
   }
   ##
   # What ~10 ticks on 5 or multiple sec intervals
   #
   set trackTime  [soundControl::duration]
   $gui::tsScale configure -to $trackTime
   # Miniplayer (catch if not up)
   if {[info exists miniPlayer::mpSS] && [winfo exists $miniPlayer::mpSS]} {
      $miniPlayer::mpSS configure -to $trackTime
   }
   ##
   # Song Title is shown in the title bar
   #
   set songTitle [string map {// /} [file rootname [join [lrange [file split $TA(currentMF)] end-1 end] /]]]
   # If the folder name is the same as the first part of file name,
   # Then we can ignore the folder name
   if {[string match -nocase "[file dirname $songTitle]*" [file tail $songTitle]]} {
      set songTitle [file tail $songTitle]
   }
   wm title $::SA "$songTitle"
   set TA(songTitle) $songTitle

   set gui::songDuration $::songArray(Duration)
   if {$snackAmpSettings(autoSavePlaylist)} {
      catalog::SavePlaylistCache
   }
   soundControl::Timer
}
#----------------------------------------------------------------------------
# Play next sound file in list
#----------------------------------------------------------------------------
proc macros::Next {} {
   Trace
   Stop
   Play next
}
#----------------------------------------------------------------------------
# Play the next track right now
#----------------------------------------------------------------------------
proc macros::NextNow {} {
   Trace
   Stop
   Play next
}
#----------------------------------------------------------------------------
# Play previous sound file in list
#----------------------------------------------------------------------------
proc macros::Prev {} {
   Trace
   Stop
   after idle Play prev
}
#----------------------------------------------------------------------------
# Shuffle the selected files in the playlist
#----------------------------------------------------------------------------
proc macros::shuffleSelection {args} {
   LB::shuffleSelection ::LB
}
#----------------------------------------------------------------------------
# Sort the selected files in the playlist
#----------------------------------------------------------------------------
proc macros::sortSelection {args} {
   LB::sortSelection ::LB
}
#----------------------------------------------------------------------------
# Stop playback of a file and cancle the update timer
#----------------------------------------------------------------------------
proc macros::Stop {args} {
   Trace
   global TA LB
   killAfter soundControl::Timer
   if {[snack::audio active]} {
      soundControl::stop
      gui::setTime 0
   }
   soundControl::pauseRelief 2
   playurl::cleanupLast
}
#------------------------------------------------------------------------------
# Move the cursor to the currently playing file,
# assign to <Home> key
#------------------------------------------------------------------------------
proc macros::locatePlaying {args} {
   LB::seePlaying
}

#------------------------------------------------------------------------------
# Pause/resume playback
#------------------------------------------------------------------------------
proc macros::Pause {args} {
   Trace
   global TA LB
   if {[snack::audio active]} {
      if {$soundControl::paused==1} {
         soundControl::pause 0
         soundControl::Timer
      } else {
         killAfter soundControl::Timer
         soundControl::pause 1
      }
   }
}
#------------------------------------------------------------------------------
# Callback to remove the current selections from the display
#------------------------------------------------------------------------------
proc macros::Cut {{playNext 0}} {
   Trace
   global TA LB
   set playActive 0
   set playing [LB::idxPlaying]
   if {[llength [LB::idxSelected]] > 0} {
      set Index [LB::idxAnchor]
      foreach {ind} [lsort -decreasing -integer [LB::idxSelected]] {
         if {$playing==$ind} {
            set playActive [soundControl::isPlaying]
            if {$playActive} {
               Stop
            }
         }
         LB::deleteIndex $ind
      }
      # Stay on 0 if that is what we were on
      ##macros::Highlight [expr {$Index-0}]
      gui::updateFileDisplay
      # Set selection to current active
      LB::setSelection [LB::idxActive]
      # This will play the next song
      if {$playNext && $playActive} {
         Play
      }
   }
}
#------------------------------------------------------------------------------
# Cut the listbox to the clibboard
# (only used for main playlist)
#------------------------------------------------------------------------------
proc macros::clipboardCut {lb {append 1}} {
   Trace
   global TA
   if {[llength [$lb curselection]] > 0} {
      clipboardCopy $lb $append
      Cut 1
      gui::updateFileDisplay
   }
}
#------------------------------------------------------------------------------
# Copy the listbox files to the clipboard
#------------------------------------------------------------------------------
proc macros::clipboardCopy {lb {append 1}} {
   Trace
   global TA LB
   if {[llength [$lb curselection]] > 0} {
      if {$append == 0} {
         clipboard clear
      } else {
         clipboard append " "
      }
      set data [getSelectedTracks $lb]
      clipboard append $data
   }
}
#------------------------------------------------------------------------------
# Paste the listbox files to the playlist
# Note: Only works with LB playlist
#------------------------------------------------------------------------------
proc macros::clipboardPaste {lb {clear 1}} {
   Trace
   global TA
   if {[catch {selection get -selection CLIPBOARD} files]} {
      return
   }
   set current [$lb index anchor]
   set processed [list]
   foreach {file} $files {
      # Only insert files that are in the master list
      # And not already processed (i.e. no duplicates)
      #W: move to shared
      if {[lsearch -exact $processed $file] < 0 && [db::whereInDb $file] >= 0} {
         lappend processed $file
      }
   }
   if {[llength $processed]>0} {
      LB::insertList [LB::idxActive] $processed
   }
   if {$clear} {
      clipboard clear
   }
}
#------------------------------------------------------------------------------
# Callback to remove the current selections from the disk
#------------------------------------------------------------------------------
proc macros::Delete {{Next 1}} {
   Trace
   global TA LB
   set goNext 0
   if {[llength [LB::idxSelected]] > 0} {
      if {$::snackAmpSettings(dontMoveFiles)} {
         tk_messageBox -icon warning -type ok \
            -message "You don't have permission to remove files"
         return $goNext
      }
      set Index [LB::idxAnchor]

      set selection [LB::idxSelected]
      foreach {ind} $selection {
         lappend Files [LB::keyFromIndex $ind]
      }

      set result [tk_messageBox -type yesno -title "Please Answer" \
         -icon question -message "Are you sure you want to delete:\n[join $Files \n]?"]
      if {$result == "yes"} {
         foreach {Entry} $Files {
            set File [LB::keyToName $Entry]
            set goNext [catalog::DeleteFile $File]
            Cut $goNext
         }
         ##macros::Highlight $Index
      }
   }
   ##
   # Play the next song if we deleted the current one and we want to
   #
   if {$Next && $goNext} {
      Play
   }
   gui::updateFileDisplay
   return $goNext
}
#----------------------------------------------------------------------------
# Bring up or down the Auto Playlist window, depending on its state
#----------------------------------------------------------------------------

proc macros::AutoPlaylist {args} {
   if {![winfo exists .wFC]} {
      package require saTree
      package require autoPlayList
      $::TA(treeButton) configure -relief sunken
      # First make sure we have the proper roots for the tree viewer
      APL::FolderTree
      update  ;# let it draw
      # May not have a selection
      catch {APL::SelectFile $::TA(currentMF)}
   } elseif {[wm state .wFC] != "normal"} {
      $::TA(treeButton) configure -relief sunken
      APL::FolderTree
      update  ;# let it draw
      catch {APL::SelectFile $::TA(currentMF)}
   } else {
      wm withdraw .wFC
      $::TA(treeButton) configure -relief flat
   }
}
#------------------------------------------------------------------------------
# Function   : macros::findNextSong
# Description: Locate the next occurance of a song in the playlist based on the
#              current filter value
# Author     : Tom Wilkason
# Date       : 11/08/2001
#------------------------------------------------------------------------------
proc macros::findNextSong {{top 0}} {
   global TA snackAmpSettings
   # Get the current playlist
   set regPattern [list]                                ;#retain wildcards
   set pattern [split [string map {" or " "|" " OR " "|" "*" ".*" "?" "."} [catalog::globEscape $TA(filter)]] |]
   foreach part $pattern {
      lappend regPattern "($part)+"
   }
   set regPattern [join $regPattern |]
   if {$top} {
      set i -1
   } else {
      set i [lindex [LB::idxSelected] 0]
      if {$i=={}} {
         set i -1
      }
   }
   set ok 2
   ##
   # Allow one wrap around if at first not found
   #
   if {[catch {
      while {$ok} {
         incr i
         foreach {Entry} [LB::keyFromIndex $i end] {
            if {[regexp -nocase -- $regPattern [LB::keyToName $Entry]]} {
               macros::Highlight $i
               return
            }
            incr i
         }
         set i -1
         incr ok -1
      }
   } result]} {
      #saLog "findNextSong $result"
   }
}
#------------------------------------------------------------------------------
# Function   : macros::playSpecificSong
# Description: Locate and Play a specific song in the playist, ignore if not in playlist
# Author     : Tom Wilkason
# Date       : 4/28/2002
#------------------------------------------------------------------------------
proc macros::playSpecificSong {song} {
   Trace
   set i 0
   set i [lsearch -exact [LB::keyOrNamesAll] $song]
   if {$i >= 0} {
      LB::activate $i
      Play active
   }
}
#------------------------------------------------------------------------------
# Function   : macros::clearFilter
# Description: Clear the current filter entry box
# Author     : Tom Wilkason
# Date       : 11/9/2001
#------------------------------------------------------------------------------
proc macros::clearFilter {args} {
   set ::TA(filter) {}
}

#------------------------------------------------------------------------------
# Function: macros::selectAll
#
# Select all the files in the playlist
#------------------------------------------------------------------------------
proc macros::selectAll {args} {
   LB::setSelection 0 end
}
#------------------------------------------------------------------------------
# Function   : macros::filterMusic
# Description: Restrict files shown based on the filter & folder settings
# Author     : Tom Wilkason
# Date       : 12/26/2000
#------------------------------------------------------------------------------
proc macros::filterMusic {how {what def}} {
   filtering::filterMusic $how $what
}
#------------------------------------------------------------------------------
# Return selected tracks
#------------------------------------------------------------------------------
proc macros::selectedTracks {ListBox {what "selected"}} {
   set names [list]
   if {$ListBox == $::LB} {
      if {$what == "selected"} {
         set Indicies [LB::idxSelected]
      } else {
         set Indicies [LB::idxAll]
      }
      foreach {Index} $Indicies {
         lappend names [LB::keyToName [LB::keyFromIndex $Index]]
      }
   } elseif {$ListBox == $::PLB} {
      if {$what == "selected"} {
         set names [aplLB::nameSelected 0]
      } else {
         set names [aplLB::getAllNames 1]
      }
   } else {
      saLog "unknown listbox '$ListBox'"
   }
   return $names
}
#----------------------------------------------------------------------------
# Copy all or selected tracks to an attached media player
# TODO: Modify to pass in the names and add functions to return
# the appropriate names from the menu level
#----------------------------------------------------------------------------
proc macros::copyToMediaPlayer {mp nameFunction} {
   Trace
   Wait 1
   set names [eval $nameFunction]
   set mydirs [list]
   foreach {dir} $::snackAmpSettings(mediaDirs) {
      lappend mydirs $dir [file join $dir *] [expr {[string length $dir]+1}]
   }
   set todo [llength $names]
   set counter 0
   waitWindow::show "Copying $todo Files to Media Player..." 1
   foreach {name} $names {
      incr counter
      set dest ""
      # For each possible dest pattern, check
      foreach {dir patt len} $mydirs {
         if {[string match $patt $name]} {
            set dest [file join $mp [string range $name $len end]]
               if {[file exists $name]} {
               set smtime [file mtime $name]
               if {[file exists $dest]} {
                  set dmtime [file mtime $dest]
               } else {
                  set dmtime [expr {$smtime-1}]
               }
               # Check the file date for existing files
               if {$smtime > $dmtime} {
                  catch {file mkdir [file dirname $dest]}
                  waitWindow::percent $counter $todo "Copying\n$name" 1
                  if {[catch {file copy -force $name $dest} result] } then {
                     saLog $result
                  } else {
                     catalog::addNewEntry $dest
                  }
                  if {[waitWindow::ifbreak]} {
                     saLog "Aborting Copy..."
                     Wait 0
                     waitWindow::close
                     return
                  }
               }
            }
         }
      }
   }
   Wait 0
   waitWindow::close
}
#------------------------------------------------------------------------------
# Function   : macros::SavePlaylistFile
# Description: Save current files in list
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::SavePlaylistFile {ListBox {how "replace"} {what "selected"} {fName {}}} {
   Trace
   global TA LB
   ##
   # If a list for GlobData is passed in, the second item should be the name to save as (if different)
   #
   if {$how=="replace"} {
      set access "w"
   } else {
      set access "a+"
   }
   # Use different dialog depending on what you want to do
   array set tk_dialog {
      replace tk_getSaveFile
      append tk_getOpenFile
   }
   if {$what == "selected"} {
      set Indicies [LB::idxSelected]
   } else {
      set Indicies [LB::idxAll]
   }
   if {[string length $fName]==0} {
      set types [list [list "Playlist files" {.m3u .taf}] [list "All Files" "*"]]

      set fName [$tk_dialog($how) -title "Save [llength $Indicies] tracks to Playlist File" -filetypes $types \
         -initialdir $::HOME -defaultextension .m3u]
      if { $fName == "" } {
         return  ""
      }
      # Set default extension to m3u
      if {[file extension $fName]==""} {
         set fName "$fName.m3u"
      }
   } elseif {[file exists $fName]} {
      if {$how=="replace" && [tk_messageBox -icon question -type yesno -title "Overwrite File" -message "Is it ok to $how $fName with [llength $Indicies] tracks?"]!="yes"} {
         return {}
      }
   }
   if {[catch {open $fName $access} InFile]} {
      tk_messageBox -icon warning -type ok -title "File Error" -message "Save Error: Unable to open '$fName' for writing"
      return {}
   }
   set TA(PlayListSelection) [file rootname [file tail $fName]]
   set TA(PlayListFile) $fName
   set ext [file extension $fName]
   if {[string equal $ext .m3u]} {
      puts $InFile "#\n# Playlist file: [file nativename $fName]"
      puts $InFile "# Saved on [clock format [clock seconds]] by user $::tcl_platform(user) on host [info hostname]\n#"
   }
   if {$InFile!=0} {
      ##
      # Process either selected or entire list
      #
      set names [list]
      set fldPatt [file join [file dirname $fName] *]
      set ml [string length [string trimright [file dirname $fName] / ]]
      incr ml
      if {$ListBox == $::LB} {
         foreach {Index} $Indicies {
            set File [LB::keyToName [LB::keyFromIndex $Index]]
            # Todo: Strip front part if it can be made relative to pwd
            if {[string equal $ext .m3u]} {
               # Make file relative if possible
               if {[string match $fldPatt $File]} {
                  set File [string range $File $ml end]
               }
               set File [file nativename $File]
            }
            puts $InFile $File
         }
      } elseif {$ListBox == $::PLB} {
         if {$what == "selected"} {
            set names [aplLB::nameSelected 0]
         } else {
            set names [aplLB::getAllNames 1]
         }
         foreach {File} $names {
            if {[string equal $ext .m3u]} {
               if {[string match $fldPatt $File]} {
                  set File [string range $File $ml end]
               }
               set File [file nativename $File]
            }
            puts $InFile $File
         }
      } else {
         saLog "unknown listbox '$ListBox'"
      }
      #W: Only save if playing is in list
      if {[lsearch -exact $names $::TA(currentMF)]>= 0} {
         puts $InFile "#Playing=$::TA(currentMF)"
      }
      close $InFile
      return $fName
   } else {
      return -1
   }
}
#------------------------------------------------------------------------------
# Function   : macros::ReadPlaylistFile
# Description: Verify playlist file and notify user of any problems
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::ReadPlaylistFile {fName {how "replace"}} {
   # If file doesn't exist, then prompt for it
   if {$fName == {} || ![file exists $fName]} {
      set types [list [list "Playlist files" {.m3u .taf}] [list "All Files" "*"]]
      set fName [tk_getOpenFile -title "Open Playlist File" -filetypes $types \
         -initialdir $::HOME]
      if { $fName == "" } {
         return  ""
      }
   }
   set failed [ProcessPlaylistFile $fName $how 1]
   set numFailed [llength $failed]
   if {$numFailed > 0} {
      ##
      # Tell user how many failed (if any)
      #
      if {$numFailed > 10} {
         set failed [lrange $failed 0 9]
         lappend failed "...."
      }
      tk_messageBox -icon warning -type ok -title "Invalid Files" \
         -message "A total of $numFailed files do no exist, and could not be added to the playlist\n[join $failed \n]"
   }
   return $numFailed
}
#------------------------------------------------------------------------------
# Function   : macros::ProcessPlaylistFile
# Description: Read a file and append/replace current list
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::ProcessPlaylistFile {fName {how "replace"} {resume 0}} {
   Trace
   global TA LB snackAmpSettings
   if {[catch {
      set fid [open $fName r]
   } result] \
   } then {
      tk_messageBox -icon warning -type ok -title "Open Error" -message $result
      return
   }
   Wait 1
   set full [split [string trim [read $fid]] \n]
   close $fid
   # Insert the files if they are not already in the list
   # Translate Removable media if necessary.
   # Also support winamp format playlists, this means no files can
   # start with # or else they will be ignored (that is a comment in m3u)
   set here [file dirname $fName]
   set push [pwd]
   cd $here
   set comment "\#*"
   set failed [list]
   set i 0
   set cp [LB::idxActive]
   set toinsert [list]
   set processLines [llength $full]
   foreach file $full {
      incr i
      if {! [string match $comment $file]} {
         # If a url, create a usk and process the usk file
         if {[string match -nocase "http://*" $file]} {
            # Create a usk (url) shortcut, then append it
            # Note: trying tail of file name, works if a file if a stream
            # may not work as well.
            set file [playurl::addUrl $::snackAmpSettings(uskDir) [file tail $file] $file]
         } else {
            # This will be ignored if abs path names
            set file [file join $here $file]
         }
         # Verify existence (can be slow on a net)
         if {[set rfile [catalog::TranslateMount $file]] != {} } {
            lappend toinsert $rfile
         } else {
            lappend failed $file
         }
         if {$i%50==0} {
            waitWindow::percent $i $processLines "Playlist Files Verified..."
         }   ;# let the music play
      } else {
         if {$resume && [string match "#Playing*" $file]} {
            set resumeFile [lindex [split $file =] 1]
            if {$::snackAmpSettings(restorePlaying) && [file exists $resumeFile]} {
               set ans [tk_messageBox -type yesno -icon question -message "Do you want to resume playing of \n$resumeFile"]
               if {$ans=="yes"} {
                  after 1000 [list PlayFile $resumeFile]
               }
            }
         }
      }
   }
   ;# Insert the tracks
   if {[llength $toinsert] > 0} {
      if {$how=="replace"} {
         LB::deleteAll
         LB::insertList 0 $toinsert
      } else {
         LB::insertList $cp $toinsert
      }
   }
   Wait 0
   waitWindow::close
   set TA(PlayListFile) $fName
   set TA(PlayListSelection) [file rootname [file tail $fName]]
   cd $push
   gui::updateFileDisplay
   return $failed
}
#------------------------------------------------------------------------------
# Function   : macros::Insert
# Description: Wrapper for LB::userInsertList for macros & registry
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::Insert {file} {
   LB::userInsertList [list $file]
}
#------------------------------------------------------------------------------
# Translate a list of files/folders into tracks, support playlists also
# if link is to a http site, read the url if it is a playlist and process it.
# TFW: Need support recursive calls for nested playlists/folders.
#------------------------------------------------------------------------------
proc macros::TranslateFiles {files} {
   Trace
   set toappend [list]
   foreach {file} $files {
      set file [string map {\\ /} $file]
      set ext [string tolower [file extension $file]]
      switch -- $ext {
         .m3u
         -
         .pls {
            # If coming from a web site, create a usk file,
            # otherwise, read in the playlist and process each line
            if {[string match -nocase "http://*" $file]} {
               playurl::checkuskfolder
               set contents [split [playurl::getUrlBody $file] \n]
            } else {
               if {[file readable $file]} {
                 set fid [open $file r]
                 set contents [split [string trim [read $fid]] \n]
                 close $fid
               } else {
                  tk_messageBox -icon warning -type ok -message "File '$file' cannot be read"
                  return $toappend
               }
            }
            ;# Handle playlist types uniquely
            set fileTail [file rootname [file tail $file]]
            switch -- $ext {
               .m3u {
                   set i 0
                   foreach {row} $contents {
                      set drow [HttpD::Decode $row]  ;# not sure if this is needed
                      if {![string match "\#*" $drow]} {
                         set fullFile [file join [file dirname $file] $drow]
                         if {[string match -nocase "http://*" $drow]} {
                            # Create a usk (url) shortcut, then append it
                            set name [file tail $drow]
                            # Note: trying tail of file name, works if a file if a stream
                            # may not work as well.
                            set fname [playurl::addUrl $::snackAmpSettings(uskDir) "$name" $row]
                            lappend toappend $fname
                            incr i
                         } elseif {[file exists $fullFile]} {
                            lappend toappend $fullFile
                         } else {
                            saLog "File:'$row' does not exist"
                         }
                      }
                   }
               }
               .pls {
                  set title $fileTail
                  set i 0
                  set Fnum 0
                  set Tnum 0
                  foreach {row} $contents {
                     if {[regexp -- {File([0-9]+)=(.*)} $row -> Fnum url]} {
                        set pls($Fnum,url) $url
                     }
                     if {[regexp -- {Title([0-9]+)=(.*)} $row -> Tnum title]} {
                        set title [string trim [file rootname $title] .]
                        set pls($Tnum,title) $title
                     }
                  }
                  foreach {index} [lsort -dictionary [array names pls *,url]] {
                     foreach {num type} [split $index ,] {break}
                     if {[info exists pls($num,title)]} {
                        set title $pls($num,title)
                     } else {
                        set title $fileTail.$num
                     }
                     lappend toappend [playurl::addUrl $::snackAmpSettings(uskDir) $title $pls($index)]
                  }
               }
            }
         }
         default {
            # can get in file:/sdfd%3fdf.mp3 type format, must convert
            if {[regexp -- {file:(.*)} $file -> url]} {
               set file [HttpD::Decode $url]
               puts "Using:'$file'"
            }
            # Should we recurse here?
            if {[file isdirectory $file]} {
               foreach {track} [glob -nocomplain -directory $file *] {
                  if {[catalog::isTrackFile $track]} {
                     lappend toappend $track
                  }
               }
            } elseif {[file readable $file] && [catalog::isTrackFile $file]} {
               # usk shortcut files also handled here
               lappend toappend $file
            } else {
               tk_messageBox -icon warning -type ok -message "File '$file' cannot be read"
            }
         }
      }
   }
   return $toappend
}
#------------------------------------------------------------------------------
# Copy a file to a different folder, used for images in files or url links
#------------------------------------------------------------------------------
proc macros::copyImageToFolder {url folderName} {
   #set fname [file join $folderName [file tail $url]]
   set fname [file join $folderName "folder.jpg"]
   if {[file exists $url]} {
      if {[file exists $fname]} {
         set resp [tk_messageBox -type yesno -icon warning -message "Is it ok to overwrite $fname"]
      } else {
         set resp yes
      }
      if {$resp=="yes"} {
         file copy -force $url $fname
      }
   } else {
      # httpget
      if {[file exists $fname]} {
         set resp [tk_messageBox -type yesno -icon warning -message "Is it ok to overwrite $fname"]
      } else {
         set resp yes
      }
      if {$resp=="yes"} {
         if {[catch {getUrlBinary $url} contents] } then {
            tk_messageBox -type ok -icon warning -message "Error retrieving $url\n$result"
            return
         }
         if {[string length $contents]} {
            if {[catch {open $fname w} fid] } then {
               saLog $fid
               return
            } else {
               fconfigure $fid -translation binary -encoding binary
               puts -nonewline $fid $contents
               close $fid
            }
         }
      }
   }
}
#------------------------------------------------------------------------------
# Function   : dndPlayFileList
# Description: Play a list of files, inserting all but the last one
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::dndPlayFileList {files} {
   Trace
   global TA
   set flag 0
   if {[llength $files]} {
      # Hack, if first file is a graphic, just replace the
      # folder image
      set ext [string tolower [file extension [lindex $files 0]]]
      switch -- $ext {
         .jpg -
         .png {
            copyImageToFolder [lindex $files 0] [file dirname $TA(currentMF)]
         }
         default {
            set tracks [TranslateFiles $files]
            set here [LB::idxPlaying]
            LB::insertList [incr here] $tracks
            LB::activate $here
            Play active
         }
      }
   }
   return 1
}
#----------------------------------------------------------------------------
# dnd drop target, insert incoming files at cursor
#----------------------------------------------------------------------------
proc macros::dndDropFileList {x y files} {
   Trace
   global LB  ;# hack
   Trace
   if {$x>0} {
      set new [$LB nearest $y]
   } else {
      set new end
   }
   LB::insertList $new [TranslateFiles $files]
   return 1
}
##
# Grab the contents of some URL and return it
# Used to grab binary data such as a jpeg
#
proc macros::getUrlBinary {url} {
   #direct to contents
   package require http
   set token [::http::geturl $url]
   set contents [::http::data $token]
   return $contents
}
#------------------------------------------------------------------------------
# Function   : PlayFile
# Description: Play a file at the top of the list (linked to DDE command on Windows)
#              Set this up early so a file association doesn't time out when launching
#              SnackAmp on a slow computer.
#              TODO: Use TranslateFiles to translate, this to insert and play
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc macros::PlayFile {file {justInsert 0}} {
   set file [string map {\\ /} $file]
   set ext [string tolower [file extension $file]]
   switch -- $ext {
      .m3u
      -
      .pls {
         # If coming from a web site, create a usk file,
         # otherwise, read in the playlist and process each line
         if {[string match -nocase "http://*" $file]} {
            playurl::checkuskfolder
            set contents [split [playurl::getUrlBody $file] \n]
         } else {
            if {[file readable $file]} {
              set fid [open $file r]
              set contents [split [string trim [read $fid]] \n]
              close $fid
            } else {
               tk_messageBox -icon warning -type ok -message "File '$file' cannot be read"
               return
            }
         }
         set fileTail [file rootname [file tail $file]]
         ;# Handle playlist types uniquely
         switch -- $ext {
            .m3u {
                set toappend [list]
                set i 0
                foreach {row} $contents {
                   set drow [HttpD::Decode $row]  ;# not sure if this is needed
                   if {![string match "\#*" $drow]} {
                      set fullFile [file join [file dirname $file] $drow]
                      if {[string match -nocase "http://*" $drow]} {
                         # Create a usk (url) shortcut, then append it
                         set name [file tail $drow]
                         # Note: trying tail of file name, works if a file if a stream
                         # may not work as well.
                         set fname [playurl::addUrl $::snackAmpSettings(uskDir) "$name" $row]
                         lappend toappend $fname
                         incr i
                      } elseif {[file exists $fullFile]} {
                         lappend toappend $fullFile
                      } else {
                         saLog "File:'$row' does not exist"
                      }
                   }
                }
                set added 0
                set end [LB::userInsert]
                ;# If we inserted some, and not just inserting, play the first one
                if {[LB::userInsertList $toappend] > 0 && ($justInsert==0) } {
                   LB::activate $end
                   Play active
                }
            }
            .pls {
               set title $fileTail
               set i 0
               set Fnum 0
               set Tnum 0
               foreach {row} $contents {
                  if {[regexp -- {File([0-9]+)=(.*)} $row -> Fnum url]} {
                     set pls($Fnum,url) $url
                  }
                  if {[regexp -- {Title([0-9]+)=(.*)} $row -> Tnum title]} {
                     set title [string trim [file rootname $title] .]
                     set pls($Tnum,title) $title
                  }
               }
               set i 1
               foreach {index} [lsort -dictionary [array names pls *,url]] {
                  foreach {num type} [split $index ,] {break}
                  if {[info exists pls($num,title)]} {
                     set title $pls($num,title)
                  } else {
                     set title $fileTail.$num
                  }
                  set fname [playurl::addUrl $::snackAmpSettings(uskDir) $title $pls($index)]
                  if {$i==1} {
                    set end [LB::userInsert]
                  }
                  LB::userInsert $fname
                  incr i
               }
               if {$justInsert==0 && [info exists end]} {
                  LB::activate $end
                  gui::updateFileDisplay
                  Play active
               }

            }
         }
      }
      default {
         # usk shortcut files also handled here
         if {[file readable $file]} {
            if {$justInsert==1} {
               LB::userInsertList [list $file]
            } else {
               LB::InsertPlay $file
            }
         } else {
            tk_messageBox -icon warning -type ok -message "File '$file' cannot be read"
         }
      }
   }
}
#------------------------------------------------------------------------------
# Function   : Exit
# Description: Exit gracefully
# Author     : Tom Wilkason
# Date       : 1/9/2001
#------------------------------------------------------------------------------
proc macros::taExit {args} {
   global TA snackAmpSettings
   catch {
      winico taskbar delete $TA(Icon)
   }
   #wm withdraw $::SA
   catch {APL::close}
   catalog::SavePlaylistCache
   # Stop after we save the cache
   soundControl::stop
   HttpD::Shutdown
   set snackAmpSettings(minWidth) [winfo width $::SA]
   foreach {w h x1 x2} [split [wm geometry $::SA] "x +"] {break}
   set snackAmpSettings(playerPos) [list $x1 $x2]
   if {[winfo exists .albumart]} {
      set snackAmpSettings(albumartPos) [wm geometry .albumart]
   }
   settings::saveSettings
   gainControl::saveSettings
   aplFolder::saveFolderNames
   db::Commit
   db::closeDB
   # close any remaining channels
   foreach {channel} [file channels] {
      catch {close $channel}
   }
   exit
}
#------------------------------------------------------------------------------
# Copy the selection to a staging file for burning to a CD
#------------------------------------------------------------------------------
proc macros::linkToStaging {lb {append 1}} {
   Trace
   global TA LB
   set data [list]
   if {[llength [$lb curselection]] > 0} {
      foreach {ind} [lsort -decreasing -integer [$lb curselection]] {
         switch -glob -- $lb {
            ".saTop*"  {lappend data [LB::keyToName [LB::keyFromIndex $ind]]}
            ".wFC*"    {lappend data [aplLB::rowIndex2file $ind]}
         }
      }
   }
   ;# Create N lists of files, for each root drive
   foreach file [lflip $data] {
      lappend toStage([fileRoot $file]) $file
   }
   ;# For each root drive, process the files
   foreach folder [array names toStage] {
      set staging [file join $folder "staging"]
      if {[file exists $staging]} {
         ;# First, find the maximum number to start with
         set i 0
         set existing [lsort -dictionary [glob -nocomplain [file join $staging *]]]
         foreach track $existing {
            set front [string trimleft [lindex [split [file tail $track] -] 0] "0"]
            if {[string is integer -strict $front] && $front > $i} {
               set i $front
            }
         }
         incr i
         saLog "Starting with $i"
         ##
         # Now link the files in
         #
         foreach file [lflip $toStage($folder)] {
            set num [format "%04d - " $i]
            set newName [file join $staging $num[file tail $file]]
            if {![file exists $newName]} {
               saLog "Creating $newName"
               if {[catch {file link $newName $file} result]} {
                  saLog $result
               } else {
                  incr i
               }
            }
         }
      } else {
         puts "$staging doesn't exist"
      }
   }
}
proc macros::showWithoutCoverArt {pattern} {
   foreach {folder} [aplFolder::foldersMatchingPattern $pattern 1] {
      if {[llength [glob -nocomplain -directory $folder *.jpg]]==0} {
         puts "Empty:$folder"
         update
      }
   }
}
##
# Utility to rename cover art to a standard name
#
proc macros::renameCoverArt {pattern} {
   set FJ "Folder.jpg"
   # AlbumArtSmall.jpg is for small images
   foreach {folder} [aplFolder::foldersMatchingPattern $pattern 1] {
      update
      if {[catch {cd $folder} result] } then {
         saLog $result
         continue
      }
      set hasFJ [file exists $FJ]
      set images [glob -nocomplain *.jpg]
      set numImages [llength $images]
      if {$numImages==1} {
         set useImage [lindex $images 0]
         if {$hasFJ==0} {
            puts "$folder Renaming '$useImage' to '$FJ'"
            if {[catch {file rename -force $useImage $FJ} result] } then {
               saLog $result
            }
         }
      } elseif {$numImages>1} {
         puts "$folder has [llength $images] images"
         set maxDate 0
         foreach {image} $images {
            # Use the latest image date
            set mtime [file mtime $image]
            if {$mtime > $maxDate} {
               set useImage $image
               set maxDate $mtime
            }
         }
         if {[catch {file rename -force $useImage $FJ} result] } then {
            saLog $result
         }
      }
   }
}
namespace eval mp3gain {
   variable STOP 0
   variable DONE 0
   variable QUEUE [list]
   variable mp3gainLoc
   upvar #0  snackAmpSettings(mp3gainLoc) mp3gainLoc
}
#----------------------------------------------------------------------------
# When a file is readable, dump output to the console
#----------------------------------------------------------------------------
proc mp3gain::isReadable {f} {
   variable STOP
   variable DONE
   # The channel is readable; try to read it.
   # Note: The initial eof check isn't really needed
   if {[eof $f]} {
      # End of file on the channel
      catch {fileevent $fid readable {}}
      catch {close $f}
      saLog "end of file"
      set DONE 1
   } else {
      set status [catch {gets $f line} result]
      if { $status != 0 } {
         # Error on the channel
         saLog "error reading $f: $result"
         set DONE 2
      } elseif {$result >= 0} {
         # Successfully read the channel
         saLog "> $line"
         set line [string trim $line]
         if {[string length $line]>5} {
            waitWindow::show $line 1
         }
      } elseif {[fblocked $f]} {
         # Read blocked.  Just return
      } elseif {[eof $f]} {
         catch {fileevent $fid readable {}}
         catch {close $f}
         saLog "end of file on read"
         set DONE 1
      } else {
         # Something else
         saLog "can't happen"
         set DONE 3
      }
   }
}

#------------------------------------------------------------------------------
# Launch and block on execution of a program
#------------------------------------------------------------------------------
proc mp3gain::launch {folder cmd} {
   variable STOP
   variable DONE
   set DONE 0
   # Open a pipe
   cd $folder
   set fid [open "|$cmd"]
   saLog "Opening $fid";update
   # Set up to deliver file events on the pipe
   fconfigure $fid -buffering line -blocking 0
   fileevent $fid readable [list mp3gain::isReadable $fid]
   # Launch the event loop and wait for the file events to finish
   vwait mp3gain::DONE
   # Close the pipe
   catch {close $fid}
   saLog "Closing $fid";update
}
#------------------------------------------------------------------------------
# Create an mp3gain command file
#------------------------------------------------------------------------------
proc mp3gain::mp3gain {recurse {opts "/s s /a /d 3.0"}} {
   variable STOP
   variable QUEUE
   variable mp3gainLoc
   lappend folders $::TA(highlightPath)
   if {$recurse} {
      foreach {folder} [aplFolder::foldersMatchingPattern [file join $::TA(highlightPath) *] 1] {
         lappend folders $folder
      }
      saLog "Processing [llength $folders]"
   }
   foreach key $folders {
      if {[waitWindow::ifbreak]} {
         break
      } else {
         if {[file exists $key]} {
            waitWindow::show "Processing '$key'" 1
            saLog "Processing '$key'";update
            mp3gain::launch $key "$mp3gainLoc $opts /k /q *.mp3"
            if {[waitWindow::ifbreak]} {
               break
            }
            gainControl::resetSome [db::filesInFolder $key]
            if {[string match "*/a*" $opts]} {
               saLog "Computing folder average";update
               gainControl::averageSome [db::filesInFolder $key]
            } else {
               saLog "Computing track averages";update
               gainControl::computeSome [db::filesInFolder $key]
            }
         }
      }
   }
   waitWindow::close
}
#----------------------------------------------------------------------------
# Create a batch file to launch mp3gain manually
#----------------------------------------------------------------------------
proc mp3gain::mp3gainbat {{pattern *}} {
   global teFolders
   set teKeys [lsort [array names teFolders $pattern]]
   set tail [list]
   set fod [open c:/mp3gain.bat w]
   foreach {key} $teKeys {
      if {[lsearch -sorted -glob $teKeys $key/*]<0} {
         puts $fod "cd \"[file nativename $key]\""
         puts $fod "mp3gain /a /d 3.0 /k /q *.mp3\""
      }
   }
   close $fod
}

#------------------------------------------------------------------------------
# Update the tag data from the database in case the file tags were
# corrupted.
#------------------------------------------------------------------------------
proc macros::fixTags {{pattern *}} {
   waitWindow::show "Fixing ... please wait"
   set rows [mk::select $db::DB -glob File $pattern -exact Mounted 1 -exact Tag "V1.2"]
   set m [llength $rows]
   set n 0
   foreach {row} $rows {
      array unset dbData
      array set dbData [mk::get $db::DB!$row]
      set fullName $dbData(File)
      set result [id3Tag::id3Label $fullName id3Data V1]
      if {[string length $dbData(Title)] > [string length $id3Data(Title)]} {
         puts "db='$dbData(Title)' tag='$id3Data(Title)'"
         foreach {field} [array names id3Data] {
            catch {set id3Data($field) $dbData($field)}
         }
         #if {[catch {id3Tag::id3V1Modify $fullName id3Data} result] } then {
         #   saLog $result
         #}
      }
      if {$row%10==0} {
         waitWindow::percent $n $m "Fixing $n of $m please wait..."
      }
      incr n
   }
   waitWindow::close
}
proc macros::fixTags1 {{pattern *}} {
   waitWindow::show "Fixing ... please wait"
   set rows [mk::select $db::DB -glob File $pattern -exact Mounted 1]
   set m [llength $rows]
   set n 0
   foreach {row} $rows {
      array unset dbData
      array unset id3V1Data
      array unset id3V2Data
      array set dbData [mk::get $db::DB!$row]
      set fullName $dbData(File)
      id3Tag::id3Label $fullName id3V1Data V1
      id3Tag::id3Label $fullName id3V2Data V2
      # Note: Need to strip .. at end of V1 tag first
      catch {set id3V1Data(Title) [string trimright $id3V1Data(Title) .]}
      if {
          [info exists id3V2Data(Title)] && [info exists id3V1Data(Title)] && \
          [string length $id3V2Data(Title)] > [string length $id3V1Data(Title)] && \
          [string match $id3V1Data(Title)* $id3V2Data(Title)]
      } {
         puts "v2='$id3V2Data(Title)'\nv1='$id3V1Data(Title)'\n";update
         if {[catch {id3Tag::id3V1Modify $fullName id3V2Data} result] } then {
            saLog $result
         }
      }
      if {$row%10==0} {
         waitWindow::show "Fixing $n of $m please wait..."
      }
      incr n
   }
   waitWindow::close
}
proc macros::tagtest {{pattern *}} {
   package require id3dll
   set rows [mk::select $db::DB -glob File $pattern -exact Mounted 1 -exact Tag "V1.2"]
   foreach {row} $rows {
      array unset dbData
      array set dbData [mk::get $db::DB!$row]
      set fullName $dbData(File)
      if {[catch {array set v2Data [id3::id3Label $fullName V2]} result] } then {
         continue
      }
      if {[catch {array set v1Data [id3::id3Label $fullName V1]} result] } then {
         continue
      }
      array set dbData [mk::get $db::DB!$row]
      if {[catch {
         if {[string length $v2Data(Title)] > [string length $dbData(Title)]} {
            puts "v2='$v2Data(Title)' db='$dbData(Title)'";update
   #          foreach {field} [array names id3Data] {
   #             catch {set id3Data($field) $dbData($field)}
   #          }
   #          if {[catch {id3Tag::id3V1Modify $fullName id3Data} result] } then {
   #             saLog $result
   #          }
         }
      } result] } then {
         saLog "$fullName ->$result"
         return
      }
      if {$row%100==0} {puts -nonewline .;update}
   }
}

##
# Find the start/ending point of the track
#
proc macros::fixStart {pattern} {
   set i 0
   set DB  $db::DB
   set result [list]
   set 2weekago [expr {[clock seconds]-2*7*24*60*60}]
#   set rows [mk::select $DB -exact Mounted 1 -glob File $pattern -min date $2weekago]
   set rows [mk::select $DB -exact Mounted 1 -glob File $pattern]
   set i 0
   foreach {row} $rows {
      set fname  [mk::get $DB!$row File]
      waitWindow::percent [incr i] [llength $rows] "Finding Start/Stop\n$fname..." 1
      if {[waitWindow::ifbreak]} {
         break
      }
      set snd [soundControl::infoObj $fname]
      ##
      # Save the start and stop positions
      #
      set Start [soundControl::findStart $snd]
      if {[string is integer -strict $Start]} {
         db::setVal $fname Start $Start
      } else {
         saLog "gainControl::create Got '$Start' for $fname Start Position"
         set Start 10000
      }
      set Stop [soundControl::findStop $snd]
      if {[string is integer -strict $Stop]} {
         db::setVal $fname Stop $Stop
      } else {
         saLog "gainControl::create Got '$Stop' for $fname Stop Position"
         set Stop 10000
      }
   }
   waitWindow::close
   return  1
}
##
# Correct incorrect ID3V2 tag lengths at the beginning of a file
#
proc macros::fixTag {pattern {fix 0}} {
   set DB  $db::DB
   set i 0
   set result [list]
   #set 2weekago [expr {[clock seconds]-2*7*24*60*60}]
   #set rows [mk::select $DB -exact Mounted 1 -glob File $pattern -max date $2weekago]
   set rows [mk::select $DB -exact Mounted 1 -glob File $pattern]
   set i 0
   foreach {row} $rows {
      set file  [mk::get $DB!$row File]
      waitWindow::percent [incr i] [llength $rows] "Checking $file" 1
      if {[waitWindow::ifbreak]} {
         break
      }
      if {[id3Tag::checkTagLength $file $fix data]} {
         if {[array size data]>0} {
            #array unset data APIC
            ##puts "[array size data]-----$file"
            #puts "------$file"
            #parray data
         }
      }
   }
   waitWindow::close
   return
}

package provide macros 1.0
