##################################################################################
# snackAmp - metakit package 1.0 for snackAmp
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
##################################################################################
set [file rootname [file tail [info script]]]_Metakit {
   $Header: /cvsroot/snackamp/snackamp/lib/metakit.tcl,v 1.18 2004/07/12 23:42:50 wilkason Exp $
}
;##
;# Metakit Interface
;#
package require Mk4tcl

namespace eval db {
   variable isSorted 0
   variable isDirty 0 ;# Can be used to flush database on a routine basis
   variable Cols
   variable dbObj
   ;# Metakit but on linux, won't work with ~ as part of name, need full path
   variable dbLocation [file nativename [file join $::HOME snackDB].mk]
   variable csvLocation [file nativename [file join $::HOME snackDB].mke]
   variable dbLock     [file nativename [file join $::HOME snackDB].lck]
   variable fieldSpec [list File Folder Mounted:I FullData:I Depth:I Title Artist Album Track Media Year Genre Note Comp Enc Desc Rate:S sampRate:D Duration:D size:I date:I Gain Tag Played LastPlay:I Rating Start:I Stop:I Votes:I]
   variable fields
   variable changesList
   variable checkTime "mtime"
   variable state "close"
   variable type "metakit"
   variable DB "db.catalog"
   variable mods 0
   variable ratingOptions [list -- * ** *** **** *****]
   variable commitRate 1000
   variable commitAfter 0
   # Build up variables for substitution
   lappend changesList "\[list"
   foreach {fld} $fieldSpec {
      set f [lindex [split $fld :] 0]
      if {$f=="File"} {
         lappend changesList $f \$$f
      } else {
         lappend fields $f
         lappend changesList $f \$Details($f)
      }
   }
   lappend changesList "\]"
   # Make it a nice list for substitution
   set changesList [join $changesList]
}
#----------------------------------------------------------------------------
# Create a lock file
#----------------------------------------------------------------------------
proc db::lock {args} {
   Trace
   variable dbLock
   if {![file exists $dbLock]} {
      set fod [open $dbLock w]
      puts $fod "[list $::tcl_platform(user) [clock seconds] [info hostname]]"
      close $fod
      return 1
   } else {
      return 0
   }
}
#----------------------------------------------------------------------------
# Release a lock file
#----------------------------------------------------------------------------
proc db::unlock {args} {
   Trace
   variable state
   variable dbLock
   if {[file exists $dbLock]} {
      if {[catch {file delete -force $dbLock} result] } then {
         saLog $result
      }
   }
}
#----------------------------------------------------------------------------
# Return info on a lock filed (if one exists)
#----------------------------------------------------------------------------
proc db::lockInfo {args} {
   Trace
   variable dbLock
   if {[file exists $dbLock]} {
      foreach {var} {user time host} {set $var {}}
      set fid [open $dbLock r]
      foreach {user time host} [read $fid] {break}
      close $fid
      return [list $user $time $host]
   } else {
      return [list]
   }
}
proc db::exists {args} {
   variable dbLocation
   return [file exists $dbLocation]
}
#----------------------------------------------------------------------------
# Function   : db::open
# Description: Create the user database if needed
# Author     : Tom Wilkason
# Date       : 9/21/2002
#----------------------------------------------------------------------------
proc db::openDB {args} {
   variable dbLocation
   variable dbLock
   variable dbObj
   variable fieldSpec
   variable state
   variable DB
   # Ignore if db already open
   if {[string equal $state "close"]} {
      # Check for lock file
      set linfo [db::lockInfo]
      if {[llength $linfo]>0} {
         foreach {user time host} $linfo {break}
         waitWindow::close
         set res [tk_messageBox -title "Database Locked" -type yesno -icon warning \
            -message "The SnackAmp database is locked by user $user on host $host,\ndo you want to override the lock?"]
         if {$res=="yes"} {
            db::unlock
         } else {
            exit
         }
      }
      # First time, must create $dbLocation
      if {[catch {mk::file open db $dbLocation -nocommit} result]} {
         waitWindow::close
         set res [tk_messageBox -title "Database Open Error" -type ok -icon warning \
            -message "The SnackAmp database '$dbLocation' could not be opened\n\n$::errorInfo"]
         exit
      }

      set state "open"
      ##
      # main db to hold data, uses blocking
      #
      mk::view layout db.catalog $fieldSpec
      mk::view open db.catalog __db_data ;# Don't touch __db_data command
      ##
      # db.catalog_map is secondary view for hashing
      #
      mk::view layout db.catalog_map {_H:I _R:I}
      ##
      # Create a hash map view for direct access on find
      # combines the base and hash map views
      mk::view open db.catalog_map __db_map
      # __db_data is raw command, __db_map is just map view, dbObj is hashed view
      # Create an ordered/hash view, all udpates are made in an ordered fashion
      # so we don't have to do any sorting during the query
      # use sorted view
      set dbObj [[__db_data view hash __db_map 1] view ordered 1]
      # use unsorted view
      #set dbObj [__db_data view hash __db_map 1]
      ### Needs sorting to return something from select
      db::lock
      ##
      # To create a mapped view on Mounted, create a new view (v2) and
      # populate it with a single integer value containing all the
      # rows returned from a select Mounted 1 from the base view (v1)
      # set v3 [$v1 view map $v2]. Then use the v3 view to access all
      # of the data from $v1. Not sure if adds/deletes will work correctly?
   }
}
#----------------------------------------------------------------------------
# Create a music listing of all the tracks
#----------------------------------------------------------------------------
proc db::document {expfile} {
   variable dbObj
   variable DB
   set catFiles [list]
   if {[catch {open $expfile w} fid]} {
      return
   }
   fconfigure $fid -buffersize 500000
   set rows [mk::select $DB]
   puts $fid {"Artist","Album","Track","Title","File"}
   foreach {row} $rows {
      foreach {File Artist Album Track Title} [mk::get $DB!$row File Artist Album Track Title] {break}
      puts $fid [subst {"$Artist","$Album","$Track","$Title","$File"}]
   }
   close $fid
   tk_messageBox -icon info -type ok -message "csv list of files created as $expfile"
   saLog "[llength $rows] records exported"
}
##
# Export all of the data to a flat file for later import
#
proc db::exportOnePath {expfile {path /}} {
   global TA snackAmpSettings
   variable dbObj
   variable DB
   set pattern [file join $path *]
   set catFiles [list]
   if {[catch {open $expfile w} fid]} {
      saLog "$fid"
      return 0
   }
   fconfigure $fid -buffersize 500000 -buffering full -translation lf
   set trim [string length $path]
   set rows [mk::select $DB -glob File $pattern]
   set i 0
   set tr [llength $rows]
   foreach {row} $rows {
      set Row [mk::get $DB!$row]
      set File [lindex $Row 1]
      set Folder [lindex $Row 3]
      # Make it relative to base folder
      lset Row 1 [string trimleft [string range $File $trim end] /]
      lset Row 3 [string trimleft [string range $Folder $trim end] /]
      set Row [string map {\n " "} $Row] ;# no embedded cr or lf (0xA=LF is used in ID3 tags)
      puts $fid $Row
      incr i
      if {($i%100)==0} {
         waitWindow::percent $i $tr "Exporting\n$path to\n$expfile" 1
         if {[waitWindow::ifbreak]} {
            break
         }
      }
   }
   close $fid
   saLog "$i records exported to $expfile"
   return $i
}
##
# Export all of the data to a flat file for later import
#
proc db::exportRAW {args} {
   variable dbObj
   variable DB
   variable csvLocation
   set catFiles [list]
   set im 0
   if {[catch {open $csvLocation w} fid]} {
      saLog "$fid"
      return $im
   }
   set size [$dbObj size]
   fconfigure $fid -buffersize 500000 -buffering full -translation lf
   set rows [mk::select $DB -sort File]  ;# sort just in case not kept sorted
   # BUG: Some Note fields have CRLF in them causing a problem, how to deal
   foreach {row} $rows {
      set ROW [string map {\n " "} [mk::get $DB!$row]]
      puts $fid $ROW
      if {($im%200)==0} {
         waitWindow::percent $im $size "Exported $im records ..."
      }
      incr im
   }
   close $fid
   saLog "[llength $rows] records exported"
   return $im
}
##
# Import a previously exported file, append to the database
#
# set f C:/Incoming/snackDB.mke

proc db::importRAW {size} {
   global TA snackAmpSettings
   variable dbObj
   variable csvLocation
   ##
   # Restore the cached files from each media Dir
   #
   set catFiles [list]
   if {[catch {open $csvLocation r} fid]} {
      return
   }
   fconfigure $fid -buffersize 100000 -buffering full -translation lf
   set im 0
   while {[gets $fid row]>0} {
      incr im
      if {[catch {eval $dbObj insert end $row} result]} then {
         saLog "$result:\n\t$row"
         continue
      }
      if {($im%100)==0} {
         waitWindow::percent $im $size "Imported $im records ..."
      }
   }
   close $fid
   saLog "$im records imported"
   mk::file commit db
}
#----------------------------------------------------------------------------
# Import a single group of files within a path
# Fixup the relative path names properly
#----------------------------------------------------------------------------
proc db::importOnePath {file path} {
   global TA snackAmpSettings
   variable dbObj
   ##
   # Restore the cached files from each media Dir
   #
   set catFiles [list]
   set expfile [file join $path $file]
   if {[catch {open $expfile r} fid]} {
      saLog $fid
      return [list 0 0]
   }
   fconfigure $fid -buffersize 100000 -buffering full -translation lf
   set im 0
   set rp 0
   set i 0
   # Flag all in this branch as unmounted
   db::unmountBranch [file join $path *]
   while {[gets $fid row]>0} {
      if {[catch {file join $path [lindex $row 1]} File] } then {
         saLog "$File -> $row"
         continue
      }
      set Folder [file dirname $File]
      array unset Info
      array set Info $row
      set Info(File) $File
      set Info(Folder) $Folder
      set Info(Depth) [llength [file split $File]]
      ##
      # Append or replace, depending if entry exists
      #
      if {[catch {$dbObj find File $File} index] } then {
         incr im
         set row [array get Info]
         eval $dbObj insert end $row
      } else {
         incr rp
         # Only import certain fields if they are replacing default fields
         #
         array set OldData [$dbObj get $index]
         set Info(Mounted) 1
         # Gain, only if mine is --, otherwise gain, start & stop
         if {$OldData(Gain)!="--"} {
            unset Info(Gain)
            unset Info(Start)
            unset Info(Stop)
         }
         # Rating, only if mine is --
         if {$OldData(Rating)!="--"} {
            unset Info(Rating)
         }
         # Played, import larger
         if {$OldData(Played) > $Info(Played)} {
            unset Info(Played)
         }
         # LastPlay, import larger
         if {$OldData(LastPlay) > $Info(LastPlay)} {
            unset Info(LastPlay)
         }
         # Some fields may now be gone, so we won't set them
         set row [array get Info]
         eval $dbObj set $index $row
      }
      if {([incr i]%100)==0} {
         waitWindow::show "Added $im, and replaced $rp records for $path ..."
      }
   }
   close $fid
   set removed [db::removeUnmountedFiles [file join $path *]]
   saLog "$im records added, $rp records replaced, $removed records removed"
   return [list $im $rp $removed]
}
##
# Return a list of folders in a tree
#
proc db::folderList {{pattern *}} {
   variable DB
   set rows [mk::select $DB -glob Folder $pattern -exact Mounted 1]
   foreach {row} $rows {
      set hash([mk::get $DB!$row Folder]) {}
   }
   return [lsort [array names hash]]
}
#----------------------------------------------------------------------------
# Function   : restart
# Description: Close and delete the databases
# Author     : Tom Wilkason
# Date       : 12/26/2002
#----------------------------------------------------------------------------
proc db::restart {args} {
   Trace
   variable dbLocation
   variable csvLocation
   waitWindow::show "Exporting Database..."
   Wait 1
   set exported [db::exportRAW]
   if {$exported > 0} {
      waitWindow::show "Recreating Database..."
      db::closeDB
      # TBD: Rename file, then delete it when import succeeds
      file delete -force -- ${db::dbLocation}.bak
      file rename -force -- $db::dbLocation ${db::dbLocation}.bak
      db::openDB
      waitWindow::show "Importing Database..."
      db::importRAW $exported
      waitWindow::release
      waitWindow::close
      Wait 0
   } else {
      saLog "No files exported"
   }
}
#----------------------------------------------------------------------------
# Function   : db::lookup
# Description: Return details for a file, if not in db the optionally add it
# Author     : Tom Wilkason
# Date       : 12/25/2002
#----------------------------------------------------------------------------
proc db::lookup {File _Data {cache 1}} {
   upvar $_Data Data
   variable dbObj
   set data [list]
   if {[catch {$dbObj find File $File} row] } then {
      ;# Didn't exist, add the data if needed
      if {$cache} {
         set cr [db::cacheFile $File Data 1]
      } else {
         set cr [db::cacheFile $File Data 0]
      }
   } else {
      # Existed
      array set Data [$dbObj get $row]
      if {$Data(FullData)==0 && $cache==1} {
         set cr [db::cacheFile $File Data 1 $row]
      } else {
         set cr 1
      }
   }
   # If cache failed, return 0
   if {$cr==0} {
      return 0
   }
   # This happens on file not existing
   set Data(Gain) [gainControl::levelText $Data(Gain)]
   #set Data(Folder) [file dirname $Data(File)]
   return [array size Data]
}

#----------------------------------------------------------------------------
;# Cache a single file name looking up info for it
;# If it exists, do nothing
#----------------------------------------------------------------------------
proc db::cacheFileName {File {chkMount 1}} {
   Trace
   variable checkTime
   variable dbObj
   ;# If file not in db, then cache just basic part
   if {[catch {$dbObj find File $File} row] } then {
      if {[catch {file stat $File Stat} result]} {
         debug $result
         return -1
      } else {
         set row [$dbObj size]
         $dbObj insert end \
            File     $File        \
            Folder   [file dirname $File] \
            Depth    [llength [file split $File]]   \
            FullData 0            \
            size     $Stat(size)  \
            date     $Stat($checkTime) \
            Played   0  \
            LastPlay 0  \
            Rating   "--"  \
            Votes    0  \
            Mounted  1  \
            Gain     "--"
         db::dirty
      }
   } else {
      # Found, flag as mounted
      if {$chkMount} {
         $dbObj set $row Mounted 1
      }
   }
   return $row
}
#----------------------------------------------------------------------------
# Increment the play counter for some file
#----------------------------------------------------------------------------
proc db::bumpPlayed {File {num 1}} {
   Trace
   variable dbObj
   if {[catch {$dbObj find File $File} row] } then {
      debug $row
      return
   }
   foreach {played last} [$dbObj get $row Played LastPlay] {break}
   if {![string is integer -strict $played]} {
      set played 1
   } else {
      incr played $num
   }
   # Update play count and last played (round to hour)
   set v [expr {$::snackAmpSettings(playResolution)*3600}]
   $dbObj set $row Played $played LastPlay [expr {([clock seconds]/$v)*$v}]
   db::dirty
}
#----------------------------------------------------------------------------
# Increment the rating for some file by some amount
#----------------------------------------------------------------------------
proc db::bumpRating {File {num 1}} {
   variable dbObj
   if {[catch {$dbObj find File $File} row] } then {
      debug $row
      return
   }
   set rating [$dbObj get $row Rating]
   if {$rating=="--"} {
      if {$num < 1} {set num 1} elseif {$num > 5} {set num 5}
      set rating [string repeat * $num]
   } else {
      set cr [string length $rating]
      incr cr $num
      if {$cr < 1} {set cr 1} elseif {$cr > 5} {set cr 5}
      set rating [string repeat * $cr]
   }
   $dbObj set $row Rating $rating
   db::dirty
}
#------------------------------------------------------------------------------
#  Set a single element to some value in the database
#------------------------------------------------------------------------------
proc db::setVal {File Field val} {
   variable dbObj
   variable mods
   if {[catch {$dbObj find File $File} row] } then {
      debug $row
      return ""
   }
   db::dirty
   $dbObj set $row $Field $val
   return $val
}
#------------------------------------------------------------------------------
#  Get a single element to some value in the database
#------------------------------------------------------------------------------
proc db::getVal {File Field} {
   variable dbObj
   if {[catch {$dbObj find File $File} row] } then {
      return ""
   }
   set val [$dbObj get $row $Field]
   return $val
}
#------------------------------------------------------------------------------
#  Get a single element to some value in the database
#------------------------------------------------------------------------------
proc db::getValues {File Fields} {
   variable dbObj
   if {[catch {$dbObj find File $File} row] } then {
      return [list] ;# TBD: return blank
   }
   set val [eval $dbObj get $row $Fields]
   return $val
}

#----------------------------------------------------------------------------
# Increment the rating for a viewed folder (or folders)
#----------------------------------------------------------------------------
proc db::bumpFolderRating {pattern depth {num 1}} {
   foreach {file} [db::filesInFolder $pattern $depth] {
      db::bumpRating $file $num
   }
}
proc db::bumpTrackRating {tracks {num 1}} {
   foreach {file} $tracks {
      db::bumpRating $file $num
   }
}

#----------------------------------------------------------------------------
# Cache a single file looking up info for it
# If row passed in, assume it already exists
#----------------------------------------------------------------------------
proc db::cacheFile {File _Details {getId3 1} {row ""}} {
   variable dbObj
   variable changesList
   variable checkTime
   variable fields
   upvar $_Details Details
   # Start building command for eval
   set cmd [list $dbObj]
   # Save user info it not getting fresh
   if {$getId3==0} {
      array set userDetails [array get Details]
   }
   ##
   # Was row specified? If so it must exist
   # Pull known data into Details
   if {[string length $row]} then {
      array set Details [$dbObj get $row]
      lappend cmd set $row File $File
   } else {
      if {[catch {$dbObj find File $File} row] } then {
         lappend cmd insert end File $File
      } else {
         array set Details [$dbObj get $row]
         lappend cmd set $row File $File
      }
   }
   ;##
   ;# File info
   ;#
   if {[catch {file stat $File Stat} result]} {
      saLog $result
      return 0
   }
   ##
   # Read the track/id3 info on the file
   # if no data was passed in, otherwise
   # use the user passed in data to override
   # any existing data
   if {$getId3} {
      # Assume adding the file
      # replace with split calls for track/id3 capture
      if {[catch {catalog::trackFullData $File New} result] } then {
         saLog $result
         return 0
      }
      # Restore certain fields from the database not contained in the tag
      foreach {field type} {
         Gain     double
         LastPlay integer
         Played   integer
         Rating   ascii
         Start    integer
         Stop     integer
         Votes    integer
      } {
         if {[info exists Details($field)] && [string is $type -strict $Details($field)]} {
            set New($field) $Details($field)
         }
      }
      set Details(FullData) 1
      array set Details [array get New]
   } else {
      array set Details [array get userDetails]
   }
   # If play info doesn't exist, then fill it in, these are not found
   # in an ID3 tag or snack info
   foreach {field type default} {
      FullData integer 0
      Gain     double  "--"
      LastPlay integer 0
      Played   integer 0
      Rating   ascii   "--"
      Start    integer 0
      Stop     integer 0
      Votes    integer 0
   } {
      if {![info exists Details($field)] || ![string is $type -strict $Details($field)]} {
         set Details($field) $default
      }
   }
   # These fields always get updated
   set Details(Depth) [llength [file split $File]]
   set Details(File)  [file tail $File]
   set Details(Folder)  [file dirname $File] ;# not sure why this is needed again
   set Details(size) $Stat(size)
   set Details(date) $Stat($checkTime)
   # Build command to update only known fields
   foreach Index $fields {
      if {[info exists Details($Index)]} {
         lappend cmd $Index $Details($Index)
      }
   }
   # Does eval ...  set [array get Details] not work properly?
   eval $cmd
   db::dirty
   return 1
}
#----------------------------------------------------------------------------
# Function   : db::removeFile
# Description: Remove a file from the database
# Author     : Tom Wilkason
# Date       : 1/19/2003
#----------------------------------------------------------------------------
proc db::removeFile {File} {
   variable dbObj
   Trace
   if {[string is integer -strict $File]} {
       $dbObj delete $File ;# Make sure row indicies work here
   } else {
      if {[catch {$dbObj find File $File} row] } then {
         return 0
      } else {
      ;# tbd: just flag as invalid
         $dbObj delete $row ;# Make sure row indicies work here
         # mk::row delete $row
      }
   }
   return 1
}

#------------------------------------------------------------------------------
# Remove all unmounted files and files that don't exist
#------------------------------------------------------------------------------
##
# Remove all tracks that are not mounted
# Have to use the object interface, using
# the mk::row delete $DB interface causes a later crash
# on append, maybe due to the hash map view.
#
proc db::removeUnmountedFiles {{pattern *}} {
   variable dbObj
   set rows [list]
   set removed 0
   $dbObj loop row {
      if {[$dbObj get $row Mounted]==0} {
         if {[string match $pattern [$dbObj get $row File]]} {
            lappend rows $row
         }
      }
   }
   foreach row [lsort -integer -decreasing $rows] {
      $dbObj delete $row
      incr removed
   }
   #saLog "Removed [llength $rows] unmounted tracks"
   mk::file commit db
   return $removed
}

##
# Remove a branch from the database, work backwards
# (there may be a better way to do this but can't find one)
proc db::removeBranch {pattern} {
   variable dbObj
   Trace
   set rows [list]
   $dbObj loop row {
      if {[string match $pattern [$dbObj get $row File]]} {
         lappend rows $row
      }
   }
   foreach row [lsort -integer -decreasing $rows] {
      $dbObj delete $row
   }
   mk::file commit db
   return [llength $rows]
}
#----------------------------------------------------------------------------
# Function   : db::renameFile
# Description: renameFile a file
# Author     : Tom Wilkason
# Date       : 1/19/2003
# Note: Because this is a sorted view, we must change the filename last
#       because it is moved within the view. (bug)
#----------------------------------------------------------------------------
proc db::renameFile {File newName} {
   variable dbObj
   if {[string is integer -strict $File]} {
      $dbObj set $File Folder [file dirname $newName] Depth [llength [file split $newName]] File $newName
      db::dirty
   } else {
      if {[catch {$dbObj find File $File} row] } then {
         saLog "$File not found"
         return -1
      } else {
         $dbObj set $row Folder [file dirname $newName] Depth [llength [file split $newName]] File $newName
         db::dirty
         return 1
      }
   }
   return 1
}
#----------------------------------------------------------------------------
# Close the database
#----------------------------------------------------------------------------
proc db::closeDB {args} {
   Trace
   variable state
   variable mods
   db::commit
   set mods 0
   db::unlock
   if {![string equal $state "close"]} {
      mk::file close db
      set state "close"
   }
}
#----------------------------------------------------------------------------
# If the database has been flaged as modified, commit it and clear flag
#----------------------------------------------------------------------------
proc db::commit {args} {
   variable isDirty
   variable state
   variable mods
   set m $mods
   if {$mods > 0 && ![string equal $state "close"]} {
      Trace
      mk::file commit db
#saLog "Commiting Changes"
      set mods 0
      return $m
   } else {
      return 0
   }
}
#----------------------------------------------------------------------------
# Flag database as dirty, use for commiting changes
#----------------------------------------------------------------------------
proc db::dirty {args} {
   Trace
   variable state
   variable mods
   variable commitRate
   variable commitAfter
   incr mods
   return
   ##
   # Push a commit into the future
   #
#    if {![string equal $state "close"]} {
#       after cancel $commitAfter
#       set commitAfter [after 5000 db::commit]
#       incr mods
#    }
}
#----------------------------------------------------------------------------
# Return the number of tracks indexed
# Note: Should limit to media folders
#----------------------------------------------------------------------------
proc db::numTracks {args} {
   variable dbObj
   return [llength [db::sortedList]]
}
##
# Relative to above: This is faster near the root, slower at the tail
# //Bertha/Music/Rock/Album Rock/Moody Blues/Time Traveller
proc db::filesInFolder {folder {depth 0}} {
   Trace
   variable DB
   variable dbObj
   set result [list]
   set folder [string trimright $folder "/*"]
   if {$depth>0} {
      set pattern [file join $folder *]
      set maxDepth [expr {[llength [file split $folder]]+$depth+1}]
      set rows [mk::select $DB -glob File $pattern -max Depth $maxDepth -exact Mounted 1]
   } else {
      set rows [mk::select $DB -exact Folder $folder -exact Mounted 1]
   }
   foreach {row} $rows {
      lappend result [mk::get $DB!$row File]
   }
   return $result
}
#----------------------------------------------------------------------------
# Return a list of files not matching a pattern
# Needs metakit select -not option
#----------------------------------------------------------------------------
proc db::filesNotMatchingPattern {pattern {ignoreCase 0}} {
   Trace
   variable DB
   set result [list]
   variable dbObj
   if {$ignoreCase} {
      foreach {row} [mk::select $DB -exact Mounted 1] {
         set name [mk::get $DB!$row File]
         if {![string match -nocase $pattern $name]} {
            lappend result $name
         }
      }
   } else {
      foreach {row} [mk::select $DB -exact Mounted 1] {
         set name [mk::get $DB!$row File]
         if {![string match $pattern $name]} {
            lappend result $name
         }
      }
   }
   return $result
}
#----------------------------------------------------------------------------
# Return files matching some pattern
#----------------------------------------------------------------------------
proc db::filesMatchingPattern {pattern {ignoreCase 0}} {
   Trace
   variable DB
   set result [list]
   if {$ignoreCase} {
      set match -globnc
   } else {
      set match -glob
   }
   # It is faster to sort the result, rather than specify in the select
   foreach {row} [mk::select $DB $match File $pattern -exact Mounted 1] {
      lappend result [mk::get $DB!$row File]
   }
   return $result
}
# ----------------------------------------------------------------------------
# Set/Clear the mounted flag as needed
# ----------------------------------------------------------------------------
proc db::flagMountedFolders {folder how} {
   variable DB
   if {$how} {
      set nothow 0
   } else {
      set nothow 1
   }
   foreach {row} [mk::select $DB  -exact Mounted $nothow -glob Folder "$folder*"] {
      mk::set $DB!$row Mounted $how
   }
}

;# This is a bit faster for a large match set, but slower for smaller sets
proc db::filesMatchingPatternx {pattern {ignoreCase 0}} {
   Trace
   variable dbObj
   set result [list]
   $dbObj loop row {
      foreach {file mounted} [$dbObj get $row File Mounted] {break}
      if {$mounted && [string match $pattern $file]} {
         lappend result $file
      }
   }
   return $result
}
##
# This is about the same as a mk::select -exact Mounted 1
#
proc db::sortedList {args} {
   Trace
   variable dbObj
   variable DB
   set result [list]
   # Looping is very fast
   # unsorted object selects don't work with metakit
   $dbObj loop row {
      foreach {File Mounted} [$dbObj get $row File Mounted] {break}
      if {$Mounted==1} {
         lappend result $File
      }
   }
   return $result
}
proc db::fixFolders {args} {
   Trace
   variable dbObj
   variable DB
   set result [list]
   # Looping is very fast
   # unsorted object selects don't work with metakit
   $dbObj loop row {
      foreach {File Folder} [$dbObj get $row File Folder] {break}
      if {![string equal $Folder [file dirname $File]]} {
         puts "$File not in $Folder"
         update
         $dbObj set $row Folder [file dirname $File]
      }
   }
   return
}
#----------------------------------------------------------------------------
# Return a list of tracks shuffled then sub-sorted by user specified field
# assumes integer sub-sorting
# modes are a list of textual modes to sort on, either 1,2 or 3 are passed in.
#----------------------------------------------------------------------------
proc db::subSorting {modes {tracks {}}} {
   variable DB
   variable ratingOptions
   global snackAmpSettings
   set db::isSorted 1 ;# not needed here
   set result [list]
   set pairs [list]
   # each mode contains {field -type -sortmode}
   set index 1

   #------------------------------------------------------------------------------
   # For Rating filtering to work, we need to apply it when the APL list is created
   # not during the sort.
   # Also, for the playlist sort case, we can only further restrict the files
   # unless we keep a copy somewhere else.
   #------------------------------------------------------------------------------

   ##
   # Append the fields to be queried
   #
   set fields "File"
   foreach {subsort} [lflip $modes] {
      foreach {field type sort} $subsort {break}
      lappend fields $field
   }
   ##
   # Build the master sorted list
   # Todo: Save only the indices, not the file names
   if {[llength $tracks]==0} {
      ##
      # Execute the query command, used for Auto Playlists
      # (Need to restrict ratings here)
      foreach {row} [mk::select $DB -exact Mounted 1] {
#         set rating [mk::get $DB!$row Rating]
#         catch {
#            if {$::snackAmpSettings(cbRating,$rating)==1} {
               lappend pairs [eval mk::get $DB!$row $fields]
#            }
#         }
      }
   } else {
      # Handling playlist, look up rating for local access
      # Set the rating text itself to a bool value
      foreach rating $ratingOptions {
         set $rating $::snackAmpSettings(cbRating,$rating)
      }
      foreach {track} $tracks {
         Unset Data
         # We will NOT catalog track if needed.
         if {[db::lookup $track Data 0] > 0} {
            catch {
               # Only allow ratings that are set true
               if {[set $Data(Rating)]} {
                  set row [list]
                  foreach {field} $fields {
                     lappend row $Data($field)
                  }
                  lappend pairs $row
               }
            }
         } else {
            saLog "No data found for '$track'";update
         }
      }
   }
   ##
   # Determine if a shuffle is in the inner
   #
   if {$snackAmpSettings(sortShuffle)} {
      Update
      set pairs [shuffle $pairs]
   }
   ##
   # Then subsort, in reverse order
   #
   foreach {subsort} [lflip $modes] {
      foreach {field type sort} $subsort {break}
      Update
      set pairs [lsort $sort $type -index $index $pairs]
      incr index
   }
   # Return just the file names
   foreach item $pairs {
      lappend result [lindex $item 0]
   }
   ##
   # Morph the list a bit
   #
   if {$snackAmpSettings(postmorph)} {
      Update
      lipmorph result $snackAmpSettings(postmorphamount)
   }
   return $result
}
#----------------------------------------------------------------------------
;# Position in db of a file, -1 if not in db
#----------------------------------------------------------------------------
proc db::whereInDb {File} {
   Trace
   variable dbObj
   if {[catch {$dbObj find File $File} row] } then {
      return -1
   } else {
      return $row
   }
}
#------------------------------------------------------------------------------
# Return a track name at some row
#------------------------------------------------------------------------------
proc db::TrackFromDb {row} {
   Trace
   variable dbObj
   if {[catch {$dbObj get $row File} File] } then {
      return ""
   } else {
      return $File
   }
}
#----------------------------------------------------------------------------
# First position after some offset, allow glob matching
#----------------------------------------------------------------------------
proc db::matchInDb {pattern {offset 0}} {
   Trace
   variable dbObj
   set result [list]
   set view $dbObj
   set end [$view size]
   # dbObj is already sorted (not really)
   for {set row $offset} {$row < $end} {incr row} {
      set file [$view get $row File]
      if {[string match -nocase $pattern $file]} {
         return $row
      }
   }
   return -1
}

#----------------------------------------------------------------------------
# First position after some offset, allow glob matching
#----------------------------------------------------------------------------
proc db::fileByPosition {row} {
   Trace
   variable dbObj
   if {[catch {$dbObj get $row File} result] } then {
      return -1
   } else {
      return $result
   }
}
#----------------------------------------------------------------------------
# Function   : db::addToDbIfNeeded
# Description: Add a file to TFList if it is not already there
# Author     : Tom Wilkason
# Date       : 12/15/2002
#----------------------------------------------------------------------------
proc db::addToDbIfNeeded {file} {
   return [db::cacheFileName $file]
}
#----------------------------------------------------------------------------
# Function   : db::makeSorted
# Description: Can always assume the metakit is sorted
# Author     : Tom Wilkason
# Date       : 12/19/2002
#----------------------------------------------------------------------------
proc db::makeSorted {args} {
   variable isSorted 1
   return 1
}
#----------------------------------------------------------------------------

# Function   : db::dirtyList
# Description: Flag the master track list as unsorted
# Author     : Tom Wilkason
# Date       : 12/19/2002
#----------------------------------------------------------------------------
proc db::dirtyList {args} {
   set db::isSorted 0
}
#----------------------------------------------------------------------------
# Function   : db::Commit
# Description: Flush the data after cataloging
# Author     : Tom Wilkason
# Date       : 1/7/2001
#----------------------------------------------------------------------------
proc db::Commit {} {
   mk::file commit db
}
##
# Flag the files in pattern as unmounted
# Can I do this? Is DB via a hash map view?
proc db::unmountBranch {pattern} {
   variable DB
   set rows [mk::select $DB -glob File $pattern -exact Mounted 1]
   foreach {row} $rows {
      mk::set $DB!$row Mounted 0
   }
   return 1
}
#----------------------------------------------------------------------------
# Return a complete tree, including paths to the root.
# TODO: Make a database of the folders, hashed by folder name
# Have only a mounted field
#----------------------------------------------------------------------------
proc db::buildFolderTree {{pattern *}} {
   variable DB
   aplFolder::stripBranch $pattern
   set hits 0
   set how 0
   set rows [mk::select $DB -glob Folder $pattern -exact Mounted 1]
   foreach {row} $rows {
      set folder [mk::get $DB!$row Folder]
      incr hits [aplFolder::setToRoot $folder $how]
   }
   return $hits
}
##
# Return raw database statistics
#
proc db::statsTree {_var {pattern ""}} {
   Trace
   variable DB
   upvar $_var stats
   set i 0
   set stats(Cataloged) 0
   set stats(UnMounted) 0
   set stats(Size) 0
   set fields [list Folder Mounted FullData Artist Album Genre size]
   if {[string length $pattern]==0} {
      set rows [mk::select $DB]
   } else {
      set rows [mk::select $DB -glob File $pattern]
   }
   foreach row $rows {
      foreach val [mk::get $DB!$row Folder Mounted FullData Artist Album Genre size] index $fields {
         switch -- $index {
            "FullData" {
               if {$val} {
                  incr stats(Cataloged)
               }
            }
            "Mounted" {
               if {$val==0} {
                  incr stats(UnMounted)
               }
            }
            "size" {
               incr stats(Size) [expr {$val/1024}]
            }
            default {
               set hash($index,$val) {}
            }
         }
      }
      incr i
      if {$i%500==0} {
         update
      }
   }
   foreach field $fields {
      set stats($field) [llength [array names hash $field,*]]
   }
   set stats(Files) $i
   set stats(Size) [expr {$stats(Size)/1024}]
   return $i
}
##
# Return text of database statistics
#
proc db::statsText {fill {pattern ""}} {
   Wait 1
   db::statsTree Data [file join $pattern *]
   foreach {index tit} {
      Files     "Indexed Tracks"
      Cataloged "Cataloged Tracks"
      UnMounted "UnMounted Tracks"
      Size      "Size (MB)"
      Folder    Folders
      Artist    Artists
      Album     Albums
      Genre     Genres
   } {
      append result "\t[commify $Data($index)]\t$tit\n"
   }
   Wait 0
   return [string trimright $result]
}
#----------------------------------------------------------------------------
# Return a list of distinct items in the database
#----------------------------------------------------------------------------
proc db::unique {item} {
   variable dbObj
   $dbObj loop row {
      set hash([$dbObj get $row $item]) {}
   }
   return [lsort [array names hash]]
}

#------------------------------------------------------------------------------
# Not sure how to use these grouped views yet, maybe for a media
# viewer like winamp5 has.
#------------------------------------------------------------------------------
proc db::subViews {args} {
   variable dbObj
   # group by folders
   set fv [$db::dbObj view groupby ffsv:V Folder Mounted]
   # loop on folders
   $fv loop c {
      set sv [$v open $c ffsv]
      # loop on files in folder
      $sv loop csv {
         puts [$sv get $csv File]
      }
      $sv close
      unset sv
$fv close
unset fv
return
   }
}
#----------------------------------------------------------------------------
# db::tableTreeData
#
# Loop over some depth of folders and return a nested list of data specified
# in the col3Use in tableData as well as the total size and time of the tracks.
# Update locate cache files of the data for faster access next time.
# todo:: Relies on a couple TA varibles and tableList::variables, need
# to decouple them
# TBD: Is filtering on all tracks optimum here? Not for file names only, but
#      for also filtering tag data it is the only way.
#----------------------------------------------------------------------------
proc db::tableTreeData {baseIndex col2Use checkStat depth filter filterHow _totalTime _totalSize _tableData _Index2File} {
   Trace
   variable dbObj
   variable DB
   variable checkTime
   # These are returned data
   upvar $_totalTime totalTime
   upvar $_totalSize totalSize
   upvar $_tableData tableData
   upvar $_Index2File Index2File
   ##
   # Grab all the rows for this depth
   #
   set indLen [string length $baseIndex]
   incr indLen
   set maxDepth [expr {[llength [file split $baseIndex]]+$depth+1}]
   # Tweak select for speed, filtering is against file names
   if {[string length $filter]} {
      Wait 1
      set rows [mk::select $DB -count $::snackAmpSettings(maxAPLApply) -globnc $filterHow *$filter* -exact Mounted 1]
   } elseif {[string length $baseIndex]==0} {
      # Select everything
      Wait 1
      set rows [mk::select $DB -count $::snackAmpSettings(maxAPLApply) -exact Mounted 1]
   } elseif {$depth==0} {
      # Select just one folder
      set rows [mk::select $DB -exact Folder $baseIndex -exact Mounted 1]
   } else {
      # Select multiple folders
      set rows [mk::select $DB -glob File [file join $baseIndex *] -max Depth $maxDepth -exact Mounted 1]
   }
   ##
   # Build the command line to append each row for later eval
   #
   lappend cmdLine lappend rowData
   foreach {col} $col2Use {
      lappend cmdLine "\$Data($col)"
   }
   set cmdLine [join $cmdLine]
   set records [llength $rows]
   set i 0
   # Iterate over each row
   # bug if only one record returned (row 0), gets a failure
   set flen [llength $baseIndex]
   foreach {row} $rows {
      ##
      # Grab the entire row into an array
      #
      array unset Data
      array set Data [mk::get $DB!$row]
      set fullFile $Data(File)
      # Only display tail
      set Data(File) [file tail $fullFile]
      # Cache if necessary
      if {$checkStat} {
         if {[catch {file stat $fullFile Stat} result] } then {
            saLog $result
            continue
         }

         set tstamp [join [list $Stat(size) $Stat($checkTime)] : ]
         set dbstamp [join [list $Data(size) $Data(date)] : ]

         # Formatted fields are shown in tablelist
         set Data(Size) $Stat(size)
         set Data(Date) [clock format $Stat($checkTime) -format $::snackAmpSettings(timeString)]
      } else {
         set Data(Size) $Data(size)
         set Data(Date) [clock format $Data(date) -format $::snackAmpSettings(timeString)]
         set tstamp ""
         set dbstamp ""
      }

      set Data(Sample) $Data(sampRate)

      if {$Data(FullData)==0} {
         waitWindow::percent $i $records "Cataloging\n${baseIndex}..." 1
         db::cacheFile $fullFile Data 1
      } elseif {![string equal $tstamp $dbstamp]} {
         waitWindow::percent $i $records "Updating\n${baseIndex}..." 1
         db::cacheFile $fullFile Data 1
      } elseif {($i%200)==0} {
         Update
      }
      # Get relative folder info
      set Data(Folder) [string range $Data(Folder) $indLen end]
      if {[string length $Data(Folder)]==0} {
         set Data(Folder) .
      }
      set Data(Gain) [gainControl::levelText $Data(Gain)]

      catch {
      ;# TFW: Note size vs Size, need to standardize
         set totalSize [expr {$totalSize+$Data(size)/1024}]
         set totalTime [expr {$totalTime+$Data(Duration)}]
      }
      # Build row of desired data
      set Data(Key) $row
      set rowData [list]

      # Build row and append row to result
      eval $cmdLine
      lappend tableData $rowData
      incr i
   }
   Wait 0
   return 0
}
#----------------------------------------------------------------------------
# Remove then rescan a folders worth of tracks
# Doesn't remove old entries
#----------------------------------------------------------------------------
proc db::cacheFolderDetails {folder {depth 0}} {
   # Remove any stale database entries
   foreach {file} [db::filesInFolder $folder $depth]  {
      if {![file exists $file]} {
         db::removeFile $file
      }
   }
   Update
   # Add any files that may exist and force recache all data
   foreach {file} [eval glob -nocomplain -directory {$folder} [globAllPattern $depth]] {
      if {[catalog::isTrackFile $file]} {
         array unset Data
         db::cacheFile $file Data 1
      }
   }
}

#----------------------------------------------------------------------------
# Catalog some path to some depth
#----------------------------------------------------------------------------
proc db::catalog {baseIndex checkStat depth} {
   Trace
   variable dbObj
   variable DB
   variable checkTime
   ##
   # Grab all the rows for this depth
   #
   set indLen [string length $baseIndex]
   incr indLen
   set maxDepth [expr {[llength [file split $baseIndex]]+$depth+1}]
   # Tweak select for speed, filtering is against file names
   if {[string length $baseIndex]==0} {
      # Select everything
      set rows [mk::select $DB -exact Mounted 1]
   } elseif {$depth==0} {
      # Select just one folder
      set rows [mk::select $DB -exact Folder $baseIndex -exact Mounted 1]
   } else {
      # Select multiple folders
      set rows [mk::select $DB -glob File [file join $baseIndex *] -max Depth $maxDepth -exact Mounted 1]
   }
   set i 0
   # Iterate over each row
   # bug if only one record returned (row 0), gets a failure
   set flen [llength $baseIndex]
   set records [llength $rows]
   foreach {row} $rows {
      ##
      # Grab the entire row into an array
      #
      array unset Data
      array set Data [mk::get $DB!$row]
      set fullFile $Data(File)
      # Cache if necessary
      if {$checkStat} {
         if {[catch {file stat $fullFile Stat} result] } then {
            saLog $result
            # should remove from the database
            continue
         }

         set tstamp [join [list $Stat(size) $Stat($checkTime)] : ]
         set dbstamp [join [list $Data(size) $Data(date)] : ]

         # Formatted fields are shown in tablelist
         set Data(Size) $Stat(size)
         set Data(Date) [clock format $Stat($checkTime) -format $::snackAmpSettings(timeString)]
      } else {
         set Data(Size) $Data(size)
         set Data(Date) [clock format $Data(date) -format $::snackAmpSettings(timeString)]
         set tstamp ""
         set dbstamp ""
      }

      if {$Data(FullData)==0} {
         waitWindow::percent $i $records "Cataloging\n$fullFile..." 1
         db::cacheFile $fullFile Data 1
      } elseif {![string equal $tstamp $dbstamp]} {
         waitWindow::percent $i $records "Updating\n$fullFile..." 1
         db::cacheFile $fullFile Data 1
      } elseif {($i%200)==0} {
         Update
      }
      if {[waitWindow::ifbreak]} {
         return $i
      }
      incr i
   }
   return $i
}

#----------------------------------------------------------------------------
# Function   : looper
# Description: Commit any changes on a 10 sec clock
# Author     : Tom Wilkason
# Date       : 12/26/2002
#----------------------------------------------------------------------------
proc db::looper {time} {
   db::commit
   after $time db::looper  $time
}
##
# Reset play counts to 0
#
proc db::resetPlayCount {{pattern *}} {
   set files [db::filesInFolder $pattern 100]
   set num [llength $files]
   if {$num > 0} {
         if {[tk_messageBox -icon question -type yesno \
         -title "Please Verify" -message "Are you sure you want to reset $num playcount entries"]=="yes"} {
         foreach {file} $files {
            db::setVal $file Played 0
         }
      }
   }
   return 1
}
##
# Reset play counts to 0
#
proc db::resetPlayTime {{pattern *}} {
   set files [db::filesInFolder $pattern 100]
   set num [llength $files]
   if {$num > 0} {
         if {[tk_messageBox -icon question -type yesno \
         -title "Please Verify" -message "Are you sure you want to reset $num playcount entries"]=="yes"} {
         foreach {file} $files {
            db::setVal $file LastPlay 0
         }
      }
   }
   return 1
}
#----------------------------------------------------------------------------
# Return tracks on an album for cover art
#----------------------------------------------------------------------------
proc db::albumTracks {Artist Album Folder} {
   Trace
   variable DB
   set result [list]             ;# Artist $Artist
   foreach {row} [mk::select $DB Album $Album Folder $Folder -exact Mounted 1] {
      lappend result [mk::get $DB!$row Track Title]
   }
   return $result
}

#------------------------------------------------------------------------------
# Function   : gainControl Namespace
# Description: Functions to access the auto-level gain control
# Author     : Tom Wilkason
# Date       : 2/27/2003
#------------------------------------------------------------------------------
namespace eval gainControl {
}
proc gainControl::levelText {value} {
   if {$::snackAmpSettings(autoLevel) && [string is double -strict $value]} {
      if {[catch {toDb $value} db] } then {
         set db 0.0
      } else {
         if {$db > $::snackAmpSettings(maxGain)} {
            set db $::snackAmpSettings(maxGain)
         } elseif {$db < $::snackAmpSettings(minGain)} {
            set db $::snackAmpSettings(minGain)
         }
      }
      #set result "[string trim [format %-5.2f $db]] dB"
      set result [format "%.2f %s" $db "dB"]
   } else {
      set result "--"
   }
   return $result
}
#------------------------------------------------------------------------------
# Remove all stale equalization settings
#------------------------------------------------------------------------------
proc gainControl::cleanStaleEqualizations {args} {
   return 1
}
proc gainControl::createIfNeeded {fname {snd ""}} {
   Trace
   if {![gainControl::exists $fname]} {
      return [gainControl::create $fname $snd]
   } else {
      return [gainControl::get $fname]
   }
}
##
# Auto Level and Find the ending point of the track
#
proc gainControl::create {fname {snd ""}} {
   Trace
   if {[string length $snd]==0} {
      set snd [soundControl::infoObj $fname]
   } else {
      $snd configure -file $fname
   }
   ##
   # Save the start and stop positions
   #
   set Start [soundControl::findStart $snd]
   if {[string is integer -strict $Start]} {
      db::setVal $fname Start $Start
   } else {
      saLog "gainControl::create Got '$Start' for $fname Start Position"
      set Start 10000
   }
   set Stop [soundControl::findStop $snd]
   if {[string is integer -strict $Stop]} {
      db::setVal $fname Stop $Stop
   } else {
      saLog "gainControl::create Got '$Stop' for $fname Stop Position"
      set Stop 10000
   }
   set gain [soundControl::logPower $snd $::snackAmpSettings(autoLevelSeconds) $Start [expr {[$snd length]-$Stop}]]
   add $fname $gain
   return $gain
}
##
# Return the value for the gain control of a track
#
proc gainControl::get {fname} {
   Trace
   set gain [db::getVal $fname Gain]
   if {[string is double -strict $gain]} {
      return $gain
   } else {
      return -code error -1
   }
}
##
# Add the gaincontrol values for a track
#
proc gainControl::add {fname amount} {
   db::setVal $fname Gain $amount
}
##
# Reset the gain control values for a track
#
proc gainControl::remove {fname} {
   db::setVal $fname Gain "--"
   return 0
}
##
# Check if the gain control values exist for a track
#
proc gainControl::exists {fname} {
   set gain [db::getVal $fname Gain]
   if {[string is double -strict $gain]} {
      return 1
   } else {
      return 0
   }
}
proc gainControl::move {from to} {
   # Not needed
   return 1
}
proc gainControl::saveSettings {args} {
   db::commit
   return 1
}
proc gainControl::restoreSettings {args} {
   return 1
}
proc gainControl::resetAll {{pattern *}} {
   set files [db::filesInFolder $pattern 100]
   set num [llength $files]
   if {$num > 0} {
         if {[tk_messageBox -icon question -type yesno \
         -title "Please Verify" -message "Are you sure you want to remove $num gain entries"]=="yes"} {
         foreach {file} $files {
            remove $file
         }
      }
   }
   return 1
}
##
# Remove some gain control settings
#

proc gainControl::adjustSome {fileNames {percent .1}} {
   global TA LB

   if {[llength $fileNames] == 0} {
      set fileNames [LB::nameSelected]
   }
   foreach {file} $fileNames {
      if {![catch {gainControl::get $file} gain]} {
         set gain [expr {$gain*(1.0+$percent)}]
         db::setVal $file Gain $gain
      } else {
         saLog $gain
      }
   }
}

proc gainControl::resetSome {{fileNames {}}} {
   global TA LB

   if {[llength $fileNames] == 0} {
      set fileNames [LB::nameSelected]
   }
   foreach {file} $fileNames {
      remove $file
   }
}
##
# Take a number of tracks and compute then apply the
# average gain. Used to equalize a complete album
# with the same gain.
#
proc gainControl::averageAlbums {folders} {
   foreach {folder} $folders {
      averageSome [db::filesInFolder $folder]
   }
}
##
# Take a number of tracks and compute then apply the
# average gain. Used to equalize a complete album
# with the same gain.
#
proc gainControl::averageSome {{fileNames {}}} {
   global TA LB

   if {[llength $fileNames] == 0} {
      set fileNames [LB::nameSelected]
   }
   set total 0
   set found 0
   set done [list]
   foreach {file} $fileNames {
      if {[catch {get $file} val]} {
         waitWindow::show "Auto-Leveling [file tail $file]" 1
         set val [gainControl::create $file]
      }
      if {[string is double -strict $val]} {
         set total [expr {$val+$total}]
         incr found
      }
      if {[waitWindow::ifbreak]} {
         break
      }
      lappend done $file
   }
   ##
   # Compute the average and apply that to each of the selected files
   #
   waitWindow::show "Averaging Gains"
   if {$found > 1} {
      set gain [expr {$total/double($found)}]
      foreach {file} $done {
         update
         add $file $gain
      }
   }
   waitWindow::close

}
##
# Take a number of tracks and compute then apply the
# average gain. Used to equalize a complete album
# with the same gain.
#
proc gainControl::computeSome {{fileNames {}}} {
   global TA LB
   if {[llength $fileNames] == 0} {
      set fileNames [LB::nameSelected]
   }
   set total 0
   set found [llength $fileNames]

   foreach {file} $fileNames {
      incr total
      if {![gainControl::exists $file]} {
         waitWindow::percent $total $found "Auto-Leveling\n$file" 1
         set val [gainControl::create $file]
         if {($total%10)==0} {
            db::commit
         }
         if {[waitWindow::ifbreak]} {
            break
         }
      }
   }
   waitWindow::close
}
proc gainControl::getkey {fname} {
   # Use key containing folder, prevents duplicates
   return [file join [file tail [file dirname $fname]] [file tail $fname]]
}
##
# Return the first point in the track that contains music
#
proc gainControl::getStart {fname} {
   set st [db::getVal $fname Start]
   if {$::snackAmpSettings(autoTruncate) && [string is integer -strict $st] && $st > 0} {
      return $st
   } else {
      return 0
   }
}
##
# Return the last point in the track that contains music
#
proc gainControl::getStop {fname {snd ""}} {
   set endp [db::getVal $fname Stop]
   if {$::snackAmpSettings(autoTruncate) && [string is integer -strict $endp] && $endp > 0} {
      return $endp
   } else {
      if {[string length $snd]==0} {
         set snd [soundControl::infoObj $fname]
      }
      return [$snd length]
   }
}

package provide snackMetakit 1.0
