#------------------------------------------------------------------------------
# snackAmp - playList package 1.0 for snackAmp
# provides access to the main playlist
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2003 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#------------------------------------------------------------------------------
set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/playList.tcl,v 1.6 2004/07/17 12:22:33 wilkason Exp $
}
#------------------------------------------------------------------------------
# LB Namespace wrapper for the ListBox
#------------------------------------------------------------------------------
namespace eval LB {
   global snackAmpSettings
   variable indexPlaying -1
   variable webInsert -1
   variable TF
   variable oldPosition 0
   variable fields [list Duration Artist Title]
   set TF() ""
   variable LBdata
}
proc LB::updateFields {} {
   global snackAmpSettings
   variable fields
   ##
   # Listbox fields
   #
   set LB::fields [list]
   foreach num {1 2 3 4} {
      if {[info exists snackAmpSettings(pl$num)] && $snackAmpSettings(pl$num)!="None"} {
         lappend LB::fields $snackAmpSettings(pl$num)
      }
   }

}
#----------------------------------------------------------------------------
# Create the listbox
#----------------------------------------------------------------------------
proc LB::create {W} {
   variable data
   updateFields
   set lb [Scrolled_Listbox $W \
   -selectmode extended \
   -listvariable     LB::LBdata                         \
   -exportselection  no                                 \
   -activestyle      dotbox                             \
   -height           $::snackAmpSettings(playListLines) \
   -foreground       $::snackAmpSettings(fg)            \
   -selectforeground $::snackAmpSettings(fg)            \
   -selectbackground $::snackAmpSettings(highlight)     \
   -background       $::snackAmpSettings(bg)]

   settings::regColor $lb  \
      -foreground       ::snackAmpSettings(fg) \
      -selectforeground ::snackAmpSettings(fg) \
      -selectbackground ::snackAmpSettings(highlight) \
      -background       ::snackAmpSettings(bg)

   # Background validate the playlist tracks
      #after 5000 [list LB::catalogBackground 0 end]

   return $lb
}
proc LB::body {} {
   global LB
   return $LB
}
proc LB::Y {y} {
   global LB
   return $y
}
#------------------------------------------------------------------------------
# Function   : LB::moveByIndex
# Description: Move an entry restore any highlighting it may have
# Author     : Tom Wilkason
# Date       : 6/3/2002
#------------------------------------------------------------------------------
proc LB::moveByIndex {old new} {
   variable indexPlaying
   global LB
   if {$old == $new} {
      return 0
   } elseif {$old < 0} {
      return 0
   }
   set item [$LB get $old]
   set fg  [$LB itemcget $old -foreground]
   set bg  [$LB itemcget $old -background]
   set sfg [$LB itemcget $old -selectforeground]
   set sbg [$LB itemcget $old -selectbackground]
   $LB delete $old
   $LB insert $new $item
   $LB activate $new
   if {[string length $fg]>0} {
      $LB itemconfigure $new -background $bg -foreground $fg
      set indexPlaying $new
   }
   if {[string length $sfg]>0} {
      $LB itemconfigure $new -selectbackground $sbg -selectforeground $sfg
      set indexPlaying $new
   }
   fixWebInsert
}
##
# Make sure the queue insert position is after the playing index
#
proc LB::fixWebInsert {args} {
   variable indexPlaying
   variable webInsert
   if {$indexPlaying >= $webInsert} {
      set webInsert [expr {$indexPlaying+1}]
   }
}
##
# Reset the queue insert position to just after the playing position
#
proc LB::resetWebInsert {args} {
   variable indexPlaying
   variable webInsert
   set webInsert [expr {$indexPlaying+1}]
}
#------------------------------------------------------------------------------
# Make sure some file is viewable
#------------------------------------------------------------------------------
proc LB::seeFile {File} {
   global LB
   if {$File == {}} {
      return 0
   }
   set index [LB::idxFile $File]
   if {$index > -1} {
      $LB see $index
      return 1
   } else {
      return 0
   }
}
#------------------------------------------------------------------------------
# Make sure the playing file is viewable
#------------------------------------------------------------------------------
proc LB::seePlaying {args} {
   global LB
   variable indexPlaying
   if {$indexPlaying > -1} {
      $LB see $indexPlaying
      return 1
   } else {
      $LB see [idxActive]
      return 0
   }
}
#----------------------------------------------------------------------------
# Return the playlist values of the indicies "start ?end?"
#----------------------------------------------------------------------------
proc LB::keyFromIndex {args} {
   global LB
   return [eval $LB get $args]
}
#----------------------------------------------------------------------------
# Return the index at the index specified
#----------------------------------------------------------------------------
proc LB::index {index} {
   global LB
   return [$LB index $index]
}
#----------------------------------------------------------------------------
# Active the entry specified by index
#----------------------------------------------------------------------------
proc LB::activate {index} {
   global LB
   if {[catch {$LB activate $index} result] } then {
      debug $result
   } else {
      $LB see $index
   }
}
#----------------------------------------------------------------------------
# Select the entries specified by the parameters
#----------------------------------------------------------------------------
proc LB::setSelection {args} {
   global LB
   eval $LB selection set $args
}
#------------------------------------------------------------------------------
# Function   : sortSelection
# Description: Sort the selection (only used for playlist)
# Author     : Tom Wilkason
# Date       : 12/4/2001
#------------------------------------------------------------------------------
proc LB::sortSelection {{_listBox ::LB}} {
   upvar $_listBox lb
   global TA
   set selection [lsort -decreasing -integer [$lb curselection]]
   set current [$lb index anchor]
   if {[llength $selection] > 1} {
      ##
      # Cut the data out and sort it
      #
      foreach {ind} $selection {
         lappend data [$lb get $ind]
      }
      $lb delete [lindex $selection end] [lindex $selection 0]
      set data [lsort -decreasing -dictionary $data]
      ##
      # Now paste it back in
      #
      foreach {file} $data {
         $lb insert $current $file
      }
      $lb selection anchor $current
      $lb selection set [lindex $selection end] [lindex $selection 0]
   }
}
#------------------------------------------------------------------------------
# Function   : shuffleSelection
# Description: Shuffle the selection (only used for playlist)
# Author     : Tom Wilkason
# Date       : 12/4/2001
#------------------------------------------------------------------------------
proc LB::shuffleSelection {{_listBox ::LB}} {
   upvar $_listBox lb
   set selection [lsort -decreasing -integer [$lb curselection]]
   set current [$lb index anchor]
   if {[llength $selection] > 1} {
      ##
      # Cut the data out and sort it
      #
      foreach {ind} $selection {
         lappend data [$lb get $ind]
      }
      $lb delete [lindex $selection end] [lindex $selection 0]
      set data [shuffle $data]
      ##
      # Now paste it back in
      #
      foreach {file} $data {
         $lb insert $current $file
      }
      $lb selection anchor $current
      $lb selection set [lindex $selection end] [lindex $selection 0]
   }
}
#------------------------------------------------------------------------------
# Function   : LB::changePlaylistFiles
# Description: Change all of the instances of a file lookup in the main
#              playlist with another. Called when renaming a file.
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc LB::changePlaylistFiles {fileList} {
   variable TF
   set active [LB::idxActive]
   set data [LB::idxSelected]

   foreach {oldFile newFile} $fileList {
      ;# Oops! For database case it has already been renamed
      ;# So we get an invalid oldKey here, just swap the file pointer
      set newKey [LB::nameToKey $newFile]
      set TF($newKey) $newFile
   }
   # Restore selection
   foreach {position} $data {
      LB::setSelection $position
   }
   LB::activate $active
}
#------------------------------------------------------------------------------
# Set the rating for the selected tracks to some value
#------------------------------------------------------------------------------
proc LB::setSelectedRating {value} {
   set files [LB::nameSelected 0]
   foreach {file} $files {
      db::setVal $file Rating $value
   }
}
#------------------------------------------------------------------------------
# After the playlist has been altered, try and restore
# the playing song with highlighting.
# This assumes the current highlighted file
# was removed first.
# TFW: If duplicates exist in the playlist
#      this will find only the first one.
#------------------------------------------------------------------------------
proc LB::tryRestorePlaying {args} {
   global TA snackAmpSettings
   variable TF
   update
   if {![LB::seeFile $TA(currentMF)]} {
#saLog "Can't See $TA(currentMF)"
      LB::activate 0
   } else {
      if {[soundControl::playState]!="Stopped"} {
         LB::inverseVideoFile $TA(currentMF) 1
         if {$snackAmpSettings(moveToTop)} {
#saLog "restoring playing $TA(currentMF) at [LB::idxPlaying]"
            LB::moveByIndex [LB::idxPlaying] 0
            LB::seePlaying
            LB::resetWebInsert
         }
      } else {
#saLog "[soundControl::playState]"
         LB::activate 0 ;# will be last item after insert
      }
   }
}
#------------------------------------------------------------------------------
# Select/Drag, set on mouse click
#------------------------------------------------------------------------------
proc LB::Select {y} {
   global TA LB
   variable TF
   variable oldPosition
   destroy [body].balloon
   set oldPosition [$LB nearest $y]
   set Value [$LB get $oldPosition]
   if {$Value != {} && [info exists TF($Value)]} {
      set TA(entryName) $TF($Value)
      if {$::snackAmpSettings(updateAPL)} {
         catch {APL::SelectFile $TF($Value)}
      }
   }
}
#------------------------------------------------------------------------------
# Drag an entry up/down
#------------------------------------------------------------------------------
proc LB::Drag {y} {
   global TA LB
   variable oldPosition
   set new [$LB nearest $y]
   if {$new == -1} return
   LB::moveByIndex $oldPosition $new
   LB::setSelection $new
   $LB selection anchor $new
   set oldPosition $new
   resetWebInsert  ;# why? Only needed if currently playing is moved
}
#------------------------------------------------------------------------------
# Function   : LB::highlightIndex
# Description: Highlight entry $i in the playlist
# Author     : Tom Wilkason
# Date       : 8/8/2003
#------------------------------------------------------------------------------
proc LB::highlightIndex {i {inverse 0}} {
   global LB
   $LB selection clear 0 end
   set i [expr {$i < 0 ? 0 : $i}]
   set i [expr {$i >= [LB::idxEnd] ? [expr {[LB::idxEnd] - 1}] : $i}]
   LB::setSelection $i
   $LB activate $i
   $LB see $i
   $LB selection anchor $i
   if {$inverse} {
      LB::inverseVideoIndex $i
   }
}
#----------------------------------------------------------------------------
# Mark an entry as invalid (e.g. if the file doesn't exist)
#----------------------------------------------------------------------------
proc LB::invalidateIndex {i} {
   global LB
   if {$i >= 0 && $i < [$LB size]} {
      $LB itemconfigure $i -bg darkgray
   }
}
#------------------------------------------------------------------------------
# Inverse video a file
# note Can get rid of all the LB::inverseVideoFile xx 0 calls, no longer needed
#------------------------------------------------------------------------------
proc LB::inverseVideoFile {File {On 1}} {
   Trace
   variable indexPlaying
   global LB
   if {$On} {
      set index [LB::idxFile $File]
      if {$index >= 0} {
         LB::inverseVideoIndex $index
      }
   } else {
      LB::resetIinverseVideo
   }
}
#------------------------------------------------------------------------------
# Inverse video a selection
# TBD: Can remove the off option
#------------------------------------------------------------------------------
proc LB::inverseVideoIndex {index {On 1}} {
   Trace
   variable indexPlaying
   LB::resetIinverseVideo
   global LB
   if {$index > -1} {
      set bg  [$LB cget -background]
      set fg  [$LB cget -foreground]
      set sbg [$LB cget -selectbackground]
      set sfg [$LB cget -selectforeground]
      ##
      # Color or uncolor the file
      #
      if {$On} {
         $LB itemconfigure $index -background $fg -foreground $bg \
            -selectforeground $sbg -selectbackground $sfg
         set indexPlaying $index
      } else {
         $LB itemconfigure $index -background "" -foreground "" \
            -selectforeground "" -selectbackground ""
      }
   }
   LB::activate $index
   fixWebInsert
}
##
# Reset all of the inverse video (very fast)
# Better than trying to keep track of what is inversed
proc LB::resetIinverseVideo {args} {
   global LB
   set end [$LB size]
   for {set idx 0} {$idx < $end} {incr idx} {
      if {[string length [$LB itemcget $idx -selectforeground]]} {
         $LB itemconfigure $idx -background "" -foreground "" \
            -selectbackground "" -selectforeground ""
      }
   }
}
#----------------------------------------------------------------------------
# Returan all names in the playlist, either the short or full names
#----------------------------------------------------------------------------
proc LB::keyOrNamesAll {{long 0}} {
   global LB
   variable TF
   variable LBdata
   set files [list]
   if {$long} {
      foreach {item} $LBdata {
         lappend files $TF($item)
      }
      return $files
   } else {
      return $LBdata
   }
}
#----------------------------------------------------------------------------
# Return the ending index of the listbox
#----------------------------------------------------------------------------
proc LB::idxEnd {args} {
   global LB
   return [$LB size]
}
#------------------------------------------------------------------------------
# Locate a file in the playlist from the anchor position
#------------------------------------------------------------------------------
proc LB::idxFile {File} {
   variable LBdata
   global LB
   if {$File == {}} {
      return -1
   }
   set lnFile [LB::nameToKey $File]
   set anchor [$LB index active]
   set index [lsearch -exact [$LB get $anchor end] $lnFile]
   if {$index == -1} {
      set index [lsearch -exact [$LB get 0 $anchor] $lnFile]
      if {$index == -1} {
         return -1
      }
   } else {
      # Search must be offset from anchor
      incr index $anchor
   }
   return $index
}
# this is no faster
# proc LB::idxFile1 {File} {
#    variable LBdata
#    global LB
#    if {$File == {}} {
#       return -1
#    }
#    set lnFile [LB::nameToKey $File]
#    set anchor [$LB index active]
#    set index [lsearch -exact [lrange $LBdata $anchor end] $lnFile]
#    if {$index == -1} {
#       set index [lsearch -exact [lrange $LBdata 0 $anchor] $lnFile]
#       if {$index == -1} {
#          return -1
#       }
#    } else {
#       # Search must be offset from anchor
#       incr index $anchor
#    }
#    return $index
# }
#----------------------------------------------------------------------------
# Return the indicies of the selected tracks in the playlist
#----------------------------------------------------------------------------
proc LB::idxSelected {args} {
   global LB
   return [$LB curselection]
}
#----------------------------------------------------------------------------
# Return an array of all the indices in the playlist
#----------------------------------------------------------------------------
proc LB::idxAll {args} {
   global LB
   set result [list]
   set end [$LB size]
   for {set i 0} {$i < $end } {incr i} {
      lappend result $i
   }
   return $result
}
#----------------------------------------------------------------------------
# Return the index of the anchor within the current selection
#----------------------------------------------------------------------------
proc LB::idxAnchor {args} {
   global LB
   return [$LB index anchor]
}
#----------------------------------------------------------------------------
# Return the index of the currently active track
#----------------------------------------------------------------------------
proc LB::idxActive {args} {
   global LB
   return [$LB index active]
}
#----------------------------------------------------------------------------
# Return the index of the currently playing track
#----------------------------------------------------------------------------
proc LB::idxPlaying {args} {
   variable indexPlaying
   return $indexPlaying
}
#----------------------------------------------------------------------------
# Return the full file name at some index, optionally translate
# for removable media.
#----------------------------------------------------------------------------
#------------------------------------------------------------------------------
# Function   : LB::nameToKey
# Description: Return a song name with the appropriate folders prepended
#              Remove the file extension if needed
# Author     : Tom Wilkason
# Date       : 1/3/2001
#------------------------------------------------------------------------------
proc LB::nameToKey {File} {
   variable fields
   global snackAmpSettings TA
   set name [list]
   set result [db::getValues $File $fields]
   # If file not found
   if {[llength $result]==0} {
      return [nameToShort $File]
   }
   ##
   # Duration is always present
   #
   foreach $fields $result {break}
   set row ""
   ##
   # var fields
   #
   set f 0
   foreach {field} $fields {
      switch -- $field {
         None {continue}
         Duration {
            if {$Duration>0} {
               append row "[timify $Duration]   "
            }
         }
         ShortName {
            append row "[nameToShort $File]$TA(spcar)"
            incr f
         }
         default {
            set val [string trim [set $field]]
            if {[string length $val]} {
               append row "$val$TA(spcar)"
               incr f
            }
         }
      }
   }
   # If no tag fields, return from file name
   if {$f} {
      set row [string trimright $row $TA(spcar)]
   } else {
      append row [nameToShort $File]
   }
   return $row
}

#----------------------------------------------------------------------------
# Return the full name of an index, optionally translate for rem media
#----------------------------------------------------------------------------
proc LB::nameFromIdx {Index {translate 0}} {
   global LB
   variable TF
   if {[catch {
      if {$translate} {
         set res [catalog::TranslateMount $TF([$LB get $Index])]
      } else {
         set res $TF([$LB get $Index])
      }
   }]} then {
      set res ""
   }
   return $res
}
#----------------------------------------------------------------------------
# Return the long file name from the track entry and optionally
# translate for removable media.
#----------------------------------------------------------------------------
proc LB::nameFromKey {hash {translate 0}} {
   variable TF
   if {[catch {
      if {$translate} {
         set res [catalog::TranslateMount $TF($hash)]
      } else {
         set res $TF($hash)
      }
   }]} then {
      set res $hash
   }
   return $res
}
#----------------------------------------------------------------------------
# Return the full file name of the the currently active index
#----------------------------------------------------------------------------
proc LB::nameActive {args} {
   global LB
   variable TF
   if {[catch {
      set active [$LB index active]
      set res $TF([$LB get $active])    ;# lookup long name
   }]} then {
      set res ""
   }
   return $res
}
#----------------------------------------------------------------------------
# Return the full file names of the selected items in the playlist,
# optionally translate for removable media.
#----------------------------------------------------------------------------
proc LB::nameSelected {{translate 1}} {
   global LB
   variable TF
   set fullNames [list]
   if {$translate} {
      foreach {Index} [LB::idxSelected] {
         lappend fullNames [catalog::TranslateMount $TF([$LB get $Index])]
      }
   } else {
      foreach {Index} [LB::idxSelected] {
         lappend fullNames $TF([$LB get $Index])
      }
   }
   return $fullNames
}
#----------------------------------------------------------------------------
# Remove all of the entries in the playlist
#----------------------------------------------------------------------------
proc LB::deleteAll {args} {
   global LB
   variable TF
   variable indexPlaying
   variable webInsert
   $LB delete 0 end
   set indexPlaying -1
   set webInsert -1
   if {[info exists TF]} {
      unset TF
      set TF() ""
   }
   gui::updateFileDisplay
}
#----------------------------------------------------------------------------
# Remove an entry at some position
#----------------------------------------------------------------------------
proc LB::deleteIndex {position} {
   global LB
   variable indexPlaying
   variable webInsert
   $LB delete $position
   if {[string is double $position]} {
      if {$indexPlaying >= $position} {
         incr indexPlaying -1
      }
      if {$webInsert >= $position} {
         incr webInsert -1
      }
   }
   return $position
}
#----------------------------------------------------------------------------
# Replace a track entry at position i
#----------------------------------------------------------------------------
proc LB::replaceKeyAtIndex {i newKey} {
   ;# Fast using lset
   global LB
   variable LBdata
   lset LBdata $i $newKey
}
#------------------------------------------------------------------------------
# Function   : LB::removeFile
# Description: Remove all instances of an entry in the playlist
# Author     : Tom Wilkason
# Date       : 8/7/2003
#------------------------------------------------------------------------------
proc LB::removeFile {File} {
   variable TF
   ##
   # Since the playlist can have multiple entries for the same file, we have
   # to take some care and remove each of them from the bottom up.
   #
   set lnFile [LB::nameToKey $File]
   set end [expr {[LB::idxEnd]-1}]
   for {set i $end} {$i >= 0} {incr i -1} {
      set Entry [LB::keyFromIndex $i]
      if {[string equal $Entry $lnFile]}  {
         LB::deleteIndex $i
      }
   }
   if {[info exists TF($lnFile)]} {
      unset TF($lnFile)
   }
}
#------------------------------------------------------------------------------
# Allow a user to insert a file for playing, uses an internal
# pointer to keep track of where the insert should take place.
# It is placed after the last user insert file, once those
# files are played, it is placed just after the currently selected
# file.
#------------------------------------------------------------------------------
proc LB::userInsert {{track ""}} {
   # webInsert should always be at least one more than indexPlaying
   # If not, then make it so.
   variable indexPlaying
   variable webInsert
   if {$track!=""} {
      if {$webInsert > $indexPlaying} {
         # Append to web insert (webInsert is incremented automatically)
         LB::InsertFile $webInsert $track
      } else {
         # Insert just after playing position
         LB::InsertFile [expr {$indexPlaying+1}]  $track
      }
   }
   return $webInsert
}
#----------------------------------------------------------------------------
# Insert a list of items in the playlist following the current user play
# queue position.
#----------------------------------------------------------------------------
proc LB::userInsertList {items} {
   Trace
   # webInsert should always be at least one more than indexPlaying
   # If not, then make it so.
   variable indexPlaying
   variable webInsert
   foreach {track} $items {
      if {$webInsert > $indexPlaying} {
         # Append to web insert (webInsert is incremented automatically)
         LB::InsertFile $webInsert $track
      } else {
         # Insert just after playing position
         LB::InsertFile [expr {$indexPlaying+1}]  $track
      }
   }
   gui::updateFileDisplay
   return [llength $items]
}
#------------------------------------------------------------------------------
# Insert a song at current position of the playlist and then play it
#------------------------------------------------------------------------------
proc LB::InsertPlay {file} {
   Trace
   global TA LB
   variable indexPlaying
   if {$file != {}} {
      ##
      # If file exists in playlist, jump to it instead
      # of inserting it
      #
      set curPost [LB::idxFile $file]
      if {$curPost < 0} {
         set curPost [LB::idxActive]
         incr curPost
         # Insert into playlist adjusting the que position
         LB::insertList $curPost [list $file]
         # If a new incoming file, make sure it is also in the master list
         db::addToDbIfNeeded $file
      } else {
         # This is a hack, adjust the playing index then make sure the queue is adjusted
         set indexPlaying $curPost
         LB::resetWebInsert
      }
      LB::activate $curPost
      Play active
   }
}
#------------------------------------------------------------------------------
# Insert a file into the playlist
#------------------------------------------------------------------------------
proc LB::InsertFile {curPost file} {
   if {$file != {}} {
      # Make sure it is in the master list
      LB::insertList $curPost [list $file]
   }
}

#------------------------------------------------------------------------------
# Apend a file to the bottom of the play list
#------------------------------------------------------------------------------
proc LB::AppendFile {file} {
   if {[string length $file] > 0} {
      LB::insertList end [list $file]
      # If a new incoming file, make sure it is also in the master list
      db::addToDbIfNeeded $file
   }
}
#------------------------------------------------------------------------------
# Append a file to the bottom of the play list
#------------------------------------------------------------------------------
proc LB::AppendList {files} {
   variable TF
   set apList [list]
   foreach {item} $files {
      if {[string length $item] > 0} {
         set key [LB::nameToKey $item]
         if {![info exists TF($key)]} {
            set TF($key) $item
            db::addToDbIfNeeded $item
         }
         lappend apList $key
      # If a new incoming file, make sure it is also in the master list
      }
   }
   if {[llength $apList] > 0} {
      LB::insertList end $apList
   }
}
#------------------------------------------------------------------------------
# Insert a list of entries into the listbox
#------------------------------------------------------------------------------
proc LB::insertList {position files} {
   global LB
   variable TF
   variable indexPlaying
   variable webInsert
   set items [list]
   # Translate to short name and make sure
   # they are in the lookup table
   foreach {item} $files {
      set key [LB::nameToKey $item]
      if {![info exists TF($key)]} {
         set TF($key) $item
      }
      lappend items $key
   }
   # This is a faster insert
   eval $LB insert $position $items
   # Adjust play position here if needed
   if {[string is double $position]} {
      if {$indexPlaying >= $position} {
         incr indexPlaying [llength $items]
      }
      if {$webInsert >= $position} {
         incr webInsert [llength $items]
      }
   }
   gui::updateFileDisplay
}
#------------------------------------------------------------------------------
# Function   : LB::keyToName
# Description: Convert a short playlist name to a full file name
# Author     : Tom Wilkason
# Date       : 8/8/2003
#------------------------------------------------------------------------------
 proc LB::keyToName {Selection} {
    variable TF
#    if {[info exists TF($Selection)]} {
      return $TF($Selection)
#    } else {
#       salog "No $Selection found"
#      return ""
#    }
 }
# proc LB::tout {_var args} {
#    upvar $_var var
#    variable $_var
#    puts  "$_var changed to $var"
# }
# trace variable LB::webInsert w LB::tout
# trace variable LB::indexPlaying w LB::tout
;# This is about the same speed
proc LB::totTime {args} {
   variable LBdata
   set totalTime 0
   foreach row $LBdata {
      if {[scan $row "%d:%d:%d" h m s]==3} {
         set totalTime [expr {$s+$m*60+$h*3600+$totalTime}]
      }
   }
   set days [expr {int($totalTime)/86400}]
   if {$days == 0} {
      return "([timify $totalTime])"
   } else {
      return "($days days [timify $totalTime])"
   }
}


#------------------------------------------------------------------------------
# Make sure all of the tracks in the viewed portion of the playlist are
# cataloged.
#------------------------------------------------------------------------------
proc LB::catalogViewed {args} {
   Trace
   variable LBdata
   global LB
   foreach {top bot} [$LB yview] {
      set len [$LB size]
      set topIndex [expr {int($len*$top)}]
      set botIndex [expr {int($len*$bot)}]
   }
   after idle [list LB::catalogBackground $topIndex $botIndex]
}
##
# TBD: Add break code and ability to restart if
# when the playlist changes (due to sorting or apl apply)
#
proc LB::catalogBackground {topIndex botIndex} {
   Trace
   variable LBdata
   global LB
   ##
   # Scan through each entry and schedule an update
   #
   set sch 0
   for {set i $topIndex} {$i <= $botIndex } {incr i} {
      set row [lindex $LBdata $i]
      if {[string length $row]== 0} {
         break
      }
      if {[scan $row "%d:%d:%d" h m s]==3} {
         set totalTime [expr {$s+$m*60+$h*3600}]
         if {$totalTime==0} {
            LB::catalogIndex $i
            #after 200
            update
            incr sch
         }
      } else {
         LB::catalogIndex $i
         #after 200
         update
         incr sch
      }
   }
   if {$sch>0} {
      after 1000 [list LB::catalogBackground 0 end]
      gui::updateFileDisplay
      puts "Complete indexing playlist with $sch changes"
   } else {
      after 10000 [list LB::catalogBackground 0 end]
   }
   return

   # check again in three seconds
   # TBD: Tie this to a configure or scroll event
}
#------------------------------------------------------------------------------
# Catalog a single index, and update the entry
# Bug; if the same key is reused for multiple files
# can get one missing.
#------------------------------------------------------------------------------
proc LB::catalogIndex {index} {
   variable LBdata
   variable TF
   global LB
   if {[catch {
      set oldKey [lindex $LBdata $index]
      ##
      # Catalog the track
      #
      if {[info exists TF($oldKey)]} {
         set file $TF($oldKey)
         # Cache the track
         if {[db::lookup $file Data 1] > 0} {
            set newKey [LB::nameToKey $file]
            lset LBdata $index $newKey
#puts "fixing $file"
# Don't unset, there may be more entries in the playlist
#            unset TF($oldKey)
            set TF($newKey) $file
         }
      }
   } result] } then {
      saLog $result
   }
}
##
# Remove unkeyed entries from listbox (debugging)
#
proc LB::verify {args} {
   global LB
   variable TF
   for {set i [expr {[$LB size]-1}]} {$i >= 0} {incr i -1} {
      set key [$LB get $i]
      if {![info exists TF($key)]} {
         puts "'$key' not found"
         $LB delete $i
      }
   }
}
proc LB::relayout {args} {
   global LB
   variable TF
   variable indexPlaying
   variable LBdata
   ;# Save current tracks
   set active [LB::idxActive]
   set data [LB::idxSelected]
   set files [list]
   foreach key $LBdata {
      lappend files $TF($key)
   }
   #replace them
   # Translate to short name and make sure
   # they are in the lookup table
   $LB delete 0 end
   Unset LBdata
   Unset TF
   foreach {item} $files {
      set key [LB::nameToKey $item]
      if {![info exists TF($key)]} {
         set TF($key) $item
      }
      lappend LBdata $key
   }
   foreach {position} $data {
      LB::setSelection $position
   }
   LB::activate $active
   LB::inverseVideoIndex $indexPlaying
}
#------------------------------------------------------------------------------
# Callback when dragging the entire playlist
#------------------------------------------------------------------------------
proc LB::dndSrcLBDrag {src} {
   Trace
   set Indicies [LB::idxSelected]
#   if {[llength $Indicies] <= 1} {
#      set Indicies [LB::idxAll]
#   }
   set tracks [list]
   foreach {Index} $Indicies {
      lappend tracks [file nativename [LB::keyToName [LB::keyFromIndex $Index]]]
   }
   return [dndFileList $tracks]
}
package provide playList 1.0

