
#------------------------------------------------------------------------------
# subSort namespace
# Manage the overall shuffling and subsorting for SnackAmp
# Labeled frame to show the various subsorting options
#------------------------------------------------------------------------------
namespace eval subSort {
   variable settings
   variable sortSelection "Default"
   ;# Valid fields to sort on
   array set subsortLookup {
      "Last Played" {LastPlay -integer}
      "Play Count"  {Played   -integer}
      "Rating"      {Rating   -ascii}
      "Votes"       {Votes    -integer}
      "Genre"       {Genre    -ascii}
      "Artist"      {Artist   -ascii}
      "Album"       {Album    -ascii}
      "Folder"      {Folder   -ascii}
      "Year"        {Year     -ascii}
      "Track"       {Track    -dictionary}
      "Date"        {date     -integer}
      "Size"        {size     -integer}
      "Year"        {Year     -dictionary}
      "Sample Rate" {sampRate -real}
      "Bit Rate"    {Rate     -dictionary}
      "Duration"    {Duration -real}
   }

   for {set i 0} {$i < 3} {incr i} {
      set settings(sortHow,$i) "-increasing"
      set settings(sortPlay,$i) "None"
   }
}
proc subSort::dialog {args} {
   ##
   # Sub-sorting radio buttons
   #
   destroy .subsort
   toplevel .subsort
   wm title .subsort "Sorting Dialog"
   pack [subSort::create .subsort 1] -fill x -expand true
   update idletasks
   myIcon .subsort
   raise .subsort
   focus .subsort
   bind .subsort <Return> [list destroy .subsort]
}
##
# Create the user interface inside the pass in frame/toplevel
#
proc subSort::create {W {close 0}} {
   variable settings
   set hull [labeledFrame $W.subsort -text "Sorting Configuration" -bd 2 -relief groove -font listbox]
   #pack $W.subsort -side top -anchor w -fill x
   set todo "None"
   set i 0
   # Some frames
   set guts [frame $hull.[incr i]]
   set bf [frame $hull.[incr i]]
   # Labels

   foreach {stage} [list "Shuffle" "Then Sort/Group" "Then Sort/Group" "Then Sort/Group" "Morph" "Amount"] {
      lappend row [label $guts.lb[incr i] -text $stage]
   }
   eval grid $row -sticky w
   # What to sort on & how to sort
   set row [checkbutton $guts.lb[incr i] -variable subSort::settings(sortShuffle)  -font listbox]
   balloon::define $guts.lb$i "Check this if you want to shuffle the track list prior to sorting on the fields to the right"
   for {set j 0} {$j < 3} {incr j} {
      set opts [concat "None" [lsort [array names subSort::subsortLookup]]]
      listOptionMenu $guts.op$j subSort::settings(sortPlay,$j) $opts
      $guts.op$j configure -font listbox
      lappend row $guts.op$j
      lappend how [checkbutton $guts.cb$j -font listbox\
         -onvalue "-decreasing" -offvalue "-increasing" \
         -text "Decreasing" -variable subSort::settings(sortHow,$j)]
   }
   set snackAmpSettings(postmorph) 0
   set snackAmpSettings(postmorphamount) 25
   lappend row [checkbutton $guts.cbpm -font listbox -variable subSort::settings(postmorph)]
   balloon::define $guts.cbpm "Check this if you want to slightly shuffle the sorted track list, adjust the amount with the scale to the right"
#   lappend row [scale $guts.scpm -from 0 -to 100 -orient horiz \
#      -width 5 -bd 2 -sliderlength 15 -resolution 10 -font balloon -relief groove \
#      -variable subSort::settings(postmorphamount) \
#      -activebackground blue -troughcolor $::snackAmpSettings(bg)]
   lappend row [listOptionMenu $guts.scpm subSort::settings(postmorphamount) {0 10 20 30 40 50 60 70 80 90 100} 3]
   balloon::define $guts.scpm "Adjust the amount of morphing to take place, 0 perc is none, 100 perc is a complete shuffle"
   eval grid $row -sticky w
   eval grid x $how -sticky w
   pack $guts -side top -fill x -expand true

   ##
   # TBD: Add code/widgets to limit the ratings here
   # cb                   mb    ent
   # [] Limit to Ratings [<|>] [__]
   #
   set f [frame $hull.other]
   set togrid [label $hull.other.cblimit -text "Include Ratings"]
   set i 0
   foreach {n} $db::ratingOptions {
      lappend togrid [checkbutton $hull.other.[incr i] -indicatoron 0 -text $n -variable subSort::settings(cbRating,$n) -font listbox]
   }
   eval grid $togrid -sticky w -ipadx 3 -padx 3
   pack $f -side top -fill x -expand true
   # Line
   pack [frame $hull.line -height 2 -borderwidth 2 -relief sunken] -pady 4 -side top -expand true -fill x
   # Buttons at bottom
   # Button frame & buttons
   button $bf.apply  -relief raised -text "Apply"       -width 7 -command {subSort::sort}    -font listbox
   button $bf.save   -relief raised -text "Save"       -width 7 -command {subSort::save}    -font listbox
   button $bf.remove -relief raised -text "Remove"     -width 7 -command {subSort::remove}  -font listbox
   button $bf.saveas -relief raised -text "Save As..." -width 9 -command {subSort::saveas}  -font listbox
   button $bf.close  -relief raised -text "Close"      -width 7 -command [list destroy $W]  -font listbox

   # Menu button to allow selection of sorting options
   menubutton $bf.opt -anchor w -direction flush -width 25 -font listbox\
      -indicatoron 1 -menu $bf.opt.menu -textvariable subSort::sortSelection ;#subSort::sortSelection
   menu $bf.opt.menu \
      -postcommand [list subSort::populate $bf.opt.menu 0]
   pack $bf.save $bf.saveas $bf.remove $bf.opt $bf.apply -side left -padx 2 -pady 1
   pack $bf -side bottom -fill x -expand true
   if {$close} {
      pack $bf.close -side left -padx 3 -pady 1
      $bf.close configure -default active
   }
   return $W.subsort
}
##
# Populate the subsorting frame
#
proc subSort::getAllModes {menu} {
   variable sortSelection
   $menu delete 0 end
   ##
   # Standard Options
   #
   foreach {item} $filtering::sortModes {
      $menu add radiobutton -label $item \
         -value $item -variable ::snackAmpSettings(sortModeTxt) \
         -command {filtering::sortPlaylist}
   }
   ##
   #  Extended Options
   #
   set choices [subSort::getFiles]
   if {[llength $choices] > 0} {
      $menu add separator
      foreach {choice} $choices {
         $menu add radiobutton -label $choice -variable ::snackAmpSettings(sortModeTxt) -command {
            set subSort::sortSelection $::snackAmpSettings(sortModeTxt)
            subSort::readFile $::snackAmpSettings(sortModeTxt)
            filtering::sortPlaylist
         }
      }
   }
   $menu add separator
   $menu add command -label "Configure Sorting..." -command {subSort::dialog}
}

#------------------------------------------------------------------------------
# List of all sorting modes
#------------------------------------------------------------------------------
proc subSort::allSortModes {args} {
   foreach {item} $filtering::sortModes {
      lappend results $item
   }
   foreach {item} [subSort::getFiles] {
      lappend results $item
   }
   return $results
}
##
# Populate the subsorting frame
#
proc subSort::populate {menu showFileModes} {
   variable sortSelection
   $menu delete 0 end
   ##
   #  Extended Options
   #
   set choices [subSort::getFiles]
   foreach {choice} $choices {
      $menu add radiobutton -label $choice -variable subSort::sortSelection -command {
         subSort::readFile $subSort::sortSelection
      }
   }
}

#------------------------------------------------------------------------------
# Return the sorting modes from the gui to be used in a query
# up to three lists, each list like {Played -integer -increasing}
#------------------------------------------------------------------------------
proc subSort::getModes {args} {
   global snackAmpSettings
   set tosort [list]
   foreach {index} [lsort [array names ::snackAmpSettings sortPlay,*]] {
     set item $::snackAmpSettings($index)
     set num [lindex [split $index ,] 1]
     if {[string length $item] && $item != "None"} {
        foreach {field type} $subSort::subsortLookup($item) {break}
        lappend tosort [list $field $type $snackAmpSettings(sortHow,$num)]
     }
   }
   return $tosort
}
##
# Sort the playlist
#
proc subSort::sort {args} {
   filtering::sortPlaylist
}
##
# Copy current settings to snackAmpDefaults
#
proc subSort::apply {{name {}}} {
   variable settings
   variable sortSelection
   array set ::snackAmpSettings [array get settings]
   if {$name==""} {
      set ::snackAmpSettings(sortSelection) $sortSelection
      set ::snackAmpSettings(sortModeTxt)   $sortSelection
   } else {
      set ::snackAmpSettings(sortSelection) $name
      set ::snackAmpSettings(sortModeTxt)   $name
   }
}
##
# Get the current snackAmpSettings (during init)
#
proc subSort::fetch {args} {
   variable settings
   variable sortSelection
   array set settings [array get ::snackAmpSettings sortHow*]
   array set settings [array get ::snackAmpSettings sortPlay*]
   array set settings [array get ::snackAmpSettings sortShuffle]
   array set settings [array get ::snackAmpSettings postmorph*]
   array set settings [array get ::snackAmpSettings cbRating,*]
   if {[info exists ::snackAmpSettings(sortSelection)]} {
      set sortSelection $::snackAmpSettings(sortSelection)
   }
}
##
# Return a list of all subsorting file names
#
proc subSort::getFiles {args} {
   variable settings
   set result [list]
   foreach name [lsort [glob -nocomplain -directory $::HOME *.ss]] {
      lappend result [file tail [file root $name]]
   }
   return $result
}

#------------------------------------------------------------------------------
#
#------------------------------------------------------------------------------
proc subSort::saveas {args} {
   set name ""
   if {[getEntryDialog "Enter Name" name "Sort/Shuffle Setting"]=="ok"} {
      saveFile $name
   }
}
##
# Save the current settings to the file
#
proc subSort::save {{name ""}} {
   variable sortSelection
   saveFile $sortSelection
}
proc subSort::remove {{name ""}} {
   variable sortSelection
   if {[removeFile $sortSelection]} {
      set sortSelection ""
   }
}
##
# Write a settings file
#
proc subSort::saveFile {{name ""}} {
   variable sortSelection
   variable settings
   if {$name==""} {
      if {[getEntryDialog "Enter Name" name "Sort/Shuffle Setting"]!="ok"} {
         return 0
      }
   }
   set fname [file join $::HOME $name.ss]
   # Check for file exists
   if {[catch {
      set fid [open $fname w]
   } result] \
   } then {
      tk_messageBox -icon warning -type ok -title "Open Error" -message $result
      return 0
   } else {
      puts $fid [array get settings]
      close $fid
      set sortSelection $name
      apply $name
      return 1
   }
}
proc subSort::removeFile {{name ""}} {
   variable sortSelection
   variable settings
   set fname [file join $::HOME $name.ss]
   if {[file exists $fname]} {
      file delete -force $fname
   }
   if {[file exists $fname]} {
      return 0
   } else {
      return 1
   }
}
##
# Read a settings file into settings
#
proc subSort::readFile {name} {
   variable settings
   variable sortSelection
   set fname [file join $::HOME $name.ss]
   if {[catch {
      set fid [open $fname r]
   } result] \
   } then {
      tk_messageBox -icon warning -type ok -title "Open Error" -message $result
      return 0
   } else {
      set len [gets $fid row]
      close $fid
      if {$len > 0} {
         array set settings $row
         set sortSelection $name
         apply $name
         return 1
      }  else {
         return 0
      }
   }
}
;# Get latest settings from snackamp
subSort::fetch
package provide subSorting 1.0
