proc id3Tag::id3V2Test {file _data} {
   Update
   upvar $_data data
   variable ID3_2_3_types
   variable ID3_2_2_types
   variable id3v1Lookup
   variable v2_2_IDs
   variable v2_3_IDs
   set result [list]
   if {[catch {open $file r} fid]} {
      return $fid
   }
   # Have ID, get data
   fconfigure $fid -translation binary -encoding binary
   if {[catch {set block [read $fid 10]}] } then {
      close $fid
      return $result
   }
# set info {
# The first part of the ID3v2 tag is the 10 byte tag header, laid out
#    as follows:
#
#      0-2 ID3v2/file identifier      "ID3"
#      3-4 ID3v2 version              $04 00  (V2.4.0)
#      5   ID3v2 flags                %abcd0000
#      6-9 ID3v2 size             4 * %0xxxxxxx
#
#    The first three bytes of the tag are always "ID3", to indicate that
#    this is an ID3v2 tag, directly followed by the two version bytes. The
#    first byte of ID3v2 version is its major version, while the second
#    byte is its revision number. In this case this is ID3v2.4.0. Version or
#    revision will never be $FF.
# }
   if {[string equal [string range $block 0 2] "ID3"]} {
      set Triplets [list]
      # Determine the frame length and read the rest of the id3 header
      binary scan [string range $block 3 9] "h1h1h1c1c1c1c1" majVer minVer flag a b c d
      # 7 bits of each byte are used for a total of 28 bits (256Mb)
      # clamp at 1M (2^20) (for safety)
      set length [expr {($a<<21 | $b<<14 | $c<<7 | $d) & 0xFFFFF}]
      if {$length <= 0} {
         close $fid
         return 0
      }
      # read and trim padding
      set block [string trimright [read $fid $length] \0]
      set length [string length $block]  ;# should this be bytelength?
      close $fid

      # determine tag version
      if {[string equal "$majVer.$minVer" "2.0"]} {
         set 3byte 1
         set types $ID3_2_2_types
         set validIDs $v2_2_IDs
      } else {
         set types $ID3_2_3_types
         set validIDs $v2_3_IDs
         set 3byte 0
      }
      # Scan for tags and save their location
      # Note: This can be a bit slow if the tag space is large
      # May want to bail if we have the basic tags
      # Note: Since there is no tag order in the ID3V2
      #       we have to search all of it for each tag
      #       (there may be multiple instances of the same tag)
      # A faster method is to rely on the lengths and "walk" the tag chain.
      foreach {type} $types {
         set offset 0
         set maxLoops 20
         while {[incr maxLoops -1]>0} {
            set loc [string first $type $block $offset]
            if {$loc >= 0} {
               # add len of tag to offset for search
               if {$3byte} {
                  binary scan [string range $block [expr {$loc+3}] [expr {$loc+6}] ] "c1c1c1" b c d
                  set taglen [expr {(($b&0xFF)<<16 | ($c&0xFF)<<8 | ($d&0xFF))& 0xFFFFF}]
                  set headlen 5
               } else {
                  binary scan [string range $block [expr {$loc+4}] [expr {$loc+8}] ] "c1c1c1c1" a b c d
                  # syncsafe version below
                  #set taglen [expr {(($a&0x7F)<<21 | ($b&0x7F)<<14 | ($c&0x7F)<<7 | ($d&0x7F))& 0xFFFFF}]
                  set taglen [expr {(($a&0xFF)<<24 | ($b&0xFF)<<16 | ($c&0xFF)<<8 | ($d&0xFF))& 0xFFFFF}]
                  set headlen 9
               }
               # Handle invalid lengths in the tag
               if {$taglen <= 0 || $taglen > $length} {
                  incr taglen
               }
               set offset [expr {$loc+$taglen}] ;# start of next search
               if {[lsearch -sorted -exact $validIDs $type] == -1} {
                  set data($type) [string range $block $loc [expr {$offset+$headlen}]]
               }
            } else {
               # no more found
               break
            }
         } ;# end while
      }
      return 1
   } else {
      close $fid
      return 0
   }
}
foreach {file} {
   {//Bertha/Music/Rock/Country Rock/David Crosby/If I Could Only Remember My Name/02 - Cowboy Movie.mp3}
   {//Bertha/Media/Jazz/Bop/Joshua Redman/Wish/01 - Turnaround.mp3}
   {//Bertha/Music/Rock/Country Rock/Neil Young/Decade/09 - The Old Laughing Lady.mp3}
    {//Kids/Music/Rock/Adult Contemporary/Tori Amos/Tales of a Librarian/03 - Silent All These Years.mp3} {//Kids/Music/Rock/Adult Contemporary/Tori Amos/Tales of a Librarian/04 - Cornflake Girl.mp3} {//Bertha/Media/Jazz/Bop/Joshua Redman/Wish/04 - The Deserving Many.mp3} {//Kids/Music/Holiday/Christmas/We Three Kings/05 - Christmas Passing Throu.mp3} {//Kids/Music/Rock/Adult Contemporary/Tori Amos/Tales of a Librarian/05 - Mary.mp3} {//Bertha/Media/Jazz/Bop/Joshua Redman/Wish/05 - We Had a Sister.mp3}
} {
   array unset data
   id3Tag::id3V2Test $file data
   if {[array size data]>0} {
      puts $file
      parray data
   }
}

